#!/usr/bin/python
#coding: iso-8859-1;

############################################################################
#                                                                          #
#    This file 'Harvest.py'                                                #
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################


import SPyRO as spyro
import warn
import shelve
import os, sys
#import config
import commands
import time
import sre
import urllib

from monarca import mime as MIME

def filenametourl(path):
    return "file://localhost/"  + path.rstrip()

class AppendDocument:
    def __init__(self, harvest = None, index = None, host = "www2.eluniversal.com.mx"):
        self.harvest = harvest
        self.index = index
        self.host = host

    def insertlist(self, filenamelist, harvest = None, index = None, host = None):
        for name in filenamelist:
            name = name.rstrip()
            print name
            self.insert(name, name, [], harvest, index, host)

    def insert(self, name, keywords, comment, harvest = None, index = None, host = None):

        datafile = file(name, 'r')
        url = self.insertstream(datafile, filenametourl(name), comment, harvest, index, host)
        datafile.close()
        return url

    def insertstream(self, stream, name, keywords, comment, harvest = None,
                   index = None, host = None):
        if harvest is None: harvest = self.harvest
        if index is None: index = self.index
        if host is None: host = self.host

        data = stream.readlines()
        if stream and len(data) > 0:
            while True:
                try:
                    line = data.pop(0)
                except:
                    print repr(data)
                    break
                # print repr(line)
                if line == '\n': break
                keywords.append(line.rstrip())
            datatext = "".join(data)
        else:
            datatext = comment
        
        last_fetched = time.time()
        #url = "push://" + host + "/" + name
        url = name
        site = {'url' : url,
                'last_fetched': last_fetched,
                'keywords': keywords, }
        harvest.save(site, datatext, "text/plain", False)
        try:
            datalist = (  {'data': datatext},
                          {'data': " ".join(keywords),
                           'wordreg': r'[^\s,]+',
                           'stemmer': False,
                           'stopwords': False,
                           #'normalizer': None
                           },
                          )
            index.vectorizechunkedandappend(url, datalist)
        except:
            warn.warn()
            return url
        return url
    
def runinclient(config, listofcontainers, listoffiles = ()):
    """
    |config| The configuration object
    |listofcontainers| A list of files with a list of image files
    |listoffiles| A list of image files
    """
    comm = spyro.TClient(config.SPyRO_Server, "/SPyRO?PICKLE2")
    harvest = comm.getobj("Repository")
    index = comm.getobj("realindex")
    for filename in listofcontainers:
        f = file(filename)
        lines = f.readlines()
        f.close()
        if len(lines) > 0:
            add = AppendDocument(harvest, index)
            add.insertlist(lines)
    if len(listoffiles) > 0:
        add = AppendDocument(harvest, index)
        add.insertlist(listoffiles)
    comm.close()

def runinserver(config, listofcontainers, listoffiles = ()):
    """
    |config| The configuration object
    |listofcontainers| A list of files with a list of image files
    |listoffiles| A list of image files
    """
    comm = spyro.TClient(config.SPyRO_Server, "/SPyRO?PICKLE2")
    appdoc = comm.getobj("AppendDocument")
    for filename in listofcontainers:
        f = file(filename)
        lines = f.readlines()
        f.close()
        if len(lines) > 0:
            appdoc.insertlist(lines)
    if len(listoffiles) > 0:
        appdoc.insertlist(listoffiles)
    comm.close()

if __name__ == '__main__':
    try:
        import psyco
        psyco.profile()
    except:
        warn.debug("doesn't exists the psyco module")

    import news
    runinserver(news.config, ['list-of-files.txt','list-of-files2.txt'])
