#!/usr/bin/python

import os, sys, sre, commands
import fecha
import warn
import urllib

_NEW_LINE = sre.compile(r"\n",sre.L)
_SPACES = sre.compile(r"\s+",sre.L)
_NOT_NOTE = sre.compile(r"[\[\]\|]",sre.L)
_RegBusqueda = sre.compile(r"B.squeda en:")
_RegTime = sre.compile(r"\d\d:\d\d")

def cleaner(data):
        lines = []
        paragraphs = []
        lineini = 0
        linenum = 0
        while True:
            try:
                line = data.pop(0)
            except:
                break
            line = line.rstrip()
            if line == "References": break
            line = line.lstrip()
            if _RegBusqueda.match(line): continue
            ## Paragraph's Identifier
            if _NOT_NOTE.search(line):
                continue
            if _RegTime.match(line):
                data.insert(0, line[5:])
                data.insert(0, '')
                line = line[:5]
            try:
                if line == '' == lines[-1]:
                    continue
            except:
                pass
            linenum = linenum + 1
            #print "OK: ", linenum, line
            if line == '':
                if lineini > 0:
                    paragraphs.append((lineini,linenum-1))
                lineini = 0
            else:
                if lineini == 0:
                    lineini = linenum - 1
            lines.append(line)
        return lines, paragraphs

class NewsNote:
    """ Basic class to retrieve news article info """
    def __init__(self,data=None):
        self.feed(data)

    def feed(self, data):
        """ Feeds the parser """
        self.data = data

    def feedclean(self, data):
        """ Feeds the parser """
        self.data = data

class ElUniversalOnline(NewsNote):
    def __init__(self, data = None):
        NewsNote.__init__(self, data)
        self._author = None

    def feed(self, data):
	self.data,self.paragraphs = cleaner(data)
	try:
	    self._parse()
	except:
	    self.keywords = {}

    def feedclean(self, data):
        linenum = 0
        lineini = 0
        paragraphs = []
        lines = []
        for line in data:
            line = line.rstrip()
            linenum = linenum + 1
            if line == '':
                if lineini > 0:
                    paragraphs.append((lineini,linenum-1))
                lineini = 0
            else:
                if lineini == 0:
                    lineini = linenum - 1
        self.data = data
        self.paragraphs = paragraphs
        try:
            self._parse()
        except:
            self.keywords = {}
        
    def _parse(self):
        counter = 0
        ini,fin = self.paragraphs[counter]
        keywords = {}
        if fin - ini == 1: # Getting the section
            keywords['section'] = self._get_paragraph(counter)
            counter = counter + 1
        ini,fin = self.paragraphs[counter]
        if fin - ini > 1: # Getting the abtract
            keywords['abstract'] = self._get_paragraph(counter)
            counter = counter + 1
        c = 0
        while True:
            x = self._get_paragraph(counter)
            if len(x) > 2:
                break
            counter = counter + 1
        if len(x) > 4:
            keywords['agency'] = x[0]
            c = 1
        try:
            keywords['journal'] = x[c]
            keywords['location'] = x[c+1]
            keywords['date'] = fecha.parse_spanish(x[c+2])
            keywords['body'] = self.data
        except:
            keywords['general'] = x
        self.keywords = keywords
    
    def _get_paragraph(self, number):
        ini,fin = self.paragraphs[number]
        return self.data[ini:fin]

    def save(self, fp):
        keys = self.keywords
        s = keys.pop('body')
        try:
            del keys['abstract']
        except:
            pass
        for k,v in keys.items():
            if k == "general":
                fp.write( "".join(v) )
            else:
	        try:
		    fp.write( k + ":" + urllib.quote("".join(v).rstrip()) + "\n" )
		except:
		    pass
        fp.write("\n")
        fp.write("".join(s))

def ElUniversalSaveFile(outdir, note, univ):
    print note
    newname = os.path.sep.join((outdir,os.path.basename(note)))
    if newname == note:
        warn.debug("cannot overwrite file")
	sys.exit(1)
    try:
        f = file(newname,"w")
	univ.save(f)
	f.close()
    except:
        warn.warn("Fail")

def ElUniversalParseListFiles():
    from cStringIO import StringIO
    import os
    if len(sys.argv) < 3:
        print "%s notes-list.txt outputdir" %( sys.argv[0] )
        sys.exit(1)
    n = None
    outdir = sys.argv[2]
    for note in file(sys.argv[1]).readlines():
        note = note.rstrip()
        datafile = file(note)
        datafile = StringIO(commands.getoutput("lynx -dump %s"%(note)))
        univ = ElUniversalOnline(datafile.readlines())
        datafile.close()
	ElUniversalSaveFile(outdir, note, univ)

def ElUniversalParseListClean():
    import os
    if len(sys.argv) < 3:
        print "%s notes-list.txt outputdir" %( sys.argv[0] )
        sys.exit(1)
    n = None
    outdir = sys.argv[2]
    for note in file(sys.argv[1]).readlines():
        note = note.rstrip()
        datafile = file(note)
        data = datafile.readlines()
        datafile.close()
        univ = ElUniversalOnline()
        univ.feedclean(data)
	ElUniversalSaveFile(outdir, note, univ)
    
if __name__ == '__main__':
    import sys, os
    ElUniversalParseListClean()
