#!/usr/bin/python
#coding: iso-8859-1;

############################################################################
#                                                                          #
#    This file 'Crawler.py'                                                #
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import Harvest
#import htmllib as html
from HTMLParser import HTMLParser
import sre
import shttp
from shttp.uri import URI
import warn

valid_html = sre.compile(r"\.([sxd]?html?|php\d?|)$",sre.I)

def default_validator(parser, url):
    site = parser.site
    if valid_html.match(url) \
           or site.path == '' or site.path[-1] == '/' \
           or site.mime.find("html") != -1 \
           or url.find("?") != -1:
        return url
    else:
        return None
    
class Parser(HTMLParser):
    def __init__(self, validator = None, links = None, site = None):
        HTMLParser.__init__(self)
        self.validator = validator
        if links is None:
            self.links = {}
        else:
            self.links = links
        self.site = site

    def reset(self, urlbase = '', links = None, site = None):
        HTMLParser.reset(self)
        self.urlbase = urlbase
        self.site = site
        if links is None:
            self.links = {}
        else:
            self.links = links

    removepage = sre.compile(r"/[^/]+$")
    #nonvalid = {'mailto', 'ssh', 'ed2k','javascript'}
    nonserverreq = sre.compile(r"^[\d\w\.]{1,20}:")
    def _append_real_url(self, href):
        if not URI.match(href):
            if Parser.nonserverreq.match(href): return
            base = URI.match(self.urlbase)
            gd = base.groupdict()
            if len(href) == 0: return
            port = gd['port']
            if port != '': port = ':' + port
            host = '%s://%s%s'%(gd['protocol'], gd['host'], port)
            if href[0] == '?':
                href = "%s%s"%(gd['path'] or '/',href)
            elif href[0] != '/':
                href = self.removepage.sub("",gd['path'] or '') + '/' + href
            href = host + href
        #print "*********"*10
        #print "CRAWLER A:",self.urlbase, href
        if self.validator:
            href = self.validator(self, href)
            if not href: return
        frompagelist = self.links.get(href)
        #print "CRAWLER B:",self.urlbase, href
        if frompagelist is None:
            frompagelist = [self.urlbase]
            self.links[href] = frompagelist
        #else:
        #    frompagelist.append(self.urlbase)
    
    def handle_starttag(self, tag, attrs):
        if tag == 'a':
            # This for is useful to use pyrex
            href=None
            for attr in attrs:
                if attr[0] == 'href':
                    self._append_real_url(attr[1])
                    break
    
class Crawler:
    def __init__(self, harvest, seeds=None, validator=None,
                 linkmemcachesize=0):
        if validator is None:
            self.validator = default_validator
        else:
            self.validator = validator
        self.harvest = harvest
        self.links = {}
        self.resetparser()
        self.linkmemcachesize = linkmemcachesize
        if seeds: harvest.regurllist(seeds)

    def resetparser(self):
        self.parser = Parser(self.validator, self.links)
    
    def callback(self, event, url):
        if event != "Saved": return
        warn.log("Crawling: ", url)
        try:
            p = self.parser
            site,data = self.harvest.getandopen(url)
            p.reset(url, self.links, site)
            p.feed(data)
            p.close()
        except:
            warn.warn("Crawler error ",url)
            self.resetparser()
        if len(self.links) > self.linkmemcachesize:
            self.flush()
    
    def flush(self):
        if len(self.links) < 1: return
        self.harvest.regurllist(self.links.keys())
        self.links.clear()

    def close(self):
        self.flush()
    
if __name__ == '__main__':
    class X:
        def regurllist(self, x):
            print "regurllist>\n", "\n".join(x)
            
    def open(self, o):
        return open('/home/sadit/information_retrieval/spydi/crawlertest.html').read()

    x = X()
    c = Crawler(x,('file://localhost/home/sadit/information_retrieval/spydi/crawlertest.html',))
    c.callback("file://localhost/home/sadit/information_retrieval/spydi/crawlertest.html", None)
