#!/usr/bin/python
#coding: iso-8859-1;

############################################################################
#                                                                          #
#    This file 'Harvest.py'                                                #
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################


from threadutils import PeriodicExecution, PCExecution
import warn
from dbsupport import bsddb3
import shelve
import time
import httplib, urllib, urllib2, urlparse
import os, sys
import distributor

class Harvest:
    """ Queue to harvest sites. The harvest can be done by timer or petition.
    The timed harvesting can be done by configuration (every |N| seconds or a
    specified time (evey Monday at 12:00 PM, in a python format)), solved
    under petition (greedy).    
    The Queue can be controled remotely with the Remote Object Protocol"""

    def __init__(self, reprouter, harvestEvery=36000,
                 harvestCounter=100, harvestCounterTime=30):
        self.reprouter = reprouter
        # Harvest All every |harvestEvery| seconds
        self.replaceTimer(harvestEvery, harvestEvery, False)
        # Harvest every |harvestCounter| operations
        # or |harvestCounterTime| seconds. Only harvest the first
        # |harvestCounter| sites in the queue
        self.replaceCounter(harvestCounter, harvestCounterTime, False)
        # More than zero (initial values of SiteInformation)

    def replaceTimer(self,begin, every, stop=True):
        self.harvestEvery = every
        self.harvestBegin = begin
        if stop: self.timer.stop()
        self.timer = PeriodicExecution(self.harvest, begin, every)

    def replaceCounter(self,counter, countertimer, stop=True):
        self.harvestCounter = counter
        self.harvestCounterTime = countertimer
        if stop: self.countertimer.stop()
        self.countertimer = PeriodicExecution(self.__harvestcounter,
                   self.harvestCounterTime, self.harvestCounterTime)

    def close(self):
        self.countertimer.stop()
        self.timer.stop()
    
    def regurllist(self, urls):
        self.reprouter.route(None, urls, "regurllist", urls)

    def regurl(self, url):
        #warn.debug("Put URL", url)
        self.reprouter.route(None, url, "regurl", url)
        #self.countertimer.changed()

    def callback(self, action, url):
        warn.debug("Callback %s: %s"%(action, url))
        if action == 'XRegistered':
            self.countertimer.changed()

    def __harvestcounter(self):
        """ Control the harvesting of the repositories """
        try:
            self.reprouter.route(None, None, 'harvest', None, self.harvestCounter)
        except Exception, e:
            warn.warn(e)
    
    def harvest(self):
        """ Control the harvesting of the repositories """
        try:
            self.reprouter.route(None, None, 'harvest')
        except Exception, e:
            warn.warn(e)

    # The Repository Interface
    def getandopen(self, url):
        return self.reprouter.routeone(None, url, "getandopen", url)

    def open(self, url):
        return self.reprouter.routeone(None, url, "open", url)

    def get(self, url, default = None):
        return self.reprouter.routeone(None, url, "get", url, default)

    def save(self, site, data, mime, docallback=True):
        try:
            url = site.url
        except:
            url = site['url']
        self.reprouter.route(None, url, "save", site, data, mime, docallback)

if __name__ == '__main__':
    def filenametourl(path):
        return "file://localhost"  + path.rstrip()
    from Repository import Repository
    distcallback = distributor.Dummy()
    distrepository = distributor.Dummy()
    rep = Repository('datafiles/repository/', distcallback)
    distrepository.register(router.Local(rep), None)
    harvest = Harvest(router.Router(distrepository),200000,300)
    distcallback.register(router.Local(harvest), None)
    harvest.regurllist(
        ('http://www.google.com/', "http://lsc.fie.umich.mx/~sadit/",
         "http://scfie.fie.umich.mx/", "http://scfie.fie.umich.mx/~sadit"
         "file://localhost/home/sadit/information_retrieval/spydi/TODO"))
        
    f = file('list-of-files.txt')
    harvest.regurllist(map(filenametourl, f.readlines()))
    f.close()
    harvest.harvest()
    distrepository.close()
    harvest.close()
    rep.close()
