#include <Python.h>

#include "structmember.h"
#include "modsupport.h"
#include "tupleobject.h"

#include <idx_man.h>
#include <document_env.h>
#include <document.h>
#include <cache1.h>
#include <query.h>
#include <vector_model.h>
#include <boolean_model.h>
#include <text_trie.h>
#include <super_malloc.h>
#include <id_wrd_t.h>
#include <id_doc_t.h>
#include <string.h>
/*==========================================*/
/***********  pycache1_object   *************/
/*------------------------------------------*/
typedef struct {
  PyObject_HEAD
  cache1 *c;
} pycache1;

#define idx_ReturnNone() Py_INCREF(Py_None); return Py_None

static void general_dealloc(PyObject* self)
{
  /*fprintf(stderr, "\n\n******* DEALLOC on: '%s', '%d'\n",
    self->ob_type->tp_doc, self->ob_refcnt);*/
  self->ob_type->tp_free((PyObject*)self);
}

static PyObject *NEW(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
  PyObject *self;
  self = type->tp_alloc(type, 0);
  /*fprintf(stderr, "\n\n******* ALLOC on: '%s', '%d'\n",
    self->ob_type->tp_doc, self->ob_refcnt);*/
  return self;
}

static int
pycache1_init(pycache1 *self, PyObject *args, PyObject *kwds)
{
  char *name;
  int perm, spages, dpages, divs;
  
  static char *kwlist[] = {"name", "perm", "spages", "dpages", "divs",NULL};
  if (! PyArg_ParseTupleAndKeywords(args, kwds, "siiii", kwlist, 
				    &name, &perm, &spages, &dpages, &divs)) {
    return -1;
  }
  cache1_open(&self->c, name, perm, spages, dpages, divs);
  return 0;
}

#define MEMBER_SENTINEL   {NULL, 0, 0, 0, NULL}  /* Sentinel */
#define METHOD_SENTINEL   {NULL, NULL, 0, NULL}  /* Sentinel */

static PyMemberDef pycache1_members[] = {
    {"cache", T_INT, offsetof(pycache1, c), 0, "cache1 pointer"},
    MEMBER_SENTINEL
};

PyObject *pycache1_close(pycache1 *self, PyObject *args)
{
  cache1_close(self->c);
  idx_ReturnNone(); /*We don't need Py_INCREF in newer versions of python*/
}

PyObject *pycache1_sync(pycache1 *self, PyObject *args)
{
  cache1_sync_all_pages(self->c);
  idx_ReturnNone(); /*We don't need Py_INCREF in newer versions of python*/
}

static PyMethodDef pycache1_methods[] = {
  {"close", (PyCFunction)pycache1_close, METH_NOARGS, "Close object"},
  {"sync", (PyCFunction)pycache1_sync, METH_NOARGS, "Sync cache "},
  METHOD_SENTINEL  
};

static PyTypeObject pycache1Type = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "idx.cache1",              /*tp_name*/
    sizeof(pycache1),          /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)general_dealloc, /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Wrapper cache1",          /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    pycache1_methods,             /* tp_methods */
    pycache1_members,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)pycache1_init,      /* tp_init */
    0,                         /* tp_alloc */
    NEW,                 /* tp_new */
};


/*============================================*/
/*************  pyid_doc_object   *************/
/*--------------------------------------------*/
typedef struct {
  PyObject_HEAD
  id_doc_t id_doc;
} pyid_doc;


static int
pyid_doc_init(pyid_doc *self, PyObject *args, PyObject *kwds)
{
  int minor, major;  
  static char *kwlist[] = {"major", "minor",NULL};

  if (!PyArg_ParseTupleAndKeywords(args, kwds, "ii", kwlist, &major, &minor)) {
    return -1;
  }
  self->id_doc.minor = minor;
  self->id_doc.major = major;
  return 0;
}

PyObject *pyid_doc_pair(pyid_doc *self, PyObject *args, PyObject *kwds) {
  return Py_BuildValue("(ii)",self->id_doc.major, self->id_doc.minor);
}

PyObject *pyid_doc_set_minor(PyObject *self, PyObject *args, PyObject *kwds) {
  int minor;
  static char * kwlist[] = {"minor", NULL};
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "i", kwlist, &minor)) {
    return NULL;
  }
  ((pyid_doc *)self)->id_doc.minor = minor;
  Py_INCREF(self);
  return self;
}

PyObject *pyid_doc_set_major(pyid_doc *self, PyObject *args, PyObject *kwds) {
  int major;
  static char * kwlist[] = {"major", NULL};
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "i", kwlist, &major)) {
    return NULL;
  }
  self->id_doc.major = major;
  return self;
}

PyObject *pyid_doc_get_minor(pyid_doc *self, PyObject *args, PyObject *kwds) {
  return Py_BuildValue("i",self->id_doc.minor);
}
PyObject *pyid_doc_get_major(pyid_doc *self, PyObject *args, PyObject *kwds) {
  return Py_BuildValue("i",self->id_doc.major);
}

static PyMemberDef pyid_doc_members[] = {
  {"id_doc", T_INT, offsetof(pyid_doc, id_doc), 0, "id_doc_t pointer"},
  MEMBER_SENTINEL
};

static PyMethodDef pyid_doc_methods[] = {
  {"pair", (PyCFunction)pyid_doc_pair, METH_NOARGS,
   "returns a pair (major,minor)"},
  {"set_minor", (PyCFunction)pyid_doc_set_minor, METH_VARARGS|METH_KEYWORDS,
   "set minor"},
  {"set_major", (PyCFunction)pyid_doc_set_major, METH_VARARGS|METH_KEYWORDS,
   "set major"},
  {"get_major", (PyCFunction)pyid_doc_get_major, METH_NOARGS,
   "get major"},
  {"get_minor", (PyCFunction)pyid_doc_get_minor, METH_NOARGS,
   "get minor"},
  METHOD_SENTINEL  
};

static PyTypeObject pyid_docType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "idx.id_doc",              /*tp_name*/
    sizeof(pyid_doc),          /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)general_dealloc, /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Wrapper id_doc_t",          /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    pyid_doc_methods,             /* tp_methods */
    pyid_doc_members,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)pyid_doc_init,      /* tp_init */
    0,                         /* tp_alloc */
    NEW,                 /* tp_new */
};


/*============================================*/
/**************  pydocument_env  **************/
/*--------------------------------------------*/
typedef struct {
  PyObject_HEAD
  document_env *env;
} pydocument_env;


/*#define PYDOCENV_TRIE -99*/
#define PYDOCENV_INCR 0
#define PYDOCENV_AINC 1
#define PYDOCENV_NAME 2

static int
pydocument_env_init(pydocument_env *self, PyObject *args, PyObject *kwds)
{
  int size, perm, tipo;
  pycache1* cache;
  char *basename;

  static char *kwlist[] = {"cache", "perm", "docenv_type",
			   "basename", NULL};
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "Oiis", kwlist, 
				   &cache, &perm, &tipo, &basename)) {
    return -1;
  }

  switch (tipo) {
  case PYDOCENV_NAME:
    document_env_open_unique_name(&self->env, cache->c, perm, basename);
    break;
  case PYDOCENV_INCR:
    document_env_open_increment(&self->env, cache->c, perm, basename);
    break;
  case PYDOCENV_AINC:
    document_env_open_always_increment(&self->env, cache->c, perm, basename);
    break;
  default:
    fprintf(stderr, "Tipo de document_env desconocido: %d\n", tipo);
    return -1;
  }
  return 0;
}

static PyMemberDef pydocument_env_members[] = {
  {"env", T_INT, offsetof(pydocument_env, env), 0, "document_env pointer"},
  MEMBER_SENTINEL
};


PyObject *pydocument_env_close(pydocument_env *self, PyObject *args)
{
  self->env->close(self->env);
  idx_ReturnNone();
}

PyObject *pydocument_env_sync(pydocument_env *self, PyObject *args)
{
  document_env_sync(self->env);
  idx_ReturnNone();
}

PyObject *pydocument_env_get_id(pydocument_env *self, PyObject *args)
{
  char *name;
  id_doc_t id;
  if (!PyArg_ParseTuple(args, "s", &name)) return NULL;
  self->env->get_id(self->env, name, &id);
  return Py_BuildValue("ii",id.major,id.minor);
}

PyObject *pydocument_env_get_name(pydocument_env *self, PyObject *args)
{
  pyid_doc *doc;
  char *name;
  PyObject *pyname;
  if (!PyArg_ParseTuple(args, "O", &doc)) return NULL;
  name = self->env->get_name(self->env, &doc->id_doc, NULL, 0);
  pyname = Py_BuildValue("s",name);
  free(name);
  return pyname;
}

static PyMethodDef pydocument_env_methods[] = {
  {"sync", (PyCFunction)pydocument_env_sync, METH_NOARGS,
   "sync()"},
  {"close", (PyCFunction)pydocument_env_close, METH_NOARGS,
   "close()"},
  {"get_id", (PyCFunction)pydocument_env_get_id, METH_VARARGS,
   "pointer get_id(name). Get Id of doc with name |name|"},
  {"get_name", (PyCFunction)pydocument_env_get_name, METH_VARARGS,
   "str get_name(iddoc). Get name of doc with id |iddoc|"},
  METHOD_SENTINEL
};

static PyTypeObject pydocument_envType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "idx.document_env",              /*tp_name*/
    sizeof(pydocument_env),          /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)general_dealloc, /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Wrapper document_env",          /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    pydocument_env_methods,             /* tp_methods */
    pydocument_env_members,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)pydocument_env_init,      /* tp_init */
    0,                         /* tp_alloc */
    NEW,                 /* tp_new */
};

/*============================================*/
/*************  pyidx_man_object   *************/
/*--------------------------------------------*/
typedef struct {
  PyObject_HEAD
  idx_man *idm;
} pyidx_man;


static int
pyidx_man_init(pyidx_man *self, PyObject *args, PyObject *kwds)
{
  char *name;
  pycache1 *idx_cache, *dat_cache;
  int perm, pages, mutex_level = 1;
  static char *kwlist[] = {"name","idx_cache","dat_cache","perm","pages",
			   "mutex_level",NULL};
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "sOOii|i", kwlist,
				   &name,&idx_cache,&dat_cache,&perm,&pages)) {
    return -1;
  }
  text_trie_open_master(&self->idm, name, perm, pages,
			LONGITUD_ALFABETO,mutex_level);
  text_trie_set_caches(self->idm, idx_cache->c, dat_cache->c);
  return 0;
}

static PyMemberDef pyidx_man_members[] = {
    {"idm", T_INT, offsetof(pyidx_man, idm), 0, "idx_man pointer"},
    MEMBER_SENTINEL  /* Sentinel */
};


PyObject *pyidx_man_close(pyidx_man *self, PyObject *args)
{
  text_trie_close_caches(self->idm);
  self->idm->close(self->idm);
  idx_ReturnNone();
}

PyObject *pyidx_man_sync(pyidx_man *self, PyObject *args)
{
  text_trie *trie = (text_trie *)self->idm;
  cache1_sync_all_pages(trie->atrie->cache);
  idx_ReturnNone();
}

static PyMethodDef pyidx_man_methods[] = {
  {"close", (PyCFunction)pyidx_man_close, METH_NOARGS, "Close object"},
  {"sync", (PyCFunction)pyidx_man_sync, METH_NOARGS, "Synchronize object"},
  METHOD_SENTINEL
};

static PyTypeObject pyidx_manType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "idx.idx_man",              /*tp_name*/
    sizeof(pyidx_man),          /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)general_dealloc, /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Wrapper idx_man",          /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    pyidx_man_methods,             /* tp_methods */
    pyidx_man_members,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)pyidx_man_init,      /* tp_init */
    0,                         /* tp_alloc */
    NEW,             /* tp_new */
};

/*============================================*/
/*************  pydocument_object   *************/
/*--------------------------------------------*/
typedef struct {
  PyObject_HEAD
  document *doc;
  PyObject *string;
  int doctype;
} pydocument;

#define PYDOCUMENT_STRING 3333
#define PYDOCUMENT_FILE 444444

static int
pydocument_init(pydocument *self, PyObject *args, PyObject *kwds)
{
  char *string;
  int len, blen;
  static char *kwlist[] = {"string", "doctype", "bufflen", NULL};
  self->doctype = PYDOCUMENT_STRING;
  blen = 4096;
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|ii", kwlist, &self->string,
				   &self->doctype, &blen)) return -1;
  PyString_AsStringAndSize(self->string, &string, &len);
  switch (self->doctype) {
  case PYDOCUMENT_STRING:
    Py_INCREF(self->string);
    document_open_string(&self->doc, string, len);
    break;
  case PYDOCUMENT_FILE:
    document_open_file(&self->doc, string, blen);
    break;
  default:
    fprintf(stderr,"pydocument's type doesn't exists\n");
    return -1;
  }
  return 0;
}

static PyMemberDef pydocument_members[] = {
    {"doc", T_INT, offsetof(pydocument, doc), 0, "document pointer"},
    MEMBER_SENTINEL  /* Sentinel */
};

PyObject *pydocument_close(pydocument *self, PyObject *args)
{
  self->doc->close(self->doc);
  if (self->doctype == PYDOCUMENT_STRING) {
    Py_DECREF(self->string);
  }
  idx_ReturnNone();
}

PyObject *pydocument_insert(pydocument *self, PyObject *args, PyObject *kwds)
{
  id_wrd_t wordread;
  word_freq *wfreq;
  pyidx_man *idm;
  pyid_doc *doc_id;
  pydocument_env *env;
  int freq = 0;
  skip_list *wlist;
  skip_node *wnode;

  static char *kwlist[] = {"env", "idm", "docid", NULL};
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "OOO", kwlist,
				   &env, &idm, &doc_id)) {
    return NULL;
  }
  wordread.key = (char *)super_malloc(sizeof(char)*64);
  wlist = document_read_document(self->doc, &wordread, &doc_id->id_doc);
  free(wordread.key);
  wnode = wlist->head;
  fprintf(stderr,"DOC: \n");
  while (skip_node_next(wnode)->ntype == NODE) {
    wnode = skip_node_next(wnode);
    wfreq = (word_freq *)wnode->data;
    wordread.key = wfreq->wrd;
    wordread.freq = wfreq->num_docs;
    wordread.len = strlen(wfreq->wrd);
    id_wrd_encode(&wordread, ENCODING_TABLE);
    idm->idm->insert(idm->idm, &wordread, &doc_id->id_doc);
    if (wfreq->num_docs> freq) freq = wfreq->num_docs;
  }
  document_read_close(wlist);
  env->env->set_maxfreq(env->env, &doc_id->id_doc, freq);
  idx_ReturnNone();
}

static PyMethodDef pydocument_methods[] = {
  {"close", (PyCFunction)pydocument_close, METH_NOARGS, "Close object"},
  {"insert", (PyCFunction)pydocument_insert,METH_VARARGS|METH_KEYWORDS,"Insert into index"},
  METHOD_SENTINEL  /* Sentinel */
};

static PyTypeObject pydocumentType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "idx.document",              /*tp_name*/
    sizeof(pydocument),          /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)general_dealloc, /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Wrapper document",          /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    pydocument_methods,             /* tp_methods */
    pydocument_members,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)pydocument_init,      /* tp_init */
    0,                         /* tp_alloc */
    NEW,             /* tp_new */
};

/*============================================*/
/**********  pyquery_result_object   **********/
/*--------------------------------------------*/
typedef struct {
  PyObject_HEAD
  query_result *qres;
} pyquery_result;

static PyMemberDef pyquery_result_members[] = {
  {"qres", T_INT, offsetof(pyquery_result, qres), 0, "query_result pointer"},
  MEMBER_SENTINEL  /* Sentinel */
};

PyObject *pyquery_result_close(pyquery_result *self, PyObject *args)
{
  query_result_free(self->qres);
  idx_ReturnNone();
}

static PyMethodDef pyquery_result_methods[] = {
  {"close", (PyCFunction)pyquery_result_close, METH_NOARGS, "Close object"},
  METHOD_SENTINEL
};

static PyTypeObject pyquery_resultType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "idx.query_result",              /*tp_name*/
    sizeof(pyquery_result),          /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)general_dealloc, /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Wrapper query result",          /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    pyquery_result_methods,             /* tp_methods */
    pyquery_result_members,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    0,                         /* tp_init */
    0,                         /* tp_alloc */
    NEW,        /* tp_new */
};



/*============================================*/
/*************  pyquery_object   *************/
/*--------------------------------------------*/
typedef struct {
  PyObject_HEAD
  query *q;
  document_env *env;
} pyquery;


static int
pyquery_init(pyquery *self, PyObject *args, PyObject *kwds)
{
  pyidx_man *idm;
  PyObject *tuple;
  pydocument_env *env;
  char **argv;
  int argc, i;
  static char *kwlist[] = {"env","idm","kwdtuple",NULL};
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "OOO", kwlist,
				   &env, &idm, &tuple)) {
    return -1;
  }
  argc = PySequence_Size(tuple);
  if (argc < 1) return 1;
  argv = super_malloc(argc * sizeof(char *));
  for (i = 0; i < argc; i++) {
    argv[i] = strdup(PyString_AsString(PySequence_GetItem(tuple, i)));
  }
  self->env = env->env;
  query_new(&self->q, env->env, idm->idm, argv, argc);
  /* fprintf(stderr, ">>>>INIT_QUERY %x, %x\n",self,self->q);*/
  /* free(argv);  moved to query_ext_close */
  return 0;
}

static PyMemberDef pyquery_members[] = {
  {"query", T_INT, offsetof(pyquery, q), 0, "query pointer"},
  MEMBER_SENTINEL  /* Sentinel */
};

PyObject *pyquery_close(pyquery *self, PyObject *args)
{
  int i;
  query *q = self->q;
  for (i = 0; i < q->argc; i++) free(q->argv[i]);
  free(q->argv);
  query_close(q);
  idx_ReturnNone();
}

PyObject *pyquery_reset(pyquery *self, PyObject *args)
{
  int i;
  query_reset(self->q);
  idx_ReturnNone();
}

PyObject *pyquery_initialize(pyquery *self, PyObject *args, PyObject *kwds)
{
  int find_all = 0;
  static char *kwlist[] = {"findAll",NULL};
  if (! PyArg_ParseTupleAndKeywords(args, kwds, "|i", kwlist, &find_all)) {
    return NULL;
  }
  /*fprintf(stderr, "Q_INIT:> %p, %p\n", self, self->q);*/
  return Py_BuildValue("i", query_init(self->q, find_all));
}

PyObject *pyquery_set_collection_size(pyquery *self, PyObject *args, PyObject *kwds)
{
  int size;
  static char *kwlist[] = {"size",NULL};
  if (! PyArg_ParseTupleAndKeywords(args, kwds, "i", kwlist, &size)) {
    return NULL;
  }
  self->q->collection_size = size;
  idx_ReturnNone();
}

PyObject *pyquery_get_collection_size(pyquery *self, PyObject *args,
				      PyObject *kwds)
{
  return Py_BuildValue("i",self->q->collection_size);
}

PyObject *pyquery_get_gdata(pyquery *self, PyObject *args, PyObject *kwds)
{
  int i;
  query *q = self->q;
  PyObject *gdata = PyTuple_New(q->argc);
  for (i = q->argc-1; i >= 0; i--) {
    PyTuple_SetItem(gdata, i,
		    Py_BuildValue("s(iii)", q->argv[i],
				  /* max freq in all documents (fake)*/
				  q->gdata[i].global_freq, 
				  /* accumulated freq */
				  q->gdata[i].global_freq,
				  /* Number of documents */
				  q->gdata[i].number_docs));
  }
  /* incremetar referencia ? */
  return gdata;
}

PyObject *pyquery_set_gdata(pyquery *self, PyObject *args, PyObject *kwds)
{
  int i;
  PyObject *gdatalist, *gdataitem, *item;
  PyObject *gdatadict = PyDict_New();
  static char *kwlist[] = {"gdata", NULL};

  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O", kwlist, &gdatalist)) {
    return NULL;
  }
  PyDict_MergeFromSeq2(gdatadict, gdatalist, 1);
  for (i = 0; i < self->q->argc; i++) {
    /*gdataitem = PySequence_GetItem(gdatalist, i);*/
    item = PyDict_GetItemString(gdatadict, self->q->argv[i]);
    /*item = PySequence_GetItem(gdataitem, 1);*/
    self->q->gdata[i].global_freq=
      PyLong_AsLongLong(PySequence_GetItem(item, 1));
    self->q->gdata[i].number_docs=
      PyLong_AsLongLong(PySequence_GetItem(item, 2));
  }
  Py_DECREF(gdatadict);
  idx_ReturnNone();
}

PyObject *_pyquery_get_results_info(query *q, query_result *qr,
				  document_env *env, int count)
{
  id_doc_t *doc_id;
  skip_list_iterator *itr;
  char *name;
  int i,_size;
  float sim;
  _size=512;
  name = super_malloc(_size);
  PyObject *all;
  skip_list_iterator_new(qr->docs, &itr);
  all = PyTuple_New(qr->docs->num_nodes > count ? count : qr->docs->num_nodes);
  /*fprintf(stderr, "num_nodes: %d\n", qr->docs->num_nodes);*/
  while (skip_list_iterator_has_next(itr) && count-- != 0) {
    /*fprintf(stderr, " i: %d\n", count);*/
    qr->get_info(qr, skip_list_iterator_next(itr), &doc_id, &sim);
    env->get_name(env, doc_id, name, _size);
    PyTuple_SetItem(all, itr->position - 1, Py_BuildValue("(sf)", name, sim));
  }
  free(name);
  skip_list_iterator_close(itr);
  return all;
}

#define _PYQUERY_OR "rank_or"
#define _PYQUERY_AND "rank_and"
#define _PYQUERY_VECTOR "rank_vector"

PyObject *pyquery_doquery(pyquery *self, PyObject *args, PyObject *kwds)
{
  query_result *res;
  PyObject *result;
  int begin, count;
  char *qtype;
  static char *kwlist[] = {"begin","count","qtype",NULL};
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "iis", kwlist, &begin, &count,
				   &qtype)) {
    return NULL;
  }
  if ( strncmp(qtype, _PYQUERY_OR, 20) == 0) {
    res = boolean_model_or(self->q, begin, count);
  } else if (strncmp(qtype, _PYQUERY_AND, 20) == 0) {
    res = boolean_model_and(self->q, begin, count);
  } else {
    res = vector_model(self->q, begin, count);
  }
  result = _pyquery_get_results_info(self->q, res, self->env, count);
  query_result_free(res);
  return result;
}

PyObject *pyquery_or(pyquery *self, PyObject *args, PyObject *kwds)
{
  int begin, count;
  static char *kwlist[] = {"begin","count",NULL};
  pyquery_result *res;
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "ii", kwlist, &begin, &count)) {
    return NULL;
  }
  res = (pyquery_result*) NEW(&pyquery_resultType, NULL, NULL);
  /*Py_INCREF(res);*/
  res->qres = boolean_model_or(self->q, begin, count);
  return res;
}

PyObject *pyquery_vector(pyquery *self, PyObject *args, PyObject *kwds)
{
  int begin, count;
  static char *kwlist[] = {"begin","count",NULL};
  pyquery_result *res;
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "ii", kwlist, &begin, &count)) {
    return NULL;
  }
  res = (pyquery_result*) NEW(&pyquery_resultType, NULL, NULL);
  /*Py_INCREF(res);*/
  res->qres = vector_model(self->q, begin, count);
  return res;
}

PyObject *pyquery_and(pyquery *self, PyObject *args, PyObject *kwds)
{
  int begin, count;
  static char *kwlist[] = {"begin","count",NULL};
  pyquery_result *res;
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "ii", kwlist, &begin, &count)) {
    return NULL;
  }
  res = (pyquery_result*) NEW(&pyquery_resultType, NULL, NULL);
  /*Py_INCREF(res);*/
  res->qres = boolean_model_and(self->q, begin, count);
  return res;
}

PyObject *pyquery_get_results_info(pyquery *self,
				   PyObject *args, PyObject *kwds)
{
  pyquery_result *pqr;
  static char *kwlist[] = {"queryresult",NULL};
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "OO", kwlist, &pqr)) {
    return NULL;
  }
  return _pyquery_get_results_info(self->q, pqr->qres, self->env, -1);
}

static PyMethodDef pyquery_methods[] = {
  {"close", (PyCFunction)pyquery_close, METH_NOARGS, "Close object"},
  {"reset", (PyCFunction)pyquery_reset, METH_NOARGS, "Reset object"},
  {"init", (PyCFunction)pyquery_initialize, METH_VARARGS|METH_KEYWORDS,
   "Initialize query"},
  {"setCollectionSize", (PyCFunction)pyquery_set_collection_size,METH_VARARGS|METH_KEYWORDS, "Set the real size of collection (on distributed collections). "
   "Must be called after |init| method"},
  {"getCollectionSize", (PyCFunction)pyquery_get_collection_size,METH_VARARGS|METH_KEYWORDS, "Returns the collection size "},
  {"setgdata", (PyCFunction)pyquery_set_gdata,METH_VARARGS|METH_KEYWORDS, "Set global (system wide) information to this query"},
  {"doquery", (PyCFunction)pyquery_doquery,METH_VARARGS|METH_KEYWORDS, "Get the global information of this query"},
  {"getgdata", (PyCFunction)pyquery_get_gdata,METH_VARARGS|METH_KEYWORDS, "Get the global information of this query"},
  {"boolean_or", (PyCFunction)pyquery_or, METH_VARARGS|METH_KEYWORDS,
   "Query OR"},
  {"boolean_and", (PyCFunction)pyquery_and, METH_VARARGS|METH_KEYWORDS,
   "Query AND"},
  {"vector_model", (PyCFunction)pyquery_vector, METH_VARARGS|METH_KEYWORDS,
   "Vector model ranking"},
  {"get", (PyCFunction)pyquery_get_results_info, METH_VARARGS|METH_KEYWORDS,
   "Query get results"},
  METHOD_SENTINEL  /* Sentinel */
};

static PyTypeObject pyqueryType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "idx.query",              /*tp_name*/
    sizeof(pyquery),          /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)general_dealloc, /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Wrapper query",          /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    pyquery_methods,             /* tp_methods */
    pyquery_members,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)pyquery_init,      /* tp_init */
    0,                         /* tp_alloc */
    NEW,             /* tp_new */
};


static PyMethodDef module_methods[] = {
  /* Add suport to new document types (string, file, etc)*/
  {NULL}  /* Sentinel */
};

#ifndef PyMODINIT_FUNC	/* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif

PyMODINIT_FUNC
initSPyDI(void) 
{
    PyObject* m;
    if (PyType_Ready(&pycache1Type) < 0) return;
    if (PyType_Ready(&pyid_docType) < 0) return;
    if (PyType_Ready(&pydocument_envType) < 0) return;
    if (PyType_Ready(&pyidx_manType) < 0) return;
    if (PyType_Ready(&pydocumentType) < 0) return;
    if (PyType_Ready(&pyquery_resultType) < 0) return;
    if (PyType_Ready(&pyqueryType) < 0) return;
    
    m = Py_InitModule3("SPyDI", module_methods, "Index wrapper module");
    if (m == NULL) return;
    Py_INCREF(&pycache1Type);
    PyModule_AddObject(m, "cache1", (PyObject *)&pycache1Type);
    Py_INCREF(&pyid_docType);
    PyModule_AddObject(m, "id_doc", (PyObject *)&pyid_docType);
    Py_INCREF(&pydocument_envType);
    PyModule_AddObject(m, "document_env", (PyObject *)&pydocument_envType);
    Py_INCREF(&pyidx_manType);
    PyModule_AddObject(m, "idx_man", (PyObject *)&pyidx_manType);
    Py_INCREF(&pydocumentType);
    PyModule_AddObject(m, "document", (PyObject *)&pydocumentType);
    Py_INCREF(&pyquery_resultType);
    PyModule_AddObject(m, "query_result", (PyObject *)&pyquery_resultType);
    Py_INCREF(&pyqueryType);
    PyModule_AddObject(m, "query", (PyObject *)&pyqueryType);
    PyModule_AddObject(m, "DOCUMENT_STRING",
		       Py_BuildValue("i",PYDOCUMENT_STRING));
    PyModule_AddObject(m, "DOCUMENT_FILE", Py_BuildValue("i",PYDOCUMENT_FILE));

/*  PyModule_AddObject(m, "DOCENV_TRIE", Py_BuildValue("i",PYDOCENV_TRIE));*/
    PyModule_AddObject(m, "DOCENV_INCR", Py_BuildValue("i",PYDOCENV_INCR));
    PyModule_AddObject(m, "DOCENV_AINC", Py_BuildValue("i",PYDOCENV_AINC));
    PyModule_AddObject(m, "DOCENV_NAME", Py_BuildValue("i",PYDOCENV_NAME));
    PyModule_AddObject(m, "QUERY_OR", Py_BuildValue("i",_PYQUERY_OR));
    PyModule_AddObject(m, "QUERY_AND", Py_BuildValue("i",_PYQUERY_AND));
    PyModule_AddObject(m, "QUERY_VECTOR", Py_BuildValue("i",_PYQUERY_VECTOR));
}
