#include <alist.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <cache1.h>
#include <super_malloc.h>
#include <math.h>
#include <debug.h>

int alist_close (alist *list)
{
  cache1_put_element(list->cache, &list->ilast, list->modified);
  cache1_put_element(list->cache, &list->ihead, list->modified);
  pthread_mutex_destroy(&list->mutex);
  free(list);
  return 0;
}

void alist_init(alist *list)
{
  int l;
  list->modified = 0;
  l = list->cache->header.chunk_size-sizeof(alist_head);
  if (list->head->element_size > 0) {
    list->elements_per_node = (int)floor(l*1.0/list->head->element_size);
  } else {
    list->head->element_size = l;
    list->elements_per_node = 1;
  }
}

int alist_new(alist **n, cache1 *c, size_t elem_size)
{
  alist_head *head;
  alist *list = struct_malloc(alist);
  pthread_mutex_init(&list->mutex, NULL);
  cache1_get_new_element(c, (void **)&head, &list->ihead);
  list->cache = c;
  list->head = head;
#ifdef _DEVELOP_VERSION
  if (sizeof(alist_head) != sizeof(alist_node)) {
    fprintf(stderr, "sizeof head: %d, node: %d\n",sizeof(alist_head),
	    sizeof(alist_node));
    fprintf(stderr, "Los tamaos de los nodos de alista son incorrectos\n");
    exit(-1);
  }
#endif
  head->ilast = list->ihead;
  head->total = 1;
  head->element_size = elem_size;
  alist_init(list);
  if (list->elements_per_node == 0) {
    cache1_put_element(list->cache,&list->ihead, 0);
    list->modified = 0;
#ifdef _DEVELOP_VERSION
    fprintf(stderr,"alist_new: lista sin elementos");
    exit(-1);
#endif
    return ALIST_NO_ELEMENTS;
  }
  memset(&head->inext, 0, sizeof(index_t));
  head->maxpos = -1;
  list->ilast = head->ilast;
  cache1_get_by_index(c, (void **)&list->last, &list->ilast);
  *n = list;
  return 0;
}

int alist_open(alist **l, const index_t* list_head, cache1 *c)
{
  alist *list = struct_malloc(alist);
  pthread_mutex_init(&list->mutex, NULL);
  list->cache = c;
  list->ihead = *list_head;
  cache1_get_by_index(c, (void **)&list->head, list_head);
  list->ilast = list->head->ilast;
  cache1_get_by_index(c, (void **)&list->last, &list->ilast);
  alist_init(list);
  *l = list;
  return 0;
}

int alist_put_element(alist *list, const index_t *idx, int modified)
  
{
  return cache1_put_element(list->cache, idx, modified); 
}

/* Inserta un nodo, al final de la lista */
int alist_insert_new_node(alist *list, index_t *idx)
{
  alist_node *node;
  cache1_get_new_element(list->cache, (void **)&node, idx);
  idx->pos = 0;
  /* Los valores del nuevo nodo para conocer a los otros */
  node->iprev = list->ilast;
  node->ihead = list->ihead;
  memset(&node->inext, 0, sizeof(index_t));
  /*node->magic = 0;*/
  node->maxpos = -1;
  /* Poniendo los valores de los otros nodos al nuevo last */
  list->last->inext = *idx;
  /* Se cierra anterior ltimo, para dejar abierto al nuevo ltimo,
     antes de cambiar la referencia */
  cache1_put_element(list->cache, &list->ilast, 1);
  cache1_set_modified(list->cache, &list->ihead);
  list->modified = 1;
  list->ilast = list->head->ilast = *idx;
  list->last = node;
  list->head->total++;
  return 0;
}

/* Obtiene un elemento de la lista, se accede como arreglo
   (el tiempo de acceso es constante). El puntero *data apuntara
   a la posicin en memoria del dato pedido. La referencia al
   nodo utilizado queda abierta por lo que es necesario regresarla con
   cache1_put_element o alist_put_element */

#define ALIST_GET_BY_INDEX(_list, _data, _idx, _node) \
  cache1_get_by_index(_list->cache, (void **)&_node, _idx); \
  *(_data) = ALIST_DATA_POINTER(_node) + \
  (ALIST_GET_POS(_idx)*_list->head->element_size);


int alist_get_by_index(alist *list, void **data, const index_t* idx)
{
  alist_node *node;
  ALIST_GET_BY_INDEX(list,data,idx,node);
  /****** en macro para usarlo tambin en alist_iterator_jump *******
  cache1_get_by_index(list->cache, (void **)&node, idx);
  *data = ALIST_DATA_POINTER(node)+ALIST_GET_POS(idx)*list->head->element_size;
  */
  return 0;
}

int alist_get_new_element_intern(alist *list, void **data, index_t *idx)
{
  /*mutex lock*/
  list->last->maxpos++;
  if (list->last->maxpos == list->elements_per_node) {
    list->last->maxpos--;
    alist_insert_new_node(list, idx);
    *data = ALIST_DATA_POINTER(list->last);
    list->last->maxpos=0;
  } else {
    *data = (void *) ALIST_DATA_ELEM(list->last,list->head,list->last->maxpos);
    *idx = list->ilast;
    //idx->pos = 0;
    ALIST_SET_POS(idx, list->last->maxpos);
  }
  /* Adicionando referencia y marcando como modificada */
  cache1_page_set_modified(cache1_get_by_index(list->cache, (void **)&list->last, idx));
  /*mutex unlock*/
  return 0;
}

int alist_get_new_element(alist *list,void **data,index_t *idx)
{
  int r;
  pthread_mutex_lock(&list->mutex);
  r = alist_get_new_element_intern(list,data,idx);
  pthread_mutex_unlock(&list->mutex);
  return r;
}

int alist_insert_new(alist *list, void *data, index_t *idx, signed int size)
{
  void *pointer = NULL;
  /*   alist_get_new_element(list, &pointer, idx); */
  /*   memcpy(pointer, data, size); */
  /*   alist_put_element(list, idx); */
  /*   return 0; */
  u_int32_t t = list->head->element_size;
  if (size <= t){
    alist_get_new_element(list,&pointer,idx);
    memcpy(pointer,data,size);
    alist_put_element(list,idx,1);
    return 0;
  } else {
    pthread_mutex_lock(&list->mutex);
    while(size > 0){
      alist_get_new_element_intern(list,&pointer,idx);
      memcpy(pointer,data,t < size ? t : size);
      data = (char *)data + t;
      size = size - t;
      alist_put_element(list,idx,1);
    }
    pthread_mutex_unlock(&list->mutex);
    return 0;
  }
}

#define ALIST_NUMBER_OF_ELEMENTS(_list, _bytes) \
((int) ceil(_bytes * 1.0 / (_list)->head->element_size))

int alist_number_of_elements(alist *list, int bytes)
{
  return ALIST_NUMBER_OF_ELEMENTS(list, bytes);
}

int alist_iterator_reset(iterator *itr)
{
  alist_iterator *i = (alist_iterator *) itr;
  i->pos = -1;
  cache1_put_element(i->list->cache, &i->inode, i->modified);
  i->modified = 0;
  i->inode = i->list->ihead;
  cache1_get_by_index(i->list->cache, (void **)&i->node, &i->inode);
  return 1;
}

int alist_iterator_new(alist *list, iterator **itr, int close_list_when_close)
{
  alist_iterator *ai = struct_malloc(alist_iterator);
  iterator *i = (iterator *)ai;
  ai->list = list;
  ai->inode = list->ihead;
  cache1_get_by_index(list->cache, (void **)&ai->node, &ai->inode);
  ai->pos = -1;
  ai->close_list = close_list_when_close;
  ai->modified = 0; /* Peligroso */
  i->next = alist_iterator_next;
  i->prev = alist_iterator_prev;
  i->recover_element = alist_iterator_recover_element;
  i->current = alist_iterator_current;
  i->has_next = alist_iterator_has_next;
  i->has_prev = alist_iterator_has_prev;
  i->remove = alist_iterator_delete;
  i->close = alist_iterator_close;
  i->jump = alist_iterator_jump;
  i->insert = alist_iterator_insert;
  i->skip = alist_iterator_skip;
  i->skip_by_bytes = alist_iterator_skip_by_bytes;
  i->reset = alist_iterator_reset;
  i->get_current_index = alist_iterator_get_current_index;
  i->find = alist_iterator_find;
  i->find_forward = alist_iterator_find_forward;
  i->find_backward = alist_iterator_find_backward;
  *itr = i;
  return 0;
}

int alist_iterator_close(iterator *itr)
{
  alist_iterator *i = (alist_iterator *) itr;
  cache1_put_element(i->list->cache, &i->inode, i->modified);
  i->modified = 0;
  if (i->close_list) alist_close(i->list);
  free(i);
  return 0;
}

void * alist_iterator_insert(iterator *itr)
{
  alist_iterator *i = (alist_iterator *) itr;
  index_t new_node;
  void *data;
  /* lock list */
  alist_get_new_element(i->list, &data, &new_node);
  itr->jump(itr, &new_node);
  alist_put_element(i->list, &new_node, 1);
  i->modified = 1;
  /* unlock list */
  return data;
}

void * alist_iterator_next(iterator *itr)
{
  index_t inext;
  void *data;
  alist_iterator *i = (alist_iterator *) itr;
  i->pos++;
  if (i->node->maxpos < i->pos) {
    if (i->pos < i->list->elements_per_node) {
      i->node->maxpos++;
      return ALIST_DATA_ELEM(i->node, i->list->head, i->pos);
    }
    if (ALIST_IDX_NIL(i->node->inext)) {
      cache1_put_element(i->list->cache, &i->inode, i->modified);
      i->modified = 0;
      /*ini: posible problema de concurrencia*/
      alist_get_new_element(i->list, &data, &i->inode);
      i->node = (alist_node *)((char *)data - (ALIST_GET_POS(&i->inode)*
					       (i->list->head->element_size) +
					       sizeof(alist_head)));
      /*i->node = i->list->last; */
      /*fin: posible problema de concurrencia*/
    } else {
      inext = i->node->inext;
      cache1_put_element(i->list->cache, &i->inode, i->modified);
      i->modified = 0;
      i->inode = inext;
      cache1_get_by_index(i->list->cache, (void **)&i->node, &i->inode);
    }
    i->pos=0;
    return ALIST_DATA_POINTER(i->node);
  }
  return (void *)ALIST_DATA_ELEM(i->node, i->list->head, i->pos);
}

int alist_iterator_has_next(iterator *itr)
{
  alist_iterator *i = (alist_iterator *) itr;
  /*fprintf(stderr,"has_next max pos: %d\n",i->list->head->maxpos);*/
  if (i->node->maxpos < 0) return 0;
  if (i->pos == i->node->maxpos) {
    if (ALIST_IDX_NIL(i->node->inext)) return 0;
  }
  return 1;
  /*return !(i->pos+1 > i->node->maxpos && ALIST_IDX_NIL(i->node->inext));*/
}

void * alist_iterator_prev(iterator *itr)
{
  alist_iterator *i = (alist_iterator *) itr;
  if (i->pos > 0) {
    i->pos--;
    return (void *)ALIST_DATA_ELEM(i->node,i->list->head,i->pos);
  }
  cache1_put_element(i->list->cache, &i->inode, i->modified);
  i->modified = 0;
  i->inode = i->node->iprev;
  cache1_get_by_index(i->list->cache, (void **)&i->node, &i->inode);
  i->pos = i->node->maxpos;
  return (void *)ALIST_DATA_ELEM(i->node, i->list->head, i->pos);
}

int alist_iterator_has_prev(iterator *itr)
{
  alist_iterator *i = (alist_iterator *) itr;
  if (i->pos > 0) return 1;
  if (!INDEX_EQ(i->inode, i->list->ihead)) return 1;
  return 0;
}

void * alist_iterator_current(iterator *itr)
{
  alist_iterator *i = (alist_iterator *) itr;
  return (void *)ALIST_DATA_ELEM(i->node, i->list->head, i->pos);
}

int alist_iterator_delete(iterator *itr)
{
  fprintf(stderr,"No se permite borrar de alist con iterador (aun...)");
  return 1;
}

void* alist_iterator_jump(iterator *itr, const index_t *index)
{
  void *data;
  index_t *idx = (index_t *) index;
  index_t *inode;
  alist_iterator *i = (alist_iterator *) itr;
  inode = (index_t *)&i->inode;
  
  /*debug_index(idx, "index");
    debug_index(&i->inode, "inode");
    debug_index(&i->list->ihead, "ihead");
    debug_index(&i->list->ilast, "ilast"); */

  if (idx->page == inode->page) {
    if (idx->elem != inode->elem) {
      i->node = (alist_node *)((char *)i->node + (idx->elem - inode->elem) *
			       ((cache1_head *)i->list->cache)->chunk_size);
      inode->elem = idx->elem;
    }
    i->pos=ALIST_GET_POS(index);
    return ALIST_DATA_POINTER(i->node) + i->pos*i->list->head->element_size;
  }else{
    cache1_put_element(i->list->cache, inode, i->modified);
    i->modified = 0;
    ALIST_GET_BY_INDEX(i->list, &data, index, i->node);
    inode->page = idx->page;
    inode->elem = idx->elem;
    return data;    
  }
}

void alist_iterator_set_modified(iterator *itr)
{
  alist_iterator *i = (alist_iterator *) itr;
  i->modified = 1;
}

void* alist_iterator_skip_by_bytes(iterator *itr, unsigned int num_bytes)
{
  alist_iterator *i = (alist_iterator *) itr;
  return alist_iterator_skip(itr, ALIST_NUMBER_OF_ELEMENTS(i->list,num_bytes));
}

void* alist_iterator_skip(iterator *itr, int n)
{
  int m;
  alist_iterator *i = (alist_iterator *) itr;
  index_t index;
  while((m = n - (i->list->elements_per_node - i->pos ))>0){
    n =  n - m;
    index = i->node->inext;
    alist_iterator_jump(itr, &index);
  }
  i->pos = n;
  return ALIST_DATA_POINTER(i->node) + i->pos*i->list->head->element_size;
}

int alist_iterator_get_current_index(iterator *itr, index_t *idx)
{
  alist_iterator *i = (alist_iterator *) itr;
  *idx = i->inode;
  idx->pos = i->pos;
  return 0;
}

int alist_iterator_find_forward(iterator *itr, void *busq,
			int (*cmp)(void *_en_lista, void *_busq))
{
  void *data;
  int cmpres;
  alist_iterator *aitr = (alist_iterator *)itr;
  if (aitr->pos >= 0) {
    data = itr->current(itr);
    cmpres = cmp(data, busq);
    if (cmpres == 0) {
      return 1;
    } else if (cmpres > 0) {
      return 0;
    }
  }
  while (itr->has_next(itr)) {
    data = itr->next(itr);
    cmpres = cmp(data, busq);
    if (cmpres == 0) {
      return 1;
    } else if (cmpres > 0) {
      return 0;
    }
  }
  return 0;
}

int alist_iterator_find_backward(iterator *itr, void *busq,
			int (*cmp)(void *_en_lista, void *_busq))
{
  void *data;
  int cmpres;
  alist_iterator *aitr = (alist_iterator *)itr;
  if (aitr->pos == -1) {
    if (aitr->list->head->maxpos == -1) {
      return 0;
    }
    itr->prev(itr);
  }
  
  data = itr->current(itr);
  cmpres = cmp(data, busq);
  if (cmpres == 0) {
    return 1;
  } else if (cmpres < 0) {
    return 0;
  }

  while (itr->has_prev(itr)) {
    data = itr->prev(itr);
    cmpres = cmp(data, busq);
    if (cmpres == 0) {
      return 1;
    } else if (cmpres < 0) {
      return 0;
    }
  }
  return 0;
}

int alist_iterator_find(iterator *itr, void *busq,
			int (*cmp)(void *_en_lista, void *_busq))
{
  return itr->find_backward(itr, busq, cmp);
}

int alist_find(alist *list, void *buscado,
	       int(* cmp)(void *en_lista, void *buscado), index_t *index)
{
  iterator *itr;
  int res = 0;
  alist_iterator_new(list, &itr, 0);
  itr->get_current_index(itr, index);
  res = itr->find(itr, buscado, cmp);
  itr->get_current_index(itr, index);
  itr->close(itr);
  return res;
}

int alist_iterator_recover_element(iterator *itr, void *data, int size)
{
  void *pointer;
  int i, num, esize, ini = 0;
  int x = 0, y = 0;
  alist_iterator *ai = (alist_iterator *) itr;
  num = alist_number_of_elements(ai->list, size) - 1;
  esize = ai->list->head->element_size;
  for (i = 0; i < num; i++) {
    pointer = itr->next(itr);
    if (x==0) x = (int)pointer;
    memcpy((void *)((char *)data+ini), pointer, esize);
    ini+=esize;
  }
  pointer = itr->next(itr);
  memcpy((void *)((char *)data+ini), pointer, size-ini);
  itr->next(itr); /*move to the next (consume this element)*/
  return num+1;
}
