#ifndef _H_ALIST_
#define _H_ALIST_
#include <cache1.h>
#include <iterator.h>

/* Esta lista est diseada para crecer, para borrar no.
   Si es necesario que sea ms dinmica (evitar fragmentacin por ejemplo)
   se debe disear e implementar mediante indices y nodo hacia nodo anterior.
 */

/* alist_new falla si no puede alojar elementos (por tamao) */
#define ALIST_NO_ELEMENTS -899

#define ALIST_IDX_NIL(_index_t_) INDEX_ZERO(_index_t_)

/* sizeof(struct_alist_head) = sizeof(struct_alist_node)
   en la misma posicin inext y maxpos */

struct struct_alist_head {
  int16_t maxpos; /* Posicin Mxima en ste nodo */
  index_t inext; /* Siguiente nodo de alist (el primero es la cabecera)*/
  index_t ilast; /* ltimo nodo de alist de sta lista */
  //u_int32_t total; /* Nmero total de nodos */
  unsigned int total: 28; /* Nmero total de nodos */
  unsigned int element_size: 12; /* Tamao de cada elemento */
  /*void *data;*/ /* Para los datos se debe saltar sta estructura */
}  __attribute__ ((__packed__));


struct struct_alist_node {
  int16_t maxpos; /* Posicin Mxima en ste nodo */
  index_t inext; /* Siguiente nodo de alist (el primero es la cabecera)*/
  index_t iprev; /* ltimo nodo de alist de sta lista */
  index_t ihead; /* Cabecera */
  //u_int16_t magic; /* Magic Node */
  /*void *data;*/ /* Para los datos se debe saltar sta estructura */
}  __attribute__ ((__packed__));


typedef struct struct_alist_node alist_node;
typedef struct struct_alist_head alist_head;

#define ALIST_DATA_POINTER(_node) \
  ((char *)(((char *) _node) + sizeof(struct struct_alist_node)))
/* obtiene la direccin de el _idx-esimo elemento en la lista*/
#define ALIST_DATA_ELEM(_node,_head,_idx) \
  ((char *)(ALIST_DATA_POINTER(_node) + (_idx) * (_head)->element_size))

struct struct_alist {
  cache1 *cache; /* La cache donde se encuentra la lista*/
  u_int16_t elements_per_node; /* El nmero de elementos por nodo */
  index_t ihead;  /* Indice de la cabecera de la lista */
  index_t ilast; /* ltimo indice en la lista */
  alist_head *head; /* Primer nodo de la lista */
  alist_node *last; /* ltimo nodo de la lista */
  pthread_mutex_t mutex; /* Mutex */
  u_int8_t modified; /* Se modifica el last node */
};

/* Estructura para iterar nodos de lista */
typedef struct struct_alist alist;

struct struct_alist_iterator {
  iterator itr; /* Estructura a iterador sirve como interfaz. Por el cast debe
		   estar en el principio de estructura */
  alist *list; /* Lista iterando */
  alist_node* node; /* Nodo que iteramos actualmente */
  index_t inode; /* Indice del nodo actual */
  int16_t pos; /* Posicin actual en el nodo */
  u_int8_t modified; /* Se modifica el actual node */
  u_int8_t close_list;
};

/* Pos es a lo ms de 12 bits, as podemos utilizar los 4 bits restantes
   como banderas o estados
 */


#define ALIST_GET_POS(_index) ((_index)->pos)
#define ALIST_SET_POS(_index, _pos) ((_index)->pos = _pos)
/*
#define ALIST_GET_POS(_index) ((_index)->extra & 0x0fff)
#define ALIST_SET_POS(_index, _pos) \
 (_index)->extra &= 0xf000; \
 (_index)->extra |= _pos;
*/

/* Crea una lista, el nmero de nodos por pgina se ajusta al tamao de pgina
   de la cache.
   Errores:
   ALIST_NO_ELEMENTS No se pueden alojar elementos en la cache
   Si elem_size = 0, entonces se crea un solo elemento por nodo.
*/
int alist_new(alist **n, cache1 *c, size_t elem_size);

/* Abre una lista, pone el indice creado en |*head| */
int alist_open(alist **l, const index_t* head, cache1 *c);

/* Cierra una lista */
int alist_close (alist *l);

/* Quita la referencia a un elemento */
int alist_put_element(alist *list, const index_t *idx, int modified);

/* Pone como modificada la informacin del iterador */
void alist_iterator_set_modified(iterator *itr);

/* Obtiene un elemento de la lista, se accede como arreglo
   (el tiempo de acceso es constante). El puntero *data apuntara
   a la posicin en memoria del dato pedido. La referencia al
   nodo utilizado queda abierta por lo que es necesario regresarla con
   cache1_put_element o alist_put_element */


/* busca el elemento "buscado" en la lista si lo encuentra es 1 y si no es 0
   pone en index el indice donde encontro al elemento */

int alist_find(alist *list,void *buscado, int(* cmp)(void *en_lista,void *buscado),index_t *index);


int alist_get_by_index(alist *list, void **data, const index_t* idx);

/* Copia el contenido del elemento |data| (de |size| bytes) en un nuevo
   nodo, pone el indice en |idx|, no se necesita borra la referencia al
   elemento */
int alist_insert_new(alist *list, void *data, index_t *idx, signed int size);

/* Asigna memora y la pone en *data, para liberarla (o dejarla de utilizar)
   es necesario llamar a cache1_put_element o alist_put_element con el valor
   puesto en idx que es el indice del nodo al que pertenece *data.
   Deja abierta la referencia, se debe cerrar al dejar de usarla.
 */
int alist_get_new_element(alist *list, void **data, index_t *idx);

/* Calcula el nmero de elementos necesarios para bytes 'bytes' */
int alist_number_of_elements(alist *list, int bytes);


/* Iteradores */
typedef struct struct_alist_iterator alist_iterator;

/* Crea un nuevo iterador sobre list y lo pone en *itr. El nuevo iterador
   ser del tipo alist_iterator y es forzado al tipo iterator. Si close
   es verdadero entonces se cerrara la lista iterada cuando se cierre el
   iterador.
 */

int alist_iterator_new(alist *list, iterator **itr, int close);
/* Ver documentacin de iterator.h */
int alist_iterator_recover_element(iterator *itr, void *data, int size);
void* alist_iterator_next(iterator *itr);
void* alist_iterator_insert(iterator *itr);
void* alist_iterator_prev(iterator *itr);
int alist_iterator_has_next(iterator *itr);
int alist_iterator_has_prev(iterator *itr);
void* alist_iterator_current(iterator *itr);
int alist_iterator_delete(iterator *itr);
int alist_iterator_reset (iterator *itr);
int alist_iterator_close (iterator *itr);
void* alist_iterator_jump(iterator *itr, const index_t *idx);
void* alist_iterator_skip(iterator *itr, int n);
void* alist_iterator_skip_by_bytes(iterator *, unsigned int num_bytes);
int alist_iterator_get_current_index(iterator *itr, index_t *idx);
int alist_iterator_find(iterator *, void *busq,
			int (*cmp)(void *en_lista, void *busq));
int alist_iterator_find_forward(iterator *, void *busq,
			int (*cmp)(void *en_lista, void *busq));
int alist_iterator_find_backward(iterator *, void *busq,
			int (*cmp)(void *en_lista, void *busq));

#define ALIST_ITERATOR_SET_MODIFIED(itr) (((alist_iterator *)(itr))->modified = 1)
#endif
