#include <cache1.h>
#include <alist.h>
#include <stdio.h>
#include <btree.h>
#include <super_malloc.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#define FLOOR(x) (x)
#define CEIL(x) ((int)ceil(x))

static index_t __btree_index_null = {0,0,0,0};
#define __btree_index_is_null(_idx) (memcmp((_idx), &__btree_index_null, sizeof(index_t)) == 0)

int btree_open(btree **_b3, index_t *head, alist *headlist, alist *nodelist,
	       int (*cmp_function)(btree *, index_t *, index_t *),
	       int (*dup_function)(btree *, int *, int *))
{
  btree *b3 = struct_malloc(btree);
  b3->headlist = headlist;
  b3->nodelist = nodelist;
  b3->ihead = *head;
  b3->cmp = cmp_function;
  b3->dup = dup_function;
  alist_get_by_index(headlist, (void **)&b3->head, head);
  alist_get_by_index(nodelist, (void **)&b3->root, &b3->head->iroot);
  *_b3 = b3;
  return 0;
}

#warning "adivinar modificaciones"
int btree_close(btree *b3)
{
  alist_put_element(b3->nodelist, &b3->head->iroot, 1);
  alist_put_element(b3->headlist, &b3->ihead, 1);
  free(b3);
  return 0;
}

void btree_node_init(btree_node *node, u_int16_t elements, u_int8_t external)
{
  node->elements = elements;
  node->external = external;
  memset(&node->iprev, 0, sizeof(node->iprev));
  memset(&node->inext, 0, sizeof(node->inext));
}

int btree_create(btree **_b3, index_t *ihead, alist *headlist, alist *nodelist,
		 int (*cmp_function)(btree *, index_t *, index_t *),
		 int (*dup_function)(btree *, btree_pair *, btree_pair *))
{
  int n;
  btree *b3 = struct_malloc(btree);
  b3->headlist = headlist;
  b3->nodelist = nodelist;
  pthread_mutex_init(&b3->mutex, NULL);
  alist_get_new_element(headlist, (void **)&b3->head, &b3->ihead);
  //memset(&b3->head->iroot, 0, sizeof(btree_head));
  n = nodelist->head->element_size - sizeof(u_int16_t) - sizeof(index_t)*2;
  n /= sizeof(index_t);
  b3->head->maximum_pairs = FLOOR(n / 2) + (n%2);
  b3->head->minimum_pairs = FLOOR(b3->head->maximum_pairs / 2);
  b3->head->height = 1;
  alist_get_new_element(headlist, (void **)&b3->root, &b3->head->iroot);
  btree_node_init(b3->root, 0, 1);
  *_b3 = b3;
  *ihead = b3->ihead;
  b3->cmp = cmp_function;
  b3->dup = dup_function;
  return 0;
}

/* Encuentra la llave correspondiente a value */
int btree_binsearch(btree *b3, btree_node *node,
		    index_t *key, int *pos, int *cmp)
{
  int half, size, abs;
  btree_pair *pairs = &node->pairs[0];
  size = node->elements;
  abs = 0;

  while (1) {
    if (size < 1) {
      /* return 0: no se encontr, pero debera estar en la posicin |pos| */
      *pos = abs;
      return 0;
    }
    half = FLOOR(size / 2);
    /* cmp representa: value - keys[half] */
    *cmp = b3->cmp(b3, key, &pairs[half].key);
    if (*cmp == 0) {
      /* return 1 para |value| encontrado, se pone en |pos| la posicin */
      *pos = abs + half;
      return 1;
    }
    if (*cmp < 0) {
      size = half;
    } else if (*cmp > 0) {
      half++;
      pairs += half;
      size -= half;
      abs += half;
    }
  }
}

int btree_internal_search(btree *b3, btree_node *root,
			  index_t *key, btree_pair *pair,
			  btree_node **node, int *pos, int *cmp)
{
  esta mal inserta donde encuentra, no hasta el fondo
  int k = 0;
  index_t iprev;
  *node = root;
  
  iprev = b3->head->iroot;
  while (1) {
    k++;
    if (btree_binsearch(b3, *node, key, pos, cmp)) {
      /* Se encontro*/
      memcpy(pair, &(*node)->pairs[*pos], sizeof(btree_pair));
      /*se deja abierto y se cierra en la aplicacin */
      return 1;
      /*if (k > 1) alist_put_element(b3->list, &prev, 0);*/
    } else {
      /* No se encontr */
      memcpy(pair, &(*node)->pairs[*pos], sizeof(btree_pair));
      if ((*node)->external) return 0;
      /* Nodos internos */
      alist_get_by_index(b3->nodelist, (void **)node, &pair->data);
      if (k > 1) alist_put_element(b3->nodelist, &iprev, 0);
      iprev=pair->data;
    }
  }
}

int btree_split(btree *b3, int pos, index_t *item,
		btree_node *onode, btree_node *nnode)
{
  int halfceil = CEIL(onode->elements / 2),
    halffloor = FLOOR(onode->elements / 2);
  btree_node *node;
  /* total = halfceil + halffloor */
  memmove(&nnode->pairs, &onode->pairs, halffloor);
  node = (pos < halfceil) ? onode : nnode;
  memmove(&(node->pairs)+pos+1, &(node->pairs)+pos,
	  (onode->elements-pos)*sizeof(index_t));
  onode->elements = halfceil;
  nnode->elements = halffloor;
  node->pairs[pos].key = *item;
  return 0;
}

int btree_insert_normal() {
}
int btree_insert_intern(btree *b3, btree_node *root, btree_pair *insert)
{
  int pos,cmp;
  btree_node *oldnode, *newnode;
  btree_pair pair;
  index_t inew;

  if (btree_internal_search(b3, root, &insert->key, &pair,
			    &oldnode, &pos, &cmp)) {
    if (b3->dup) {
      b3->dup(b3, insert, &pair);
    } else {
      return 0;
    }
  } else {
    /* falta diferenciar cuando es externo o interno */
    if (oldnode->elements == b3->head->maximum_pairs) {
      /* cuando se parte el nodo */
      alist_get_new_element(b3->nodelist, (void **)&newnode, &inew);
      btree_node_init(newnode, 0, 1);
      btree_split(b3, pos, &insert->key, oldnode, newnode);
      if (oldnode->external) {
	if (__btree_index_is_null(&oldnode->external_iparent)) {
	  index_t iroot;
	  btree_node *newroot;
	  /* Es root */
	  alist_get_new_element(b3->nodelist, (void **)&newroot, &iroot);
	  btree_node_init(newroot, 2, 0);
	  newroot->pairs[0].key = oldnode->pairs[0].key;
	  newroot->pairs[0].data = b3->head->iroot;
	  newroot->pairs[1].key = newnode->pairs[0].key;
	  newroot->pairs[1].data = inew;
	  b3->head->iroot = iroot;
	  b3->root = newroot;
	  newnode->external_iparent = iroot;
	  oldnode->external_iparent = iroot;
	  alist_put_element(b3->nodelist, &inew, 1);
	  alist_put_element(b3->nodelist, &newroot->pairs[0].data, 1);
	} else {
	  
	}
      } else {
      }
      fprintf(stderr, "SPLIT\n");
      alist_put_element(b3->nodelist, &inew, 1);
    } else {
      /* cuando hay espacio para insertar */
      fprintf(stderr, "pos: %d, insertkey: %d\n", pos, insert->key);
      memmove(oldnode->pairs+pos+1, oldnode->pairs+pos,
	      (oldnode->elements-pos)*sizeof(oldnode->pairs[0]));
      oldnode->pairs[pos] = *insert;
      oldnode->elements++;
    }
    return 1;
  }
  return 0;
}

int btree_insert(btree *b3, btree_pair *insert)
{
  return btree_insert_intern(b3, b3->root, insert);
}

int btree_search(btree *b3, index_t *key, btree_pair *result)
{
  int pos, cmp;
  btree_node *node;
  btree_internal_search(b3, b3->root, key, result, &node, &pos, &cmp);
  if (node != b3->root) {
    alist_put_element(b3->nodelist, &result->data, 0);
  }
  return cmp == 0;
}

int btree_update(btree *b3, index_t *oldkey, index_t *newkey)
{  
}

int btree_delete(btree *b3, index_t *key)
{
}

int btree_cmp(btree *b3, index_t *a, index_t *b)
{
  int cmp;
  CACHE1_CMP_INDEX(cmp, a, b);
  return cmp;
}

#include <math.h>

int main()
{
  int i, j, k, ok;
  cache1 *c;
  alist *nodes;
  btree_node *node;
  btree *b3;
  index_t ihead, key, iprev;
  btree_pair pair;
  unlink("example");
  cache1_open(&c, "example", 0600, 1, 1, 0);
  alist_new(&nodes, c, 0);
  btree_create(&b3, &ihead, nodes, nodes, btree_cmp, NULL);
  memset(&key, 0, sizeof(key));
  for (ok = i = 0; i < 1; i++) {
    for (j = 0; j < 1400; j++) {
      //k = random()%800;
      k = j;
      fprintf(stderr,"random: %d\n",k);
      key.page = k;
      fprintf(stderr, "Search  xxantes: %d\n", btree_search(b3, &key, &pair));
      pair.key.page = k;
      ok+=btree_insert(b3, &pair);
      fprintf(stderr, "Search despues: %d\n", btree_search(b3, &key, &pair));
    }
  }
  fprintf(stderr, "Ordenados: %d\n", ok);
  node = b3->root;
  iprev = b3->head->iroot;
  for (k = i = 0; i < ok/b3->head->maximum_pairs; i++) {
    k += 1;
    for (j = 0; j < b3->head->maximum_pairs && k < ok; j++) {
      fprintf(stderr, "%04d pos %04d: %d\n", k, j, node->pairs[j].key);
    }
    iprev = node->inext;
    alist_get_by_index(b3->nodelist, (void **)&node, &iprev);
    alist_put_element(b3->nodelist, &iprev, 0);
  }
  fprintf(stderr, "DESPUES BBB\n");
  btree_close(b3);
  fprintf(stderr, "DESPUES CCC\n");
  alist_close(nodes);
  fprintf(stderr, "DESPUES DDD\n");
  cache1_close(c);
  fprintf(stderr, "DESPUES EE\n");
  return 0;
}
