#include <cache1.h>
#include <super_malloc.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/syscall.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include <pthread.h>
#include <debug.h>
#include <config_main.h>
#include <string.h>

int _seek, __find = 0, __read = 0, __write = 0, __reuse = 0, __reuseTRUE = 0,
  __reuseFALSE = 0, __get_new = 0, __get_index = 0;
 
/* Pone el  hash de |cache| con |ssize| pginas estticas (sin colisiones
   en las primeras ssize pginas) y con |dsize| pginas dinmicas. */
void cache1_hash_create(cache1 *cache, u_int32_t ssize, u_int32_t dsize)
{
  unsigned int i;
  cache1_hash *h = struct_malloc(cache1_hash);
  h->size = 0;
  h->ssize = ssize;
  /* El doble para decrementar el factor de colisiones */
  h->dsize = dsize * 2;
  if (ssize > 0) {
    h->spages = (cache1_page **)calloc(ssize, sizeof(cache1_page *));
    if (h->spages == NULL) {
      perror("cache1_hash_create: spages = NULL");
      exit(-1);
    }
  }
  /* Siempre se itera sobre el arreglo en sync_all ya que es una operacin
     que se realiza mucho menos frecuente que acceder, no vale la pena
     tantos saltos por un poco de velocidad en sync_all */
  if (h->dsize > 0) {
    h->dpages = (list_head *)calloc(h->dsize, sizeof(list_head));
    if (h->dpages == NULL) {
      fprintf(stderr,"cache1_hash_create: dpages = NULL\n");
      exit(-1);
    }
  }
  for (i = 0; i < h->dsize; i++) list_head_init(&h->dpages[i], 1, NULL);
  cache->pool = h;
}

struct cache1_hash_result {
  unsigned int index;
  list_node *result;
  cache1_page *page;
};

struct cache1_walk_info {
  cache1 *cache;
  int is_static;
};

int cache1_hash_itr_get_node(void *res,  list_node *node)
{
  struct cache1_hash_result *r = (struct cache1_hash_result *) res;
  list_node *node_page = (list_node *)node->ld_vpointer;
  if (r->index == node_page->ld_uint) {
    r->result = node;
    r->page = (cache1_page *)node_page;
    return 0;
  }
  return 1;
}

/* Encuentra la pgina asociada co |*idx| y la pone en |*p|, sin proteccin
   de concurrencia. Si son de las primeras hay problemas si no estn
   inicializadas, cada pgina debe tener page = NULL para saber si est o
   no inicializada */
#define CACHE1_HASH_GET_I(_pool, _page) \
  ((_pool)->dpages[(_page) % (_pool)->dsize])

int cache1_hash_find(cache1 *c, cache1_page **p, const index_t *idx)
{
  list_head *n;
  struct cache1_hash_result r;
  if (c->pool->ssize > idx->page) {
    *p = c->pool->spages[idx->page];
  } else {
    n = &CACHE1_HASH_GET_I(c->pool, idx->page);
    if (n->size > 0) {
      r.result = NULL;
      r.page = NULL;
      r.index = idx->page;
      list_head_iterator_next(n, cache1_hash_itr_get_node, &r);
      if (p == NULL) fprintf(stderr,"page addr: %p\n", p);
      *p = r.page;
    } else {
      *p = NULL;
    }
  }
  return *p != NULL;
}

int cache1_hash_insert(cache1 *c, cache1_page *p)
{
  list_head *h;
  /* promovido funcin arriba: pthread_mutex_lock(&c->mutex_stree);*/
  c->pool->size++;
  if (p->node.ld_uint < c->pool->ssize) {
#ifdef _DEVELOP_VERSION
    if (c->pool->spages[p->node.ld_uint] != NULL) {
      fprintf(stderr,"cache1_hash_insert: reemplazando pgina %d\n",
	      p->node.ld_uint);
      memset(NULL, 0, 32);
      exit(-3);
    }
#endif
    c->pool->spages[p->node.ld_uint] = p;
  } else {
    h = &CACHE1_HASH_GET_I(c->pool, p->node.ld_uint);
    list_head_unshift_vpointer(h, p, NULL);
  }
  /* pthread_mutex_unlock(&c->mutex_stree);*/
  return 1;
}

int cache1_hash_remove(cache1 *c, cache1_page *p)
{
  list_head *h;
  struct cache1_hash_result r;
  r.index = p->node.ld_uint;
  /* promovido arriba: pthread_mutex_lock(&c->mutex_stree);*/
  if (c->pool->ssize > r.index) {
    c->pool->spages[r.index] = NULL;
    c->pool->size--;
    /* pthread_mutex_unlock(&c->mutex_stree); */
    return 0;
  }
  r.result = NULL;
  h = &CACHE1_HASH_GET_I(c->pool, r.index);
  list_head_iterator_next(h, cache1_hash_itr_get_node, &r);
  if (r.result == NULL) {
    fprintf(stderr,"cache1_hash_remove: no existe pgina %d\n",r.index);
#ifdef _DEVELOP_VERSION
    exit(-1);
#endif
    /* pthread_mutex_unlock(&c->mutex_stree);*/
    return -1;
  }
  c->pool->size--;
  list_head_remove_node(h, r.result);
  /* pthread_mutex_unlock(&c->mutex_stree);*/
  return 0;
}

void cache1_hash_walk(cache1 *c, int (*action)(void *, list_node *))
{
  register int i;
  cache1_page *p;
  list_head *h;
  struct cache1_walk_info winfo;
  winfo.cache = c;
  /*fprintf (stderr, "COMENZANDO CACHE SYNC\n");*/
  for (i = c->pool->ssize - 1; i >= 0; i--) {
    p = c->pool->spages[i];
    if (p == NULL) continue;
    winfo.is_static = 1;
    action(&winfo, (list_node *)p);
  }
  int xx = 0;
  /*fprintf(stderr, "\nINI POOL_SIZE: %d, COUNTED: %d\n", c->pool->size, xx);*/
  for (i = c->pool->dsize - 1; i >= 0; i--) {
    h = &c->pool->dpages[i];
    if (h->size > 0) {
      winfo.is_static = 0;
      xx+=h->size;
      /*fprintf(stderr, "HISTOGRAMA POS: %d, SIZE: %d, max: %d\n",
	i, h->size, c->max_pages);*/
      list_head_iterator_next(h, action, &winfo);
    }
  }
  /*fprintf(stderr, "\nFIN POOL_SIZE: %d, COUNTED: %d\n", c->pool->size, xx);*/
}

void cache1_hash_destroy(cache1 *c)
{
  register int i,l;
  cache1_hash *hash;
  /* se debo sincronizar las pginas con anterioridad */
  cache1_page *p;
  hash = c->pool;
  l = c->pool->dsize;
  for (i = 0; i < l; i++) {
    if (hash->dpages[i].size > 0) {
      list_head_free(&hash->dpages[i], 0);
    }
  }
  if (hash->ssize > 0) {
    for (i = hash->ssize-1; i >= 0; i--) {
      p = hash->spages[i];
      if (p == NULL) continue;
      cache1_page_free(p, c);
    }
    free(hash->spages);
  }
  if (hash->dsize > 0) free(hash->dpages);
  free(hash);
}

/* Debe estar bloqueado el padre */
int cache1_page_free(cache1_page *p, cache1 *c)
{
#ifdef _DEVELOP_VERSION
  if (p->refs != 0) {
    fprintf(stderr,"Se libera pgina %p(%d), con datos %p y %d referencias\n",
	    (void *)p, p->node.ld_uint, (void *)p->page, p->refs);
    memset(NULL, 0, 32);
    exit(-1);
  }
#endif
#ifdef CACHE1_USE_MMAP
  munmap(p->page, ((cache1_head *)c)->page_size);
#else
  free(p->page);
#endif
  free(p);
  return 0;
}

int cache1_page_sync(cache1_page *p, cache1 *c)
{
  int ret;
  off_t seek;
  if (c->write_mode == 0 || p->modified == 0) {
    /*fprintf(stderr, "SALTO\n");*/
    return 0;
  }
  /*fprintf(stderr, "GUARDO\n");*/
#ifdef CACHE1_USE_MMAP
  ret = msync(p->page, ((cache1_head *)c)->page_size, MS_SYNC);
#else
  seek = CACHE1_ABSPOS(c, p->node.ld_uint);
  _seek++;
  ret = lseek(c->fd, seek,SEEK_SET);
  if (ret != seek) {
# ifdef _DEVELOP_VERSION
    perror("cache1_page_sync:");
    fprintf(stderr,"cache1_page_sync: %d\n",p->node.ld_uint);
    fprintf(stderr,"cache1_page_sync: shift=%d\n",c->shift);
    fprintf(stderr,"cache1_page_sync: ret=%d seek=%d\n",ret,(int)seek);
    fprintf(stderr,"error!!\n");
    exit(-2);
# endif 
  }
  /*__write++;*/
  write(c->fd, p->page, ((cache1_head *)c)->page_size);
  ret = 0;
#endif
  if (ret != 0) perror("cache1_page_sync");
  return ret;
}

int cache1_page_create(cache1_page **p, cache1 *c, const index_t *index)
{
  int ret;
  cache1_page *page;
  cache1_head *head = (cache1_head *)c;
  page = struct_malloc(cache1_page);
#ifdef CACHE1_USE_MMAP  
  page->page = NULL;
#else
  page->page = super_malloc(head->page_size);
#endif
  ret = cache1_page_init(page, c, index);
  if (ret == 0) {
    *p = page;
  } else {
    fprintf(stderr, "cache1_page_init fails!! %d\n", ret);
    exit(-1);
  }
  return ret;
}

int cache1_page_init(cache1_page *p, cache1 *c, const index_t *index)
{
  cache1_head *h = (cache1_head *)c;
#ifdef CACHE1_USE_MMAP
  void *data = mmap(p->page, h->page_size, PROT_WRITE | PROT_READ ,
		    MAP_SHARED, c->fd, CACHE1_ABSPOS(c, index->page));
  if (data == NULL) {
# ifdef _DEVELOP_VERSION
    fprintf(stderr,"cache1_page_init fd: %d, errno: %d\n",c->fd, errno);
    perror("cache1_page_init:NULL");
    exit(-1);
# endif
    return -1;
  }
  if (data == MAP_FAILED) {
# ifdef _DEVELOP_VERSION
    fprintf(stderr,"cache1_page_init fd: %d, errno: %d\n",c->fd, errno);
    perror("cache1_page_init:MAP_FAILED");
    exit(-1);
# endif
    return -1;
  }
  p->page = data;
#else /* No MMAP */
  off_t seek = CACHE1_ABSPOS(c, index->page), res;
  _seek++;
  res=lseek(c->fd, seek, SEEK_SET);
  if (seek != res) {
# ifdef _DEVELOP_VERSION
    if (index->page > h->last_position.page) {
      //matar otros threads y espera por gdb
      debug_index(index, "cache1_page_xXinit index:");
      fprintf(stderr,"cache1_page_initxX: bad seek: %ld Res: %ld\n", seek, res);
    }
    fprintf(stderr,"cache1_page_init: bad seek. Seek: %d Res: %d\n", seek, res);
    exit(-2);
# endif
  }
  /*__read++;*/
  res=read(c->fd, p->page, h->page_size);
#endif
  p->node.ld_uint = index->page;
  p->refs = 0;
  p->modified = 0;
  return 0;
}

int cache1_page_reuse(cache1_page *p, cache1 *c, const index_t *index)
{
  cache1_page_sync(p,c);
  return cache1_page_init(p,c,index);
}

/* Adiciona una referencia a |p|
   Debe estar protegida con mutex cache1->mutex_zero_refs desde afuera
   de sta funcin, pues puede estar ligado a otra accin.
 */
inline int cache1_page_add_ref(cache1_page *p, cache1 *c)
{
  p->refs++;
  if (p->refs > 1) return 0;
  list_head_remove_node(&c->zero_list, &p->node);
  return 0;
}

/* Remueve una referencia a |p|.
   Debe estar protegida con mutex cache1->mutex_zero_refs desde afuera
   de sta funcin, pues puede estar ligado a otra accin. */
#warning "An NO se recupera memoria de pginas, slo se reusa"
inline int cache1_page_del_ref(cache1_page *p, cache1 *c)
{
  if (p->refs == 0) {
    fprintf(stderr,"cache1_page_del_ref: se libera pagina de cache no "
	    "alojada refs=%d, idx=%d\n", p->refs, p->node.ld_uint);
    debug_backtrace();
    exit(-3);
  }
  p->refs--;
  if (p->refs == 0)list_head_unshift_integer(&c->zero_list, p->node.ld_uint, &p->node);
  return 0;
}

/* Una vez que se han colocado o ledo los valores de la cache en la
   cabecera se inicializan los valores calculados y exclusivos de memoria */
int cache1_init(cache1 *c)
{
  cache1_head *h = (cache1_head *)c;
  c->shift = (u_int8_t)ceil(log(h->page_size)/log(2));
  pthread_mutex_init(&c->mutex_idx, NULL);
  pthread_mutex_init(&c->mutex_new, NULL);
  return 0;
}

/* Habre una nueva cache, si no existe la crea.
   Errores:
   CACHE1_FAIL_OPEN: No se pudo abrir cache
   Informacin:
   CACHE1_CREATED: Se creo la cache en lugar de abrirla
   CACHE1_OPENED: Se abrio la cache */
int cache1_open (cache1** cache, const char *file, int mode,
		 int spages, int dpages, int divs)
{
  int ret, o_flags;
  cache1* c;
  cache1_head *h;
  cache1_page *p;
  index_t index;
  if (divs >= CACHE1_MAX_ELEMENTS_PER_PAGE) {
    fprintf(stderr, "No puede haber ms de %d divisiones, se piden %d\n",
	    CACHE1_MAX_ELEMENTS_PER_PAGE - 1, divs);
    exit(-1);
  }
  if (spages < 1) spages = 1;
  if (dpages < 1) dpages = 1;
  c = struct_malloc(cache1);
  h = (cache1_head *)c;
  if ((0600 & mode) == 0600) {
    o_flags = O_RDWR | O_NONBLOCK | O_ASYNC;
    c->write_mode = 1;
  } else {
    o_flags = O_RDONLY | O_NONBLOCK | O_ASYNC;
    c->write_mode = 0;
  }
  if ( (c->fd = open(file, o_flags, mode)) == -1 ) {
    if (c->write_mode==0 || (c->fd=open(file, o_flags|O_CREAT, mode)) == -1) {
      return CACHE1_FAIL_OPEN;
    }
    /* Se tuvo que crear */
    ret = CACHE1_CREATED;
    h->page_size = CACHE1_PAGESIZE;
    h->divs = divs;
    memset(&h->first_index,0,sizeof(h->first_index));
    h->elements_per_page = divs + 1;
    h->chunk_size = h->page_size / h->elements_per_page;
    /*(u_int32_t)ceil(sizeof(cache1_head)*1.0/(h->page_size/elem_per_page));*/
    h->first_index.elem = (unsigned int)ceil((sizeof(cache1)*1.0)/
      h->chunk_size);
    /*(u_int32_t)ceil(sizeof(cache1_head)*1.0/(h->page_size/elem_per_page));*/
    if(h->first_index.elem >= h->elements_per_page ){
      h->first_index.page = 1;
      h->first_index.elem = 0;
    }else{
      h->first_index.page = 0;
    }
    h->last_position = h->first_index;
    h->version = CACHE1_VERSION;
    lseek(c->fd, 0, SEEK_SET);
    write(c->fd, h, sizeof(cache1_head));
  } else {
    ret = CACHE1_OPENED;
    lseek(c->fd, 0, SEEK_SET);
    read (c->fd, h, sizeof(cache1_head));
  }
  _seek++;
  cache1_hash_create(c, spages, dpages);
  c->max_pages = dpages + spages;
  /*fprintf(stderr, "max: %d, dpages: %d, spages: %d\n", c->max_pages, dpages, spages);*/
  memset(&index, 0, sizeof(index));
  list_head_init(&c->zero_list,0,NULL);
  cache1_insert_new_page(c, &p, &index);
  memcpy(p->page, c, sizeof(cache1));
  cache1_page_set_modified(p);
  free(c);
  c = (cache1 *)p->page;
  cache1_init(c);
  *cache = c;
  return ret;
}

int _cache1_page_sync_ (void *info, list_node *node)
{
  struct cache1_walk_info *w = (struct cache1_walk_info *) info;
  if (w->is_static) {
    cache1_page_sync((cache1_page *)node, w->cache);
  } else {
    cache1_page_sync((cache1_page *)node->ld_vpointer, w->cache);
  }
  return 1;
}

int cache1_sync_all_pages(cache1 *c)
{
  if (c->write_mode == 0) return 0;
  pthread_mutex_lock(&c->mutex_idx);
  pthread_mutex_lock(&c->mutex_new);
  cache1_hash_walk(c, _cache1_page_sync_);
  pthread_mutex_unlock(&c->mutex_new);
  pthread_mutex_unlock(&c->mutex_idx);
  return 0;
}

/* Cierra una cache, supone que unicamente hay un thread de control */

int cache1_close(cache1 *c)
{
  index_t index;
  cache1_sync_all_pages(c);
  fprintf(stderr, "seeks: %d\n", _seek);
  fprintf(stderr, "PET REUSE: %d, TRUE: %d, FALSE: %d\n",
	  __reuse, __reuseTRUE, __reuseFALSE);
  fprintf(stderr, "pool size: %d, max pages: %d\n",
	  c->pool->size, c->max_pages);
  fprintf(stderr, "get_index: %d, get_new: %d, find: %d\n",
	  __get_index, __get_new, __find);
  fprintf(stderr, "writes: %d, reads: %d\n", __write, __read);

  memset(&index, 0, sizeof(index));
  cache1_put_element(c, &index, 1);
  pthread_mutex_destroy(&c->mutex_idx);
  pthread_mutex_destroy(&c->mutex_new);
  cache1_hash_destroy(c);
  /* #ifndef CACHE1_USE_MMAP
     lseek(c->fd, 0, SEEK_SET);
     write(c->fd, c, sizeof(cache1_head));
     #endif */
  /*free(c);*/
  return 0;
}

/* Regresa una pgina que puede ser reutilizada, o NULL si no es
   posible o no hay necesidad */
cache1_page* cache1_get_unused_page(cache1 *c, const index_t *index)
{
  int data;
  cache1_page *page = NULL;
  /*__reuse++;*/
  if (c->pool->size < c->max_pages) return NULL;
  /*if (c->zero_list.size > 0 ) {
    fprintf(stderr, ">>> pool size: %d, max pages: %d, zero_size: %d\n",
    c->pool->size, c->max_pages, c->zero_list.size);
    }
    if (c->zero_list.size > 14) {
    fprintf(stderr, "\n\n******************* \nzero_list %d\n",
    c->zero_list.size);
    //exit(-1);
    } */
  if (list_head_get_size(&c->zero_list) > 0) {
    /*__reuseTRUE++;*/
    page = (cache1_page *) list_head_pop(&c->zero_list, &data);
    cache1_hash_remove(c,page);
    if (cache1_page_reuse(page, c, index) == -1) {
      fprintf(stderr,"cache1_get_unused_page: fail");
      cache1_page_sync(page, c);
      cache1_page_free(page, c);
      page = NULL;
#ifdef _DEVELOP_VERSION
      exit(-2);
#endif
    }
  } /* else __reuseFALSE++;*/
  return page;
}

int cache1_insert_new_page(cache1 *c, cache1_page **p, const index_t *index)
{
  cache1_page *page = NULL;
  page = cache1_get_unused_page(c, index);
  if (page == NULL) {
    cache1_page_create(&page, c, index);
  }
  page->refs = 1;
  cache1_hash_insert(c, page);
  *p = page;
  return 0;
}

int cache1_set_modified(cache1 *c, const index_t *idx)
{
  cache1_page *page;
  pthread_mutex_lock(&c->mutex_idx);
  cache1_hash_find(c, &page, idx);
  if (page == NULL) {
    fprintf(stderr,"cache1_set_modified: no page %d\n",idx->page);
    return -1;
  }
  page->modified = 1;
  pthread_mutex_unlock(&c->mutex_idx);
  return 0;
}

int cache1_put_element(cache1 *c, const index_t *idx, int modified)
{
  cache1_page *page;
  pthread_mutex_lock(&c->mutex_idx);
  cache1_hash_find(c, &page, idx);
  if (page == NULL) {
    fprintf(stderr,"cache1_put_element: no page %d\n",idx->page);
    return -1;
  }
  if (modified && page->modified == 0) page->modified = 1;
  cache1_page_del_ref(page, c);
  pthread_mutex_unlock(&c->mutex_idx);
  return 0;
}

void cache1_page_set_modified(cache1_page *p)
{
  if (p) p->modified = 1;
}

/* Busca el elemento con indice idx en la cache1 c, pone los datos en *p.
   Regresa 0 si la operacin tuvo exito. */

cache1_page* cache1_get_by_index(cache1 *c, void **p, const index_t *idx)
{
  cache1_page *page;
  *p = NULL;
  pthread_mutex_lock(&c->mutex_idx);
#ifdef _DEVELOP_VERSION_XXX_XXX_YYY
  *if (idx->page > c->header.last_position.page) {
    fprintf(stderr, "cache1_get_by_index: idx > last_position (%d > %d)\n",
	    idx->page, c->header.last_position.page);
    *p = NULL;
    pthread_mutex_unlock(&c->mutex_idx);
    memset(NULL, NULL, 1024);
    return NULL;
    /*c->header.last_position = *idx;
      goto cache1_get_by_index_and_new_element;*/
  }
#endif
  //fprintf(stderr, "cache1_get_by_index page: %d\n", idx->page);
  __get_index++;
  /* mutex_zero_refs para que no nos quiten la pgina encontrada
     con con cache1_hash_search antes de reclamarla */
  if (cache1_hash_find(c, &page, idx)) {
    __find++;
    cache1_page_add_ref(page, c);
  }
  if (page == NULL) {
  cache1_get_by_index_and_new_element:
    cache1_insert_new_page(c,&page,idx);
  }
  *p = (void *)((char *)page->page +
		idx->elem*((cache1_head *)c)->chunk_size);

#ifdef CACHE1_ACCESS_WITHOUT_WRITE
  page->counter++;
  if (page->counter > CACHE1_ACCESS_WITHOUT_WRITE) {
    cache1_page_sync(page, c);
    page->counter = 0;
  }
#endif
  pthread_mutex_unlock(&c->mutex_idx);
  return page;
}

/* Obtiene un nuevo elemento de c, pone los datos en *p, y su indice
   en *idx (si idx != NULL). Regresa 0 si la operacin tuvo exito.
   Aqu no se aprovecha espacio perdido, eso es responsabilidad de
   datos y tipos que utilicen cache1 como cache.
   Se mantiene una referencia a |*idx| */
int cache1_get_new_element(cache1 *c, void **data, index_t *idx)
{
  cache1_page *p;
  cache1_head *h = (cache1_head *)c;
  pthread_mutex_lock(&c->mutex_new);
  __get_new++;
  if (h->last_position.elem == h->elements_per_page) {
    pthread_mutex_lock(&c->mutex_idx);
    h->last_position.page++;
    h->last_position.elem=0;
    *idx = h->last_position;
    h->last_position.elem++;
    cache1_insert_new_page(c, &p, idx);
    pthread_mutex_unlock(&c->mutex_idx);
    idx->pos = 0; /* posicin no utilizada, poniendo a cero */
    *data = p->page;
  } else if (h->last_position.elem > h->elements_per_page) {
    fprintf(stderr, "GRAN ERROR\n");
    memset(NULL,NULL,1203);
  } else {
    *idx = h->last_position;
    h->last_position.elem++;
    idx->pos = 0;
    p = cache1_get_by_index(c, data, idx);
  }
  //fprintf(stderr, "cache1_new_element page: %d\n", idx->page);
  pthread_mutex_unlock(&c->mutex_new);
  cache1_page_set_modified(p);
  return 0;
}
