#ifndef _H_CACHE1__H__
#define _H_CACHE1__H__
#include <pthread.h>
#include <sys/types.h>
#include <list_node.h>
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64
#endif
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64
#endif 


#define C_PRINT_TREE(cache,PRI) fprintf(stderr,"\n***** PRINT_TREE: %p, %s\n", \
	    *(cache1_page **)cache->search_tree, PRI)
/*
#define PRINT_TREE(cache,PRI)
*/
/*#define CACHE1_USE_MMAP*/
/* Un indice */

#define CACHE1_CMP_INDEX(_x, _0, _1) \
  (_x) = (_0)->page - (_1)->page; \
  if ((_x) == 0) { \
    (_x) = (_0)->elem - (_1)->elem; \
    if ((_x) == 0) { \
      (_x) = (_0)->pos - (_1)->pos; \
    } \
  }

struct index_t {
  unsigned int page: 23; /* Nmero de Pgina. */
  unsigned int elem: 3; /* Elemento de pgina. 2^3 divisiones posibles al
			   tamao de pgina */
  unsigned int pos: 12; /* Posicin en la lista */
  unsigned int flags: 2; /* Flags para cada aplicacin */
} __attribute__ ((__packed__)) ;

/* To put into a 32 bit chunk */
/* struct index_t { */
/*   unsigned int page: 15; /\* Nmero de Pgina. *\/ */
/*   unsigned int elem: 3; /\* Elemento de pgina. 2^3 divisiones posibles al */
/* 			   tamao de pgina *\/ */
/*   unsigned int pos: 12; /\* Posicin en la lista *\/ */
/*   unsigned int flags: 2; /\* Flags para cada aplicacin *\/ */
/* } __attribute__ ((__packed__)) ; */

typedef struct index_t index_t;

/* Ligado a campo elem */
#define CACHE1_MAX_ELEMENTS_PER_PAGE (1<<3)
#define CACHE1_MAX_DIVS (CACHE1_MAX_ELEMENTS_PER_PAGE - 1)


#define INDEX_EQ(_idxa,_idxb) ((_idxa).page == (_idxb).page && \
   (_idxa).elem == (_idxb).elem)

#define INDEX_ZERO(_idxa) ((_idxa).page == 0 && (_idxa).elem == 0)

/* Posible problema con muchas threads al borrar cmo bloquear la?
   pgina borrada para que nadie la pida?
   Posible solucin: 
   bloqueo de subarbol padre para borrado
   Esto implica que posiblemente tendremos ms de
   max_pages en algn momento, pero siempre tender
   a max_pages el nmero de pginas.
   
   Aumentar la granuladidad (ms pequeo al principio del trie)
    
   Crear un trie para las posibles palabras de n primeras letras. Es decir
   por posible:
   - De una vocal pueden seguir consonantes y viceversa (con sus excepciones)
   y que crezca como abanico (crear con BFS)

   Qu sucede cuando cuando el archivo es 1k, pedimos offset de 20k y
   guardamos? Crece el archivo? Con que lo rellena?
*/


/* Es posible no tener que recrear las cache1_page mejor unicamente
   cerrar el map y poner los nuevos valores (menos manejo de memoria
   y recursos (mutex))
 */
struct cache1_page {
  /* Para la lista de remplazos, debe ser el primer elemento
     de la estructura */
  list_node node;
  /* Datos */
  void *page;
  /* Referencias a est pgina */
  unsigned int refs: 29;
  /* La pgina fue modificada?*/
  unsigned int modified: 1;
  /* Quin decide si cambia guardar.En cache1 y adems se puede poner por
     random con probabilidad p*/


  /*Contador de accesos */
  u_int16_t counter;
};

typedef struct cache1_page cache1_page;

/* Cada cache tiene un elemento inicial cabecera */
#define CACHE1_VERSION 1

struct struct_cache1_head {
  /* Versin para futuras actualizaciones y revisiones */
  u_int16_t version;
  /* last: ltima pgina en la cual se inserto */
  index_t last_position;
  /* Primer indice. Debe garantizar suficiente espacio para sta cabecera */
  index_t first_index;
  /* Tamao de cada pgina */
  size_t page_size;
  /* Nmero de divisiones */
  u_int16_t divs;
  /* Tamao de bloque */
  u_int16_t chunk_size;
  /* Nmero de elementos */
  u_int8_t elements_per_page;
  /*registros proposito general*/
  int32_t reg32a;
  int32_t reg32b;
  int64_t reg64a;
  int64_t reg64b;
} __attribute__ ((__packed__));

typedef struct struct_cache1_head cache1_head;

/*************** Cache Hash *************/
//#define CACHE1_ACCESS_WITHOUT_WRITE 2048
struct cache1_hash {
  u_int32_t size; /* Elementos actualmente en el hash */
  u_int32_t ssize; /* Tamao del array esttico */
  u_int32_t dsize; /* Tamao del array dinmico */

  cache1_page **spages; /* Arreglo punteros de pginas estticas */
  list_head *dpages; /* Arreglo punteros de pginas dinmicas */
};

typedef struct cache1_hash cache1_hash;

/* Hace cast a *cache1_head*/
#define CACHE1H(_h) ((cache1_head *)_h)
struct struct_cache1 {
  /* Es la parte de cada cache que se guarda y contiene informacin
     importante acerca de la cache. Debe ser el primer elemento en
     la esctructura para poder hacer cast */
  cache1_head header;
  /* Descriptor del archivo a utilizar  */
  int fd;
  /* Aqu se guardarn los bloques de memora, la
     llave es la posicin del archivo recorrida en 'shift' bytes
     (modificacin de granularidad). */
  cache1_hash *pool;
  /* Multiplicador de index_t.page para obtener posicin absoluta */
  u_int8_t shift;
  /* Mximo nmero de pginas en memora. La politica de liberacin debe ser
     del mayor indice para el trie y de menos usado recientemente para alist */
  u_int32_t max_pages;
  /* Usado para el remplazado/borrado de pginas: remplaza los que tengan
     cero referencias */
  list_head zero_list;
  pthread_mutex_t mutex_idx;
  pthread_mutex_t mutex_new;
  u_int8_t write_mode; /* Escribir(1) o solo leer(0) */
};

typedef struct struct_cache1 cache1;

#define CACHE1_ABSPOS(_cache, _index_page) \
  ((_index_page) << (_cache)->shift)

/* libera pagina de cache */
int cache1_page_free(cache1_page *p, cache1 *c);
/* crea una nueva pgina y la pone en *p */
int cache1_page_create(cache1_page **p, cache1 *c, const index_t *index);
/* Inicia una pgina con nuevos valores*/
int cache1_page_init (cache1_page  *p, cache1 *c, const index_t *index);
/* Reusa una pgina con nuevos valores*/
int cache1_page_reuse (cache1_page  *p, cache1 *c, const index_t *index);
/* Sincroniza una pgina */
int cache1_page_sync(cache1_page* p, cache1 *c);
/* Marca la cache como modificada*/
void cache1_page_set_modified(cache1_page *p);
/*********** CACHE1 ************/
#define CACHE1_FAIL_OPEN -989
#define CACHE1_OPENED -988
#define CACHE1_CREATED -98887

/* Habre una nueva cache, si no existe la crea.
   divs: Nmero de divisiones en la pgina. 0 Significa un solo elemento (sin divisiones).

   Errores:
   CACHE1_FAIL_OPEN: No se pudo abrir cache
   Informacin:
   CACHE1_CREATED: Se creo la cache en lugar de abrirla
   CACHE1_OPENED: Se abrio la cache 
*/
int cache1_open (cache1** cache,
		 const char *file, int mode,
		 int spages,int dpages, int divs);
/* Obtiene un nuevo elemento de c, pone los datos en *p, y su indice
   en *idx (si idx != NULL). Regresa 0 si la operacin tuvo exito. */
int cache1_get_new_element(cache1 *c, void **p, index_t *idx);

/* Busca el elemento con indice idx en la cache1 c, pone los datos en *p.
   Regresa la cache1_page donde se encuentran los datos */
cache1_page* cache1_get_by_index(cache1 *c, void **p, const index_t *idx);

/*Marca la pgina conteniendo |idx| como modificada */
int cache1_set_modified(cache1 *c, const index_t *idx);
/* Le dice a la cache que ya no est siendo utilizado el elemento con indice
   idx, no asegura guardar los datos, unicamente permite reutilizar
   la pgina. */

int cache1_put_element(cache1 *c, const index_t *idx, int modified);

/* Sincroniza toda la cache */
int cache1_sync_all_pages(cache1 *c);

/* Sincroniza la pgina que contiene idx */
int cache1_sync(cache1 *c, const index_t *idx);

/* Cierra la cache */
int cache1_close(cache1 *c);
/* Inserta una nueva pgina */
int cache1_insert_new_page(cache1 *c, cache1_page **p, const index_t *index);

#endif /* _H_CACHE1____*/
