#include <document.h>
#include <super_malloc.h>
#include <string.h>
#include <stdio.h>
#include <alfabeto.h>
#include <id_word_t.h>
#include <string.h>
#include <debug.h>
#include <skip_list.h>
/********** string stream *********/

int document_string_getc(document *d)
{
  if (d->len == d->pos) return -1;
  return d->string[d->pos++];
}

int document_string_tell(document *d)
{
  return d->pos; 
}

int document_file_getc(document *doc)
{
  if (doc->len == doc->pos) {
    if (gzeof(doc->file)) return -1;
    doc->len = gzread(doc->file, doc->string, doc->bufflen);
    if (doc->len == 0) return -1;
    if (doc->len == -1) {
      perror("document_file_getc");
      return -1;
    }
    doc->abs += doc->pos;
    doc->pos = 0;
  }
  return doc->string[doc->pos++];
}

int document_file_tell(document *d)
{
  return d->abs + d->pos;
}

void document_file_close(document *doc)
{
  gzclose(doc->file);
  free(doc->string);
  document_close_generic(doc);
}

/* Inicia el documento leyendo desde archivo */
int document_open_file(document **d, const char *path, int bufflen)
{
  document *doc = struct_malloc(document);
  doc->string = (char *)super_malloc(sizeof(char *)*bufflen);
  doc->bufflen = bufflen;
  doc->len = 0;
  /*doc->id = *docid;*/
  doc->file = gzopen(path,"rb");
  doc->pos = 0;
  doc->abs = 0;
  if (doc->file == NULL) {
    free(doc->string);
    document_close_generic(doc);
    *d = NULL;
#ifdef _DEVELOP_VERSION
    perror("document_open_private");
    fprintf(stderr,"path='%s'\n", path);
    /*exit(-1);*/
#endif
    return 1;
  }
  doc->get_character = document_file_getc;
  doc->tell = document_file_tell;
  doc->close = document_file_close;
  *d = doc;
  return 0;  
}

int document_open_string(document **d, char *string, int len)
{
  document *doc = struct_malloc(document);
  /*doc->id = *id;*/
  doc->pos = 0;
  doc->len = len;
  doc->string = string;
  doc->get_character = document_string_getc;
  doc->tell = document_string_tell;
  doc->close = document_close_generic;
  *d = doc;
  return 0;
}

/* Cierra el documento */
void document_close_generic(document *doc)
{
  free(doc);
}

/* Lee una palabra en sbuffer, de maximo 'max' car�teres, pone en *len
   la longitud de la palabra, y en pos pone la posici� en bytes
   de la primera letra de la palabra leida (Si pos != NULL).
   Regresa 0 si no leyo nada por error o terminaci� del flujo de datos y
   uno si puede seguir leyendo. */

int document_read_word(document *doc, id_wrd_t *wrd, int bufflen)
{
  register int i, c;
  /* probar la rapidez de gztell en terminos de llevar el conteo nosotros
      de cuantos bytes hemos le�o */
  /*wrd->pos = doc->tell(doc);*/
  for (i = 0; i < bufflen; i++) {
  document_read_word_label:
    c = doc->get_character(doc);
    if (c < 0) return 0;
    /*if (c == -1 && i == 0) return 0; PUEDE OCURRIR QUE C == -1 E i != 0 */
    c = alfa_get_index_by_char(c);
    /* consume la palabra completa si WRD_MAXLEN se alcanza,
       pero no se ponen en sbuffer */
    if (c == -1) {
      //if (c == -1) {
      if (i == 0) goto document_read_word_label;
      break;
    }
    if (WRD_MAXLEN < i) goto document_read_word_label; 
    wrd->key[i]=alfa_get_char_by_index(c);
  }
  wrd->key[i]='\0'; /* para impresion chida en las pruebas */
  wrd->len = i;
  return 1;
}



/*********************************************************/
/*********** Indexamiento de un solo documento ***********/
/*********************************************************/

void* document_read_dup(void *a, void *b)
{
  ((word_freq *)b)->num_docs++;
  return b;
}

void document_read_free(void *data)
{
  free(((word_freq *)data)->wrd);
  free(data);
}

void document_read_close(skip_list *list)
{
  skip_list_free(list, document_read_free);
}

void* document_read_alloc(void *a)
{
  word_freq *wf = struct_malloc(word_freq);
  wf->wrd = strdup((char *)a);
  wf->num_docs = 1;
  return wf;
}

int document_read_print(skip_node *n, void *arg)
{
  word_freq *wf = (word_freq *) n->data;
  fprintf(stderr, "%s> %s, %d\n", (char *) arg, wf->wrd, wf->num_docs);
  return 1;
}

int document_read_cmp(void *a, void *b)
{
  return strcmp((char *)a,((word_freq *)b)->wrd);
}

skip_list* document_read_document(document *doc, id_wrd_t *wrd, id_doc_t *idoc)
{
  skip_list *wlist = skip_list_new(0.5, 4096/6, document_read_cmp,
				   document_read_dup, document_read_alloc);
  while (document_read_word(doc, wrd, 63)) {
    if (wrd->len == 0) continue;
    skip_list_insert(wlist, wrd->key);
  }
  return wlist;
}

void document_save_vectorized(skip_list *wlist, char *filename)
{
  skip_list_iterator *witer;
  skip_node *wnode;
  word_freq *wdata;
  FILE *f = fopen(filename, "w+");
  if (f == NULL) {
    perror("document_save_vectorized: ");
    fprintf(stderr,"'%s'",filename);
    return;
  }
  skip_list_iterator_new(wlist, &witer);
  while (skip_list_iterator_has_next(witer)) {
    wdata = (word_freq *)skip_list_iterator_next(witer);
    fprintf(f, "%s: %d\n", wdata->wrd, wdata->num_docs);
  }
  skip_list_iterator_close(witer);
  fclose(f);
}
