#ifndef _H_DOCUMENT_H_
#define _H_DOCUMENT_H_
/* id de documento */
#include <sys/types.h>
#include <zlib.h>
#include <stdio.h>
#include <config_main.h>
#include <id_wrd_t.h>
#include <id_doc_t.h>
#include <skip_list.h>

struct word_freq {
  char * wrd; /*palabra codificada*/
  u_int16_t num_docs; /*longitud de palabra*/
};


typedef struct word_freq word_freq;

struct document {
  int pos; /* posicin en la cadena: stream = string*/
  int abs; /* posicin absoluta */
  int len; /* longitud cadena */
  int bufflen; /* Longitud del buffer (si es de archivo )*/
  gzFile file;
  char *string;
  int (*get_character)(struct document *);
  int (*tell)(struct document *);
  void (*close)(struct document *);
};
typedef struct document document;

/* Abre el documento leyendo desde archivo.
   'id' es el identificador de archivo que se utilizara para las entradas
   'd' es la direccin del puntero donde se pondra el resultado
   'path' es la ruta al archivo
   'bufflen' tamao del buffer a leer */
int document_open_file(document **d, const char *path, int bufflen);

/* Abre un documento leyendo desde string */
int document_open_string(document **d, char *string, int len);

/* Cierra el documento */
void document_close_generic(document *doc);

/* Lee una palabra en sbuffer, de a lo ms 'max' carcteres, rellena
   los campos de |wrd| con la informacin sobre la posicin y la longitud
   de la palabra.
   Regresa 0 si no leyo nada por error o terminacin del flujo de datos y
   uno si puede seguir leyendo. */

int document_read_word(document *doc, id_wrd_t *wrd, int max);

/* Lee un documento entero, regresa su versin ordenada en un |skip_list|,
   cada elemento es del tipo |word_count| */
skip_list* document_read_document(document *doc, id_wrd_t *buffer,
				  id_doc_t *idoc);
#endif /* _H_DOCUMENT_H_*/
