#include <document_env.h>
#include <super_malloc.h>
#include <string.h>
#include <stdio.h>
#include <debug.h>
#include <id_wrd_t.h>
#include <alfabeto.h>
#include <misc.h>

void document_env_create_or_open(document_env **env, const char *basename,
				 cache1 *base, int perm,
				 int (*dbt_cmp_func)(DB *, const DBT*,
						     const DBT*))
{
  document_env *e = struct_malloc(document_env);
  int error;
  char *name_from_id_to_name, *name_from_name_to_id, *name_from_id_to_maxfreq;
  if (sizeof(id_doc_t) != sizeof(index_t)) {
    fprintf(stderr, "id_doc_t != index_t. %d != %d\nLa alineacin de los "
	    "campos 'flags' tambin debe coincidir.\n",
	    sizeof(id_doc_t), sizeof(index_t));
    exit(-1);
  }
  name_from_id_to_name = xstrcat(basename, ".from_id_to_name");
  name_from_id_to_maxfreq = xstrcat(basename, ".from_id_to_maxfreq");
  name_from_name_to_id = xstrcat(basename, ".from_name_to_id");
  pthread_mutex_init(&e->mutex, NULL);
  e->max = (id_doc_t *)&base->header.reg32a;
  e->num_docs = (u_int64_t *)&base->header.reg64b;

  if (db_create(&e->from_id_to_name, NULL, 0)) {
    perror("document_env_create_or_open: from_id_to_name");
    exit(-2);
  }

  e->from_id_to_name->set_bt_compare(e->from_id_to_name,dbt_cmp_func);

  if (db_create(&e->from_id_to_maxfreq, NULL, 0)) {
    perror("document_env_create_or_open: from_id_to_maxfreq");
    exit(-2);
  }
  e->from_id_to_maxfreq->set_bt_compare(e->from_id_to_maxfreq, dbt_cmp_func);

  if (db_create(&e->from_name_to_id, NULL, 0)) {
    perror("document_env_create_or_open: from_name_to_id");
    exit(-2);
  }

  if (0600 & perm) {
    error = xdbopen(e->from_id_to_name, name_from_id_to_name, 0, perm);
    if (!error) error = xdbopen(e->from_id_to_maxfreq, name_from_id_to_maxfreq, 0, perm);
    if (!error) error = xdbopen(e->from_name_to_id, name_from_name_to_id, 0, perm);
    if (error) { /* create a new one  */
      error = xdbopen(e->from_id_to_name, name_from_id_to_name, DB_CREATE, perm);
      if (!error) error = xdbopen(e->from_id_to_maxfreq, name_from_id_to_maxfreq, DB_CREATE, perm);
      if (!error) xdbopen(e->from_name_to_id, name_from_name_to_id, DB_CREATE, perm);
      if (error) {
	perror("document_env_create_or_open DB_CREATE");
	exit(-2);
      }
      /* El id_doc_t puede no respetar el tamao de 32 bits, asi que no se
	 deben usar los registros necesarios para que quepa, el ltimo registro
	 base->header.reg64b tiene el nmero total de documentos, controlado
	 por get_id */
      base->header.reg32a = base->header.reg32b = 0;
      base->header.reg64a = base->header.reg64b = 0;
    }
  } else {
    error = xdbopen(e->from_id_to_name, name_from_id_to_name, DB_RDONLY, perm);
    if (!error) error = xdbopen(e->from_id_to_maxfreq, name_from_id_to_maxfreq, DB_RDONLY, perm);
    if (!error) xdbopen(e->from_name_to_id, name_from_name_to_id, DB_RDONLY, perm);
    if (error) {
      perror("document_env_create_or_open DB_RDONLY");
      exit(-2);
    }
  }
  free(name_from_id_to_name);
  free(name_from_id_to_maxfreq);
  free(name_from_name_to_id);
  *env = e;
  e->set_maxfreq = document_env_set_maxfreq;
  e->get_maxfreq = document_env_get_maxfreq;
  e->get_name = document_env_get_name_by_id;
}

int _docenv_bt_cmp(DB *db, const DBT *dbt1, const DBT *dbt2)
{
  return id_doc_cmp((const id_doc_t*)dbt2->data, (const id_doc_t*)dbt1->data);
}

int _docenv_bt_cmp_major(DB *db, const DBT *dbt1, const DBT *dbt2)
{
  return id_doc_cmp_major((const id_doc_t*)dbt2->data, (const id_doc_t*)dbt1->data);
}

void document_env_open_always_increment(document_env **env, cache1 *base,
					int perm, const char *basename)
{
  document_env *e;
  document_env_create_or_open(&e, basename, base, perm, _docenv_bt_cmp);
  e->close = document_env_close;
  e->get_id = document_env_get_id_always_increment;
  *env = e;
}

void document_env_open_increment(document_env **env, cache1 *base, int perm,
				 const char *basename)
{
  document_env *e;
  document_env_create_or_open(&e, basename, base, perm,
			      _docenv_bt_cmp_major);
  e->close = document_env_close;
  e->get_id = document_env_get_id_just_increment;
  *env = e;
}

void document_env_close(document_env *e)
{
  pthread_mutex_destroy(&e->mutex);
  e->from_name_to_id->close(e->from_name_to_id, 0);
  e->from_id_to_name->close(e->from_id_to_name, 0);
  e->from_id_to_maxfreq->close(e->from_id_to_maxfreq, 0);
  free(e);
}

void document_env_open_unique_name(document_env **env, cache1 *base, int perm,
				   const char *basename)
{
  document_env *e;
  document_env_create_or_open(&e, basename, base, perm,
			      _docenv_bt_cmp);
  e->close = document_env_close;
  e->get_id = document_env_get_id_by_name;
  *env = e;
}

void document_env_sync(document_env *env)
{
  pthread_mutex_lock(&env->mutex);
  env->from_id_to_name->sync(env->from_id_to_name, 0);
  env->from_id_to_maxfreq->sync(env->from_id_to_maxfreq, 0);
  env->from_name_to_id->sync(env->from_name_to_id, 0);
  pthread_mutex_unlock(&env->mutex);
}

void document_env_set_maxfreq (document_env *env, const id_doc_t *doc,
				  u_int16_t maxfreq)
{
  int error;
  DBT key, dat;
  DB * from_id_to_maxfreq = env->from_id_to_maxfreq;
  memset(&key, 0, sizeof(key));
  memset(&dat, 0, sizeof(dat));
  key.data = (void *)doc;
  key.size = sizeof(id_doc_t);
  dat.flags = 0;
  dat.size = sizeof(maxfreq);
  dat.data = (void *)&maxfreq;
  error=from_id_to_maxfreq->put(from_id_to_maxfreq, NULL, &key, &dat, 0);
  if (error) {
    from_id_to_maxfreq->err(from_id_to_maxfreq, error, "document_env_set_maxfreq: %s",
			    db_strerror(error));
    id_doc_print(doc, "Guardando maxfreq");
  }
}

u_int16_t document_env_get_maxfreq(document_env *env, const id_doc_t *doc)
{
  int error;
  DB * from_id_to_maxfreq = env->from_id_to_maxfreq;
  u_int16_t res;
  DBT v, k;
  memset(&v, 0, sizeof(v));
  memset(&k, 0, sizeof(k));
  v.ulen = sizeof(res);
  v.data = &res;
  v.flags = DB_DBT_USERMEM;
  k.data = (void *)doc;
  k.size = sizeof(id_doc_t);
  res = 0;
  error = from_id_to_maxfreq->get(from_id_to_maxfreq, NULL, &k, &v, 0);
  if (error) {
    from_id_to_maxfreq->err(from_id_to_maxfreq, error, "document_env_get_maxfreq: %s",
		 db_strerror(error));
    id_doc_print(doc, "Fallo buscando frequecia");
    return 1;
  }
  return res;
}


void document_env_store_name(document_env *env, const id_doc_t * doc,
			     const char *name, u_int16_t len)
{
  int error;
  DBT key, dat;
  DB *from_id_to_name = env->from_id_to_name;
  memset(&key, 0, sizeof(key));
  memset(&dat, 0, sizeof(dat));
  key.data = (void *)doc;
  key.size = sizeof(id_doc_t);
  dat.data = (void *)name;
  dat.size = len;
  error=from_id_to_name->put(from_id_to_name, NULL, &key, &dat, 0);
  if (error) {
    from_id_to_name->err(from_id_to_name, error, "document_env_store_name: %s",
			 db_strerror(error));
    id_doc_print(doc, "Error guardando ");
  }
}

char * document_env_get_name_by_id (document_env *env, const id_doc_t *doc,
				    char *buffer, u_int16_t blen)
{
  int error;
  DBT v, k;
  DB *from_id_to_name = env->from_id_to_name;
  memset(&k, 0, sizeof(k));
  memset(&v, 0, sizeof(v));
  if (buffer != NULL) {
    v.data = buffer;
    v.ulen = blen;
    v.flags = DB_DBT_USERMEM;
  } else {
    v.flags = DB_DBT_MALLOC;
  }
  k.data = (void *)doc;
  k.size = sizeof(id_doc_t);
  error=from_id_to_name->get(from_id_to_name, NULL, &k, &v, 0);
  if (error) {
    from_id_to_name->err(from_id_to_name, error, "document_env_get_name_by_id: %s",
		      db_strerror(error));
    if (buffer != NULL) {
      buffer[0] = '\0';
    } else {
      return NULL;
    }
  }
  return (char *)v.data;
}


void document_env_get_id_by_name(document_env *env, const char *name, id_doc_t *d)
{
  int error;
  DBT k, v;
  DB *from_name_to_id = env->from_name_to_id;
  memset(&v, 0, sizeof(v));
  memset(&k, 0, sizeof(k));

  v.ulen = sizeof(id_doc_t);
  v.data = (void *)d;
  v.flags = DB_DBT_USERMEM;

  k.data = (void *)name;
  k.size = strlen(name) + 1;
  error = from_name_to_id->get(from_name_to_id, NULL, &k, &v, 0);
  if (error) {
    /* from_name_to_id->err(from_name_to_id, error,
			 "document_env_get_id_by_name: %s",
			 db_strerror(error)); */
    document_env_get_id_always_increment(env, name, d);
  }
}


void document_env_get_id_just_increment(document_env *env, const char *name,
					id_doc_t *doc)
{
  int len;
  pthread_mutex_lock(&env->mutex);
  (*env->num_docs)++;
  env->max->major++;
  doc->major = env->max->major;
  doc->minor = 0;
  if (name != NULL) {
    len = strlen(name);
    if (len > 0) document_env_store_name(env, doc, name, len + 1);
  }
  pthread_mutex_unlock(&env->mutex);
}

void document_env_get_id_always_increment(document_env *env, const char *name,
					  id_doc_t *doc)
{
  int len;
  pthread_mutex_lock(&env->mutex);
  (*env->num_docs)++;
  if (env->max->minor == ID_DOC_MINOR_MAX) {
    env->max->major++;
    env->max->minor=0;
  } else {
    env->max->minor++;
  }
  doc->major = env->max->major;
  doc->minor = env->max->minor;
  if (name != NULL) {
    len = strlen(name);
    if (len > 0) document_env_store_name(env, doc, name, len + 1);
  }
  pthread_mutex_unlock(&env->mutex);
}
