#ifndef _H_DOCUMENT_ENV_
#define _H_DOCUMENT_ENV_

#include <sys/types.h>
#include <pthread.h>
#include <idx_man.h>
#include <text_trie.h>
#include <id_word_t.h>
#include <id_doc_t.h>

#ifdef __FreeBSD__
#include <db42/db.h>
#else
#include <db.h>
#endif

#include <fcntl.h>
#include <cache1.h>

typedef struct document_information document_information;

struct document_env {
  pthread_mutex_t mutex;
  id_doc_t *max; /*id_doc_t mximo, util para get_id con historia*/
  u_int64_t *num_docs; /* El nmero total de documentos, controlado en get_id */
  /* Cierra el document_env */
  void (*close)(struct document_env *);
  /* Pone en |doc| el id correspondiente para |orig|, tambin controla el
     nmero total de documentos (num_docs) */
  void (*get_id)(struct document_env *env, const char *orig,  id_doc_t *doc);
  /* Regresa el nombre identificado por |doc|, lo pone en |buffer| Si es
     diferente de NULL (con un mximo de |blen| bytes). Si |buffer| == NULL
     regresa memoria dinamica creada por malloc */
  char* (*get_name)(struct document_env *env, const id_doc_t *doc,
		    char *buffer, u_int16_t blen);
  /* Pone la frecuencia maxima al documento */
  void (*set_maxfreq)(struct document_env *env, const id_doc_t *doc,
		      u_int16_t mfreq);
  u_int16_t (*get_maxfreq)(struct document_env *env, const id_doc_t *doc);

  DB *from_id_to_maxfreq;
  DB *from_id_to_name;
  DB *from_name_to_id;
};

typedef struct document_env document_env;

/* Crea un document_env que siempre incrementa el id del documento, no importa
   si ya existia en la coleccin */
void document_env_open_always_increment(document_env **env, cache1 *base,
					int perm, const char *basename);
/* Igual a always_increment, pero unicamente incrementa el campo |major| del
   id de documento, se debe poner el |minor| por la aplicacin */
void document_env_open_increment(document_env **env, cache1 *base, int perm,
				 const char *basename);

/* Sincroniza |env| con el disco */
void document_env_sync(document_env *env);

void document_env_close(document_env *env);
/* Despacha los id_doc de acuerdo con el nombre de los documentos, permite
   encontrar archivos duplicados (por nombre) */
/* void document_env_open_trie(document_env **env,cache1 *base,int size,int perm, */
/* 			    const char *basename); */
/* void document_env_close_trie(document_env *e); */

/* Frecuencia mxima del documento |doc|. u_int16_t igual a idx_data */
void document_env_set_document_information(document_env *env,
					   const id_doc_t *doc,
					   document_information *docinfo);
/* regresa los datos de un documento del document |doc|*/
document_information *
document_env_get_document_information(document_env *env, const id_doc_t *doc);
void document_env_get_id_by_name(document_env *e, const char *o, id_doc_t *d);
void document_env_get_id_just_increment(document_env *env, const char *tonull,
					id_doc_t *doc);
void document_env_get_id_always_increment(document_env *env,
					  const char *tonull, id_doc_t *doc);

/* int document_env_trie_handler(text_trie *trie, trie_node *node,
   id_wrd_t *wrd, id_doc_t *doc, int index_pos, u_int8_t isfinal); */
char * document_env_get_name_by_id(document_env *e, const id_doc_t *doc, char *buffer, u_int16_t blen);

/* Frecuencia mxima del documento |doc|. u_int16_t igual a idx_data */
void document_env_set_maxfreq(document_env *env, const id_doc_t *doc,
			      u_int16_t maxfreq);
/* regresa la Frecuencia maxima del document |doc|*/
u_int16_t document_env_get_maxfreq(document_env *env, const id_doc_t *doc);

#endif
