#ifndef _H_IDX_MAN_
#define _H_IDX_MAN_
/* tipos de indexado */
#define IDX_TRIE 0
#define IDX_DB 1

#include <iterator.h>
#include <id_word_t.h>
#include <id_doc_t.h>
#include <sys/types.h>

struct struct_idx_man;
typedef struct struct_idx_man idx_man;

struct struct_idx_data {
#ifdef TEXT_TRIE_WITH_DATLIST
  index_t index; /* indice hacia la lista invertida de la palabra */
#endif
  u_int16_t freq; /* frecuencia de la palabra en el documento */
  id_doc_t id_doc; /* ID del documento que la contiene, en la lista
		      invertida se har referencia a ste nmero */
} __attribute__ ((__packed__));
typedef struct struct_idx_data idx_data;

struct struct_idx_global_wrd_data {
  u_int64_t global_freq: 48;  /* Frecuencia global */
  u_int64_t number_docs: 48; /* Numero de documentos con la palabra */
} __attribute__ ((__packed__));
typedef struct struct_idx_global_wrd_data idx_global_wrd_data;

struct struct_idx_wrd_data {
  unsigned int doc_position : 8; /*posicin en el documento */
  unsigned int tag: 8; /* etiqueta a la que pertenece */
} __attribute__((__packed__));
typedef struct struct_idx_wrd_data idx_wrd_data;

#define MAX_WRD_STRING 32
struct struct_idx_man {
  int idxtype;
  /* Para insertar una palabra wrd asociado al documento doc_id */
  int (*insert)(idx_man *idm, id_wrd_t *wrd, id_doc_t *doc_id);
  /* Borra la liga de wrd a doc_id */
  int (*remove)(idx_man *idm, id_wrd_t *wrd, id_doc_t *doc_id);
  /* Busca los documentos asociados a la palabra  wrd, el iterador
     se pone en el apuntador *itr. Regresa 1 si se encontro, de
     otra manera 0 */
  int (*find)(idx_man *idm, id_wrd_t *wrd, iterator **itr);
  /* Cierra y libera recursos de idx_man, Si flags != 0 entonces
     tambin se cierra la cache (o contenedor) asociado */
  int (*close)(idx_man *idm);
};

/* crea un id_wrd_t a partir de la informacin dada*/
id_wrd_t* idx_man_wrd_get_id(char *word, int size, int pos);
/* funcin de comparacin en lista invertida en_lista es un *idx_data
   y buscado *id_doc_t */
int idx_man_doc_find_cmp(void *en_lista, void *buscado);
#endif /*_H_IDX_MAN_*/
