#ifndef _H__ITERATOR_
#define _H__ITERATOR_
struct iterator;
typedef struct iterator iterator;
#include <cache1.h>

/* Se deben regresar errores en skip y jump para cuando se sale del espacio
   de memoria */

struct iterator {
  /* Devuelve verdadero si an hay ms elementos hacia adelante */
  int (*has_next)(iterator *);
  /* Devuelve verdadero si an hay ms elementos hacia atrs */
  int (*has_prev)(iterator *);
  /* Devuelve la siguiente posicin del iterador, si no existe
     lo crea y lo devuelve. Si solo se requiere iterar sobre elementos
     existentes se debe utilizar en conjunto con has_next */
  void * (*next)(iterator *);
  /* Recupera un elemento de tamao size y lo pone en data */
  int (*recover_element)(iterator *itr, void *data, int size);
  /* Inserta un elemento y lo regresa, siempre inserta a diferencia
     de next */
  void * (*insert)(iterator *);
  /* Devuelve la posicin anterior a la apuntada en el iterador, slo itera
     sobre elementos existentes, en caso de llegar al principio regresa
     al final (usar has_prev para prevenir ste comportamiento) */
  void * (*prev)(iterator *);
  /* Regresa el elemento actual sin moverse */
  void * (*current) (iterator *);
  /* Borra el elemento actual */
  int (*remove)(iterator *);
  /* Cierra el iterador */
  int (*close)(iterator *);
  /* Reset al inicio iterador */
  int (*reset)(iterator *);
  /* Pone el iterador en la posicin indicada por |idx| y regresa
     el valor de ese element */
  void * (*jump)(iterator *, const index_t* idx);
  /* Salta |n| posiciones en el iterador (n puede ser positivo o negativo)
     y regresa el valor del elemento encontrado */
  void * (*skip)(iterator *, int n);
 /* Salta |ceil(num_bytes/list->element_size)| posiciones en el iterador */
  void * (*skip_by_bytes)(iterator *, unsigned int n);

  /* Pone el indice actual en idx*/
  int (*get_current_index)(iterator *, index_t *idx);
  /* Encuentra la posicin de un elemento y se queda ah, en caso de no
     encontrar se en la siguiente posicin de la posicin que debera estar.
     Regresa 1 si
     encuentra la posicin, 0 de otra forma, asume una lista ordenada.
     iterator es el iterador
     busq. elemento a buscar
     cmp. funcin para comparar, regresa 0 si en_lista = busq, mayor a 0 si
     en_lista > busq, menor que 0 si busq > en_lista. */
  int (*find)(iterator *, void *busq, int (*cmp)(void *en_lista, void *busq));
  /* Igual a find, pero con el HINT que debe buscar hacia adelante */
  int (*find_forward)(iterator *, void *busq,
		      int (*cmp)(void *en_lista, void *busq));
  /* Igual, pero con el HINT que debe buscar hacia atras */
  int (*find_backward)(iterator *, void *busq,
		       int (*cmp)(void *en_lista, void *busq));
};

#endif /*_H__ITERATOR_*/
