#ifndef ___H_LIST_NODE__
#define ___H_LIST_NODE__
#include <sys/types.h>

struct list_node
{
  struct list_node *prev;
  struct list_node *next;
#define ld_int __list_data__.idata
#define ld_uint __list_data__.uidata
#define ld_vpointer __list_data__.vpointer
  union {
    int idata;
    unsigned int uidata;
    void *vpointer;
  } __list_data__;
};

typedef struct list_node list_node;
struct list_head
{
  struct list_node *first;
  struct list_node *last;
  int free_node;
  int (*free_data)(void *);
  u_int32_t size;
};

typedef struct list_head list_head;

/* Inicializa una lista y la pone en *list
   Si free_node != 0 se llama cuando se libera un nodo con el nodo como arg
   Si free_data != NULL se llama cuando se libera un nodo con el dato como arg
*/
void list_head_init(list_head *list, int free_node, int (*free_data)(void *));
/* Crea una nueva lista y la pone en *list
   Si free_node != 0 se llama cuando se libera un nodo con el nodo como arg
   Si free_data != NULL se llama cuando se libera un nodo con el dato como arg
*/
void list_head_new(list_head **list, int free_node, int (*free_data)(void *));
/* Libera una lista */
void list_head_free(list_head *list, int free_head);
/* Libera un nodo de la lista list_pointer */
int list_node_free(void *list_pointer, list_node *node);

/* Obtiene el tamao de la lista */
u_int32_t list_head_get_size(list_head *list);
/* Inserta un elemento al final de la lista. Complejidad O(1).
   Si node == NULL se crea un nuevo nodo con malloc (util cuando
   se especifica free_node y free_data) */
void list_head_push_integer(list_head *h, int d, list_node *node);
void list_head_push_vpointer(list_head *h, void *d, list_node *node);
void list_head_push_any(list_head *h, int di, void *dv, list_node *node);

/* Inserta |d| al inicio de la lista. Complejidad O(1)
   Si node == NULL se crea un nuevo nodo con malloc (util cuando
   se especifica free_node y free_data) */
void list_head_unshift_integer(list_head *h, int d, list_node *node);
void list_head_unshift_vpointer(list_head *h, void * d, list_node *node);
void list_head_unshift_any(list_head *h, int di, void *dv, list_node *node);

/* Borra un elemento al final de la lista. Regresa el puntero al nodo
   borrado (NULL si no se borro nada) y pone en |*ret| el dato.
   Si el puntero regresado es NULL entonces no se borro nada. Si se
   est utilizando malloc (free_node, free_data) el puntero
   regresado no podr ser utilizado, pero es indicador de borrado o no
   Complejidad O(1)*/
void* list_head_pop(list_head *h, int *ret);

/* Borra un elemento al principio de la lista. Regresa el puntero al nodo
   borrado (NULL si no se borro nada) y pone en |*ret| el dato.
   Si el puntero regresado es NULL entonces no se borro nada. Si se
   est utilizando malloc (free_node, free_data) el puntero
   regresado no podr ser utilizado, pero es indicador de borrado o no
   Complejidad O(1)*/
void* list_head_shift(list_head *h, int *ret);
/* Borra el nodo |n| de la lista |h|. Complejidad O(1)*/
void list_head_remove_node(list_head *h, list_node *n);
/* Recorre hacia adelante sobre los elementos de una lista*/
void list_head_iterator_next(list_head *h,
			     int (*visit)(void *, list_node *), void *arg);
/* Recorre hacia atras sobre los elementos de una lista*/
void list_head_iterator_prev(list_head *h,
			     int (*visit)(void *, list_node *), void *arg);
/* Imprime un nodo de la lista, para propositos de debug */
int list_node_print(void *arg, list_node *node);
#endif
