#include <trie_node.h>
#include <text_trie.h>
#include <alfabeto.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <cache1.h>
#include <idx_man.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <document_env.h>
#include <document.h>
#include <config_main.h>
#include <signal.h>
#include <debug.h>
#include <skip_list.h>
#include <string.h>
#include <sys/times.h>

#ifndef _NSIG
#define _NSIG 32
#endif
/*
  #ifdef __linux__
  #warning "Linux OS"
  #endif
  #ifdef ((__FreeBSD__))
  #warning "FreeBSD OS"
  #endif */

#define MAIN_LBUFFER 4096
#define MAIN_LBUFFER_READ 4095

struct thr_data {
  /* Buffer para guardar una entrada
     4096 para indexar archivos (ruta larga) */
  char buffer[MAIN_LBUFFER];
  /* Longitud de string buffer */
  int length;
  /* Hilo asociado */
  pthread_t thr;
};

/* Hilos a correr */
#define MAIN_NUM_THREADS 16
struct thr_data main_threads[MAIN_NUM_THREADS];

pthread_mutex_t main_mutex_read = PTHREAD_MUTEX_INITIALIZER;

/* Index Manager, visible por todos los hilos */
idx_man *main_idm;

/* Document environment */
document_env *main_env;

/* Archivo de entrada */
FILE *input;

/* Lee una linea de archivo 'file' con a lo ms 'max' carcteres y
   pone la longitud en len. Regresa un cero si ya no es posible
   seguir leyendo de 'file' y un  de otra manera */

int read_line(register char *sbuffer, FILE *file, int max, int *len)
{
  int i = 0, c;
  while ((c = fgetc(file)) != EOF) {
    if (c == '\n' || i == max) {
      sbuffer[i] = '\0';
      break;
    }
    /*if (c == '\r') break;*/
    sbuffer[i++] = c;
  }
  *len = i;
  return i != 0;
}

u_int32_t main_global_counter = 0;


/* Controla un hilo indexador */
void* main_index_file(void *argvoid)
{
  int s, index = (int)argvoid;
  int count, x;
  id_wrd_t wordread, wrd;
  id_doc_t doc_id;
  document *doc;
  int max_doc_freq;
  skip_list *wlist, *windex;
  skip_node *wnode;
  word_freq *wfreq;
  struct thr_data *data = &main_threads[index];

  wordread.key = (char *)super_malloc(sizeof(char)*64);
  data->thr = pthread_self();
  while (1) {
    x = 0;
    max_doc_freq = 0;
    pthread_mutex_lock(&main_mutex_read);
    s=read_line(data->buffer, input, MAIN_LBUFFER_READ, &data->length);
    count = ++main_global_counter;
    pthread_mutex_unlock(&main_mutex_read);
    if (s == 0) break; /* no se puede seguir leyendo */
    if (data->length == 0) continue; /* pierde su turno por no poder leer */
    if (document_open_file(&doc, data->buffer, 4096) || doc == NULL) continue;
    main_env->get_id(main_env, data->buffer, &doc_id);
    fprintf(stderr,"INI DOCUMENTO %d, %d, %s\n",count, (int)data->thr,
	    data->buffer);
    wlist = document_read_document(doc, &wordread, &doc_id);
    //document_save_vectorized(wlist, "somefile.dat");
    //windex = memindex_new();
    windex = memindex_read(doc, &wordread, &doc_id);
    memindex_dump(windex,"midx");
    memindex_close(windex);
    //exit(0);
    wnode = wlist->head;
    while (skip_node_next(wnode)->ntype == NODE) {
      wnode = skip_node_next(wnode);
      wfreq = (word_freq *)wnode->data;
      wrd.key = wfreq->wrd;
      wrd.freq = wfreq->num_docs;
      wrd.len = strlen(wfreq->wrd);
      id_wrd_encode(&wrd, ENCODING_TABLE);
      main_idm->insert(main_idm, &wrd, &doc_id);
      if (wfreq->num_docs > max_doc_freq) max_doc_freq = wfreq->num_docs;
    }
    document_read_close(wlist);
    main_env->set_maxfreq(main_env, &doc_id, max_doc_freq);
    fprintf(stderr,"FIN DOCUMENTO %d, %d\n",count, (int)data->thr);
    doc->close(doc);
  }
  free(wordread.key);
  return NULL;
}

void main_exit_graceful()
{
  debug_backtrace();
  fprintf(stderr, "Tratando de guardar... ");
  /* while (1) {
    fprintf(stderr, "Esperando GDB\n");
    usleep(3000000);
  }*/
  main_env->close(main_env);
  text_trie_close_caches(main_idm);
  main_idm->close(main_idm);  
  pthread_mutex_destroy(&main_mutex_read);
  fprintf(stderr,"hecho\n");
}

void main_sigaction(int s, siginfo_t *i, void *v)
{
  fprintf(stderr, "Manejando la seal %d\n",s);
  exit(1);
}

int main(int argc, char **argv)
{
  int i, ret, tmp;
  cache1 *idx_info, *dat_info;
  float tckpsec = sysconf(_SC_CLK_TCK);
  struct tms tini, tfin;

  atexit(main_exit_graceful);
  struct sigaction action;
  action.sa_sigaction = main_sigaction;
  action.sa_flags = 0;
  for (i = 1; i < _NSIG; i++) {
    if (i == SIGPROF||i==9||SIGSTOP) continue;
    sigaction(i, &action, NULL);
    /* signal(i, exit);*/
  }
  cache1_open(&idx_info, CFG_IDX_NAME, 0640, 0,
	      CFG_CREATE_CACHE_IDX_SIZE, CFG_CACHE_IDX_DIVS);
  cache1_open(&dat_info, CFG_DAT_NAME, 0640, 0,
	      CFG_CREATE_CACHE_DAT_SIZE, CFG_CACHE_DAT_DIVS);

/*document_env_open_trie(&main_env, idx_info, CFG_CREATE_DOC_ENV_SIZE, 0640);*/
  document_env_open_unique_name(&main_env, idx_info, 0640,
				     CFG_DOCENV_BASENAME);
  fprintf(stderr, "NUM_DOCS: %lld\n",*main_env->num_docs);
  main_global_counter = *main_env->num_docs;
  text_trie_open_master(&main_idm, CFG_IDM_NAME, 0640,
			CFG_CREATE_TRIE_SIZE, LONGITUD_ALFABETO, 5);
  text_trie_set_caches(main_idm, idx_info, dat_info);
  input = fopen("list.dat", "rb");
  if (input == NULL) {
    perror("main:input:");
    exit(3);
  }  

  times(&tini);
  if (1) {
    for (i = 0; i < MAIN_NUM_THREADS; i++) {
      ret=pthread_create(&main_threads[i].thr,NULL,main_index_file,(void *)i);
      if (ret != 0) {
	perror("main:pthread_create");
	fprintf(stderr, "Error posible EAGAIN: %d, ret: %d\n", EAGAIN, ret);
	exit(1);
      }
    }
    for (i = 0; i < MAIN_NUM_THREADS; i++) {
      ret=pthread_join(main_threads[i].thr, NULL);
      if (ret != 0) {
	perror("main:pthread_join");
      }
    }
  } else {
    tmp=0;
    main_threads[tmp].thr = pthread_self();
    main_index_file((void *)tmp);
  }

  times(&tfin);
  fprintf(stderr, "==== times:\nuser time:\t%f sec\nsystem time:\t %f sec\n", 
	  (tfin.tms_utime-tini.tms_utime)/tckpsec,
	  (tfin.tms_stime-tini.tms_stime)/tckpsec);

  fprintf(stderr,"\nYa estuvo, documentos totales %lld\n",*main_env->num_docs);
  i=0;
  /*text_trie_extract_thesaurus(main_idm, NULL, &i, main_threads[0].buffer,
			      0, MAIN_LBUFFER, NULL);
  */
  return 0;
}

