#include <stdio.h>
#include <super_malloc.h>
#include <idx_man.h>
#include <string.h>
#include <id_wrd_t.h>
#include <query.h>
#include <math.h>

int query_new(query **qry, document_env *env,
	      idx_man *idm, char **argv, int argc)
{
  int s;
  query *q = struct_malloc(query);
  q->argv = argv;
  q->argc = argc;
  q->idm = idm;
  q->env = env;
  s = argc * sizeof(iterator *);
  q->itrs = (iterator **)super_malloc(s);
  q->__itrs = (iterator **)super_malloc(s);
  memset(q->itrs, 0, s);
  memset(q->__itrs, 0, s);
  q->gdata = (idx_global_wrd_data*)super_malloc(argc*
						sizeof(idx_global_wrd_data));
  *qry = q;
  return 0;
}

void query_close(query *q)
{
  int i;
  for (i = 0; i < q->argc; i++) {
    if (!q->__itrs[i]) continue;
    q->__itrs[i]->close(q->__itrs[i]);
  }
  free(q->gdata);
  free(q->itrs);
  free(q->__itrs);
  free(q);
}

int query_init(query *q, int init_all)
{
  int i, res;
  id_wrd_t wrd;
  iterator *itr;
  q->collection_size = *(q->env->num_docs);
  for (i = 0; i < q->argc; i++) {
    wrd.len = strlen(q->argv[i]);
    wrd.key = strdup(q->argv[i]);
    itr = NULL;
    id_wrd_encode(&wrd, ENCODING_TABLE);
    res = q->idm->find(q->idm, &wrd, &itr);
    id_wrd_encode(&wrd, DECODING_TABLE);
    if (!res && init_all) {
      /*YA est en el query_new for (; i < q->argc; i++) q->__itrs[i] = NULL;*/
      return 0;
    }
    q->__itrs[i] = itr;
  }
  query_reset(q);
  return 1;
}

void query_reset(query *q)
{
  int i;
  iterator *itr;
  for (i = 0; i < q->argc; i++) {
    itr = q->__itrs[i];
    q->itrs[i] = itr;
    if (itr == NULL) {
      memset(&q->gdata[i], 0, sizeof(q->gdata[i]));
      continue;
    }
    itr->reset(itr);
    /* El primero nos dice datos globales */
    itr->recover_element(itr, &q->gdata[i], sizeof(q->gdata[i]));
    /*itr->skip_by_bytes(ai,sizeof(q->gdata[i]));*/
    /*itr->next(itr); Nos pone en el primer elemento */
  }
}

/* Mximo de todos los elementos "current" en los iteradores */
id_doc_t query_get_max(query *q)
{
  int i;
  idx_data *dat;
  id_doc_t doc = {0,0,0}; /* Todos tienen mayor a 0,0 */
  for (i = 0; i < q->argc; i++) {
    if (q->itrs[i] == NULL) continue;
    dat = (idx_data *) q->itrs[i]->current(q->itrs[i]);
    if (id_doc_cmp(&dat->id_doc, &doc) > 0) doc = dat->id_doc;
  }
  return doc;
}

/* Minimo de todos los elementos "current" en los iteradores */
id_doc_t query_get_min(query *q)
{
  int i;
  idx_data *dat;
  id_doc_t doc = {~0,~0,0}; /* El mayor posible, porque no hay signo */
  for (i = 0; i < q->argc; i++) {
    if (q->itrs[i] == NULL) continue;
    dat = (idx_data *) q->itrs[i]->current(q->itrs[i]);
    if (id_doc_cmp(&dat->id_doc, &doc) < 0) doc = dat->id_doc;
  }
  return doc;
}


/*********************************************/
/* query_result structure and related methods*/
/* int __query_skip_cmp_list(void *in_skip_list, void *new_data)
   {
   list_head *in_list = (list_head *)in_skip_list,
   *to_insert = (list_head *)new_data;
   //Siempre existe al menos un nodo para que la lista exista en el skiplist
   //return id_doc_cmp((id_doc_t *)list->first->data, (id_doc_t *)new_data);
   return id_doc_cmp((id_doc_t *)in_list->first->data,
   (id_doc_t *)to_insert->first->data);
   }*/
int __query_result_generic_free__(void *data)
{
  free(data);
  return 1;
}

void query_result_new(query_result **qr, int size,
		      void (*get_info)(query_result *, void *,
				       id_doc_t **, float *),
		      int (*cmp)(void *, void*),
		      int (*free_item)(void *))
{
  query_result *q = struct_malloc(query_result);
  q->psize = size;
  q->get_info = get_info;
  q->cmp = cmp;
  q->free_item = free_item ? free_item : __query_result_generic_free__;
  q->docs = skip_list_new(0.5, size, cmp, NULL, NULL);
  q->free = query_result_free;
  /* q->push = dup_handler ? query_result_push_dup : query_result_push_no_dup;
     q->pop = dup_handler ? query_result_pop_dup : query_result_pop_no_dup;
  */
  q->push = query_result_push_no_dup;
  q->pop = query_result_pop_no_dup;
  *qr = q;
}

/* Inserta el dato 'data' en  'qr->docs' y si el dato existe lo remueve y lo
   vuelve a insertar, sigue utilizando 'dup' para hacer saber al dato
   anterior que se realizan cambios, se reincerta para reacomodarse a
   un posible cambio de peso del nodo. El tamao de la lista es regulado
   por 'max', para una lista no regulada 'max' = 0.
   Regresa el puntero removido de la lista en caso de sobrepasar el nmero
   de nodos 'qr->psize'.

   Nota: Bajo el esquema utilizado los valores exactos del ranking no
   pueden ser alcanzados e incluso un buen elemento puede no aparecer,
   pues puede ser que un elemento eliminado por falta de un buen ranking sea
   eliminado y a travs del tiempo se alcance un buen nivel. Por eso se 
   deslizan qr->psize<<2 para dar una mayor oportunidad a los elementos */

/* void query_result_push_dup(query_result *qr, void *data) */
/* { */
/*   list_head *h; */
/*   skip_node *new_node = skip_list_find_intern(qr->docs, data, qr->docs->cmp); */
/*   new_node = skip_node_next(new_node); /\*skip_list_find nos deja antes *\/ */
/*   if (new_node != qr->docs->nil && qr->docs->cmp(new_node->data, data) == 0) { */
/*     list_head_push_vpointer((list_head *)new_node->data, data, NULL); */
/*     skip_list_insert(qr->docs, skip_list_remove_by_node((skip_list *)data, new_node)); */
/*   } else { */
/*     list_head_new(&h, 1, qr->free_item); */
/*     list_head_push_vpointer(h, data, NULL); */
/*     skip_node_insert(skip_node_prev(new_node),h,skip_list_random_level(qr->docs), */
/* 		     qr->docs->nil); */
/*     qr->docs->num_nodes++; */
/*   } */
/*   /\*if (qr->psize && (qr->psize<<2) < qr->docs->num_nodes) { */
/*     list_head_free((list_head *)skip_list_remove_last(qr->docs), 1); */
/*     }*\/ */
/* } */

/* void query_result_pop_dup(query_result *qr) */
/* { */
/*   if (qr->psize) { */
/*     list_head_free((list_head *)skip_list_remove_first(qr->docs), 1); */
/*   } */
/* } */

void query_result_push_no_dup(query_result *qr, void *data)
{
#warning  "en modelo vectorial y no se deben cortar, se debe escribir y luego escoger el cachito necesario."
  skip_list_insert(qr->docs, data);
  /*if (qr->psize && (qr->psize<<2) < qr->docs->num_nodes) {
    qr->free_item(skip_list_remove_last(qr->docs));
    }*/
}

void query_result_pop_no_dup(query_result *qr)
{
  if (qr->psize) {
    qr->free_item(skip_list_remove_first(qr->docs));
  }
}


/* int __query_results_free_dup__(skip_node *node, void *arg) */
/* { */
/*   list_head_free((list_head *)node->data, 1); */
/*   return 1; */
/* } */

int __query_results_free_no_dup__(skip_node *node, void *arg)
{
  ((query_result *)arg)->free_item((void *)node->data);
  return 1;
}

void query_result_free(query_result *qr)
{
/*   skip_list_iterator_forward(qr->docs, qr->dup_handler ? */
/*   __query_results_free_dup__ : __query_results_free_no_dup__, qr); */
  skip_list_iterator_forward(qr->docs, __query_results_free_no_dup__, qr);
  skip_list_free(qr->docs, NULL);
  free(qr);
}
