#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <skip_list.h>
#include <super_malloc.h>

struct skip_list * skip_list_new(float prob, int max_elem,
				 int (*cmp)(void *, void *),
				 void* (*dup)(void *, void *),
				 void* (*alloc)(void*))
{
  int i;
  struct skip_list * new_list;
  struct skip_node * head, * nil;
  double probabilidad = prob;
  int max_level;
  srand48(time(NULL));
  prob = 1.0/prob;
  max_level = (int)ceil(log(max_elem)/log(prob));
  /* se debe inicializar el generador de nmeros aleatorios*/
  head = skip_node_new(HEADER, NULL, max_level);
  nil = skip_node_new(NIL, NULL, max_level);
  nil->backward = head;
  nil->prev_max_level = head;
  for (i = 0; i < max_level; i++) {
    head->forward[i] = nil;
    nil->forward[i] = NULL;
  }
  /*nil->backward->nodes[0] = head;*/
  new_list = struct_malloc(struct skip_list);
  new_list->num_nodes = 0;
  new_list->prob = probabilidad;
  new_list->max_level = max_level;
  new_list->cmp = cmp;
  new_list->head = head;
  new_list->nil = nil;
  new_list->dup = dup;
  new_list->alloc = alloc;
  return new_list;
}

void skip_list_free(struct skip_list *l, void (*f_free)(void *))
{
  struct skip_node * s = skip_node_next(l->head), *n;
  while(s != l->nil) {
    n = s;
    s = skip_node_next(s);
    if (f_free) f_free(n->data);
    skip_node_free(n);
  }
  skip_node_free(l->head);
  skip_node_free(l->nil);
  free(l);
}

struct skip_node * skip_list_find_prev_node(struct skip_list * list,
				  void *data, int (*f_cmp)(void *, void *))
{
  int i, cmp;
  struct skip_node *s = list->head, *tmp;
  for (i = list->head->level - 1; i >= 0; i--) {
    tmp = s->forward[i];
    cmp = (tmp->ntype == NIL) ? -1: f_cmp(data, tmp->data);
    if (cmp == 0) return tmp->backward;
    if (cmp > 0) s = tmp;
    //while (tmp->ntype != NIL && f_cmp(data, tmp->data) < 0) s = tmp;
  }
  return s;
}

void * skip_list_search(struct skip_list * list, void *value)
{
  struct skip_node *s = skip_list_find_prev_node(list,value,list->cmp);
  s = skip_node_next(s);
  return  (s->ntype != NIL && 0 == list->cmp(value, s->data)) ? s->data : NULL;
}

struct skip_node * skip_node_new(enum skip_node_type ntype, void *d, int level)
{
  struct skip_node * new_node = struct_malloc(struct skip_node);
  new_node->ntype = ntype;
  new_node->data = d;
  new_node->level = level;
  new_node->forward = array_malloc(struct skip_node *, level);
  new_node->prev_max_level = new_node->backward = NULL;
  return new_node;
}

void skip_node_free(struct skip_node * n)
{
  free(n->forward);
  free(n);
}

int skip_list_random_level(struct skip_list *list)
{
  int l = 1;
  while (list->prob < drand48() && l < list->max_level) {
    l++;
  }
  //int l = (int)(list->max_level * d + 1);
  return l;
}

/*
  void * skip_list_remove_last (struct skip_list *list) {
  if (list->num_nodes > 0) {
  return skip_list_remove_by_node(list, skip_node_prev(list->nil));
  }
  return NULL;
  }*/

void * skip_list_remove_first(struct skip_list *list)
{
  int i;
  void *d;
  skip_node *first, *head = list->head;
  first = skip_node_next(head);
  if (first->ntype == NIL) return NULL;
  first->forward[0]->backward = first->backward;
  for (i = 0; i < first->level; i++) {
    head->forward[i] = first->forward[i];
  }
  list->num_nodes--;
  skip_node_next(head)->prev_max_level = head;
  d = first->data;
  skip_node_free(first);
  return d;
}

void * skip_list_remove(struct skip_list *list, void *data)
{
  int i, cmp;
  struct skip_node *s = list->head, *tmp;
  struct skip_node *rem = NULL; /*The node to remove*/
  for (i = list->head->level - 1; i >= 0; i--) {
    tmp = s->forward[i];
    cmp = (tmp->ntype == NIL) ? -1 : list->cmp(data, tmp->data);
    if (cmp < 0) {
      continue;
    } else if (cmp > 0) {
      i++;
      s = tmp;
    } else if (cmp == 0) {
      if (rem == NULL) rem = tmp;
      if (rem != tmp) continue;
      s->forward[i] = tmp->forward[i];
      if ((i+1) == s->level) tmp->prev_max_level = s;
      if (i == 0) {
	tmp->backward = s;
	list->num_nodes--;
	data = tmp->data;
	skip_node_free(tmp);
	return data;
      }
    }
  }
  return NULL;
}

/* permite duplicados, si list->dup == NULL, los duplicados 
   se apilan */

#define MacroLinkNodes \
      if (i < level) { \
        if ((i+1) == level) node->prev_max_level = s; \
        if  (tmp->level == (i+1)) tmp->prev_max_level = node; \
        node->forward[i] = tmp; \
	s->forward[i] = node; \
      }

void skip_list_insert(struct skip_list *list, void *data)
{
  int i, level = skip_list_random_level(list), cmp;
  struct skip_node *s = list->head, *tmp, *per;
  struct skip_node *node = skip_node_new(NODE, NULL, level);
/*   fprintf(stderr, "AAAAAQUI! node:%p, level:%d\n",node,level); */
  for (i = list->head->level - 1; i >= 0; i--) {
    tmp = s->forward[i];
/*     fprintf(stderr, "I=%d, s=%p, tmp=%p, ",i,s,tmp); */
    cmp = (tmp->ntype == NIL) ? -1 : list->cmp(data, tmp->data);
/*     fprintf(stderr, "cmp=%d\n",cmp); */
    if (cmp < 0) {
      MacroLinkNodes;
    } else if (cmp > 0) {
      /*if (s->ntype == NIL) goto skip_list_insert_label;*/
      i++;
      s = tmp;
    } else if (cmp == 0) {
      // error : si node tiene menor nivel que tmp y son iguales, los punteros no se actualizan en MacroLinkNodes
      if (list->dup) {
	if (i == 0) {
	  node->data = list->dup(data, tmp->data);
/* 	  fprintf(stderr,"free: %p\n", tmp); */
/* 	  fprintf(stderr,"- level %d\n", tmp->level); */
/* 	  fprintf(stderr,"- ntype %d\n", tmp->ntype); */
	  per = tmp->forward[i];
	  skip_node_free(tmp);
	  list->num_nodes--;
	  tmp = per;
	  //fprintf(stderr, "===========WWWWW>tmp: %p, ndata: '%s', data: %s \n",
	  //  tmp, node->data , (char*)data);
	} else if (tmp->level > level && i < tmp->level) {
          if (tmp->forward[i]->level == (i+1))
	    tmp->forward[i]->prev_max_level = s;
	  s->forward[i] = tmp->forward[i]; 
	  continue;
	} else {
	  tmp = tmp->forward[i];
	}
      }
      MacroLinkNodes;
    }
  }
/*   fprintf(stderr, "===========XXXXXXX>tmp: %p, ndata: '%s', data: %s \n", */
/* 	  tmp, (char*)node->data, (char*)data); */
  if (node->data == NULL) {
    if (list->alloc) {
      node->data = list->alloc(data);
    } else {
      node->data = data;
    }
  }
/*   fprintf(stderr, "===========YYYYYYY>tmp: %p, ndata: '%s', data: %s \n", */
/* 	  tmp, (char*)node->data, (char*)data); */
  /* skip_list_insert_label:*/
  node->backward = s;
  tmp->backward = node;
  /* Limitando espacio */
  list->num_nodes++;
}


void skip_list_iterator_forward(struct skip_list *list,
				int (*visit) (skip_node *, void *), void *arg)
{
  struct skip_node *s = skip_node_next(list->head), *next;
  while (s != list->nil) {
    next = skip_node_next(s);
    visit(s, arg);
    s = next;
  }
}

int skip_list_iterator_new(skip_list *list, skip_list_iterator **itr)
{
  skip_list_iterator *i = struct_malloc(skip_list_iterator);
  i->list = list;
  i->node = list->head;
  i->position = 0;
  *itr = i;
  return 0;
}

void skip_list_iterator_close(skip_list_iterator *i)
{
  free(i);
}

int skip_list_iterator_has_next(skip_list_iterator *i)
{
  return skip_node_next(i->node)->ntype == NODE;
}

void *skip_list_iterator_next(skip_list_iterator *i)
{
  i->node = skip_node_next(i->node);
  i->position++;
  return i->node->data;
}

void skip_list_iterator_backward(struct skip_list *list,
				int (*visit) (skip_node *, void *), void *arg)
{
  struct skip_node *s = skip_node_prev(list->nil), *prev;
  while (s != list->head) {
    prev = skip_node_prev(s);
    visit(s, arg);
    s = prev;
  }
}


/*To probe SkipLists*/
#include <string.h>

int printfunctionforward(skip_node *n, void *d)
{
  fprintf(stderr, "'%s' -> ", (char *)n->data);
  if (n->forward[0]->ntype == NIL) fprintf(stderr, "(NIL)\n");
  return 1;
}

int printfunctionbackward(skip_node *n, void *d)
{
  fprintf(stderr, "'%s' -> ", (char *)n->data);
  if (n->forward[0]->ntype == HEADER) fprintf(stderr, "(NIL)\n");
  return 1;
}

int main(int argc, char **argv)
{
  skip_list *list = skip_list_new(0.5, 8, strcmp, NULL, NULL);
  fprintf(stderr, "PALABRA HOLA!\n");
  skip_list_insert(list, "hola");
  skip_list_iterator_forward(list, printfunctionforward, NULL);
  fprintf(stderr, "PALABRA ZIRA!\n");
  skip_list_insert(list, "zira");
  skip_list_insert(list, "zira");
  skip_list_iterator_forward(list, printfunctionforward, NULL);
  fprintf(stderr, "PALABRA TE!\n");
  skip_list_insert(list, "te");
  skip_list_insert(list, "xe");
  skip_list_insert(list, "mte");
  skip_list_insert(list, "ye");
  skip_list_iterator_forward(list, printfunctionforward, NULL);
  skip_list_insert(list, "b");
  skip_list_insert(list, "zzzzzzzzzzzzzxx");
  skip_list_iterator_forward(list, printfunctionforward, NULL);
  skip_list_insert(list, "cantera");
  fprintf(stderr, "------->FORWARD!!!!!\n");
  skip_list_iterator_forward(list, printfunctionforward, NULL);
  fprintf(stderr, "AL REVS\n");
  skip_list_iterator_backward(list, printfunctionbackward, NULL);
  fprintf(stderr,"===============================\n");
  fprintf(stderr, "BORRADO A: %s\nXXXXXXXX> AL DEREC\n",(char *)skip_list_remove_first(list));
  fprintf(stderr, "BORRADO B: %s\nXXXXXXXX> AL DEREC\n",(char *)skip_list_remove_first(list));
  skip_list_iterator_forward(list, printfunctionforward, NULL);
  fprintf(stderr, "XXXXXXXXX> AL REVS\n");
  skip_list_iterator_backward(list, printfunctionbackward, NULL);
  fprintf(stderr, "HASTA AQU TODO BIEN, AHORA BORRAR POR DATO\n\nXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\nXXXXXXXXXXXXXXXXXXXXXXXXX\n");
  skip_list_iterator_forward(list, printfunctionforward, NULL);
  fprintf(stderr, "remove ye, res: '%s'\n",(char *)skip_list_remove(list, "ye"));
  skip_list_iterator_forward(list, printfunctionforward, NULL);
  return 0;
}

/* Visita hacia adelante los elementos de 'list'  con la funcin 'visit',
   visit debe regresar 1 a menos que se desee finalizar la iteracin, arg es
   el segundo argumento de 'visit' */
