#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <cache1.h>
#include <trie_node.h>
#include <text_trie.h>
#include <string.h>
#include <idx_man.h>
#include <alist.h>
#include <super_malloc.h>
#include <math.h>
#include <debug.h>
#include <unistd.h>
#include <pthread.h>
#include <skip_list.h>

int text_trie_open_master(idx_man **idxm, const char *file,
			  int mode, int max_pages, int alpha_size, int mlevel)
{
  cache1 *cache;
  trie_node *root;
  index_t iroot;
  alist *atrie;
  switch(cache1_open(&cache, file, mode, max_pages/3, 2*max_pages/3, 0)) {
  case CACHE1_CREATED:
    alist_new(&atrie, cache, SIZEOF_TRIENODE(alpha_size));
    alist_get_new_element(atrie, (void **)&root, &iroot);
    trie_node_init_new(root, alpha_size);
    if (!INDEX_EQ(iroot,cache->header.first_index)) {
      fprintf(stderr,"text_trie_open_master iroot != first_element\n");
      exit(-3);    }
    text_trie_open(idxm, atrie, alpha_size, &iroot, mlevel);
    alist_put_element(atrie, &iroot, 1);
    break;
  case CACHE1_OPENED:
    alist_open(&atrie, &cache->header.first_index, cache);
    text_trie_open(idxm, atrie, alpha_size, &cache->header.first_index,mlevel);
    break;
  case CACHE1_FAIL_OPEN:
  default:
    fprintf(stderr, "text_trie_open_master: fallo al abrir archivo '%s'",file);
    exit(-1);
  }
  return 0;
}

int text_trie_open(idx_man **idxm, alist *atrie, int alpha_size, index_t *root,
		   int mutex_level)
{
  int i,j;
  text_trie *t = struct_malloc(text_trie);
  idx_man *idm = (idx_man *)t;
  /* Inicializando interfaz */
  idm->idxtype = IDX_TRIE;
  idm->close = text_trie_close;

  idm->insert = text_trie_insert_document;
  idm->remove = text_trie_remove_document;
  idm->find = text_trie_exists_word;
  
  t->atrie = atrie;
  t->node_size = alpha_size;
  t->init_node = trie_node_new(alpha_size);
  t->iroot = *root;
  t->mutex_level = mutex_level;
  t->alpha_size = alpha_size;

  if (mutex_level >= 0) {
    t->mutex_table = (pthread_mutex_t **)
      super_malloc(sizeof(pthread_mutex_t*) * mutex_level);
    for (i = 0; i < mutex_level; i++) {
      t->mutex_table[i] = (pthread_mutex_t *)
	super_malloc(sizeof(pthread_mutex_t) * alpha_size);
      for (j = 0; j < alpha_size; j++) {
	pthread_mutex_init(&t->mutex_table[i][j], NULL);
      }
    }
  }
  /* Crear los primeros n palabras ms posibles */
  *idxm = idm;
  return 0;
}

int text_trie_remove_document(idx_man *idm, id_wrd_t *wrd, id_doc_t *doc)
{
  return 0;
}

void text_trie_set_caches(idx_man* idm, cache1 *idx, cache1 *dat)
{
  text_trie *trie = (text_trie *) idm;
  cache1 **arg = (cache1 **)super_malloc(2 * sizeof(cache1 *));
  arg[0] = idx;
  arg[1] = dat;
  trie->arg = arg;
  trie->add_info = text_trie_add_doc;
  
  /* unlink("db.idx-btree");
     trie->db_index = gdbm_open("db.idx-btree", 0, GDBM_WRCREAT, 0644, NULL);
     if (trie->db_index == NULL) {
     perror("text_trie_set_caches gdbm_open db.idx-btree");
     exit(2);
     } */
}

void text_trie_close_caches(idx_man* idm)
{
  text_trie *trie = (text_trie *) idm;
  cache1_close(((cache1 **)trie->arg)[0]);
  cache1_close(((cache1 **)trie->arg)[1]);
  /*gdbm_close(trie->db_index);*/
  free(trie->arg);
}

int text_trie_close(idx_man* idm)
{
  text_trie *trie = (text_trie *) idm;
  cache1 *c;
  int i, j;
  if (trie->mutex_level > 0) {
    for (i = 0; i < trie->mutex_level; i++) {
      for (j = 0; j < trie->alpha_size; j++) {
	pthread_mutex_destroy(&trie->mutex_table[i][j]);
      }
      free(trie->mutex_table[i]);
    }
    free(trie->mutex_table);
  }
  trie_node_free(trie->init_node);
  c = trie->atrie->cache;
  alist_close(trie->atrie);
  cache1_close(c);
  free(trie);
  return 0;
}

#ifdef TEXT_TRIE_WITH_DATLIST

int text_trie_add_doc(text_trie *trie, trie_node *node, id_wrd_t *wrd,
		      id_doc_t *doc, int index_pos, u_int8_t is_final)
{
  index_t tmp;
  idx_global_wrd_data global_wrd, *global_wrd_ref;
  idx_wrd_data *wrd_data;
  int frec;
  /*datum dat, key;
    struct __db_idx dbx; */
  idx_data *data;
  cache1 *idx_info = ((cache1 **)trie->arg)[0];
  cache1 *dat_info = ((cache1 **)trie->arg)[1];
  alist *idx_list, *dat_list;

  if ((is_final) ? TNODE_FINAL_IS_NIL(node) : TNODE_IS_NIL(node, index_pos)) {
    alist_new(&idx_list, idx_info, sizeof(idx_data));
    alist_new(&dat_list, dat_info, sizeof(idx_wrd_data));
    global_wrd.global_frec = 1;
    global_wrd.number_docs = 1;
    alist_insert_new(idx_list, &global_wrd, &tmp, sizeof(global_wrd));
    alist_get_new_element(idx_list, (void **)&data, &tmp);
    data->freq = 1;
    data->index = dat_list->ihead;
    data->id_doc = *doc;
    if (is_final) {
      trie_node_set_at_final_field(node, &idx_list->ihead);
    } else {
      trie_node_set_word_end(node, index_pos, &idx_list->ihead);
    }
 } else {
    if (is_final) {
      tmp = *trie_node_get_final(node);
    } else {
      tmp = *trie_node_get_symbol_by_index(node, index_pos);
    }
    alist_open(&idx_list, &tmp, idx_info);
    /*key.dptr = &dbx;
      dbx.doc =*doc;
      dbx.wrd = tmp;
      key.dsize= sizeof(dbx);
      dat = gdbm_fetch(trie->db_index, key);
      if (dat.dptr != NULL) {
      tmp = *((index_t *)dat.dptr);
      free(dat.dptr);*/

    /* El primer elemento de la lista es el contador de la palabra para
       todos los documentos */
    global_wrd_ref = (idx_global_wrd_data *)ALIST_DATA_POINTER(idx_list->head);
    global_wrd_ref->global_frec++;
    if (alist_find(idx_list, doc, idx_man_doc_find_cmp, &tmp)) {
      alist_get_by_index(idx_list, (void **)&data, &tmp);
      data->freq++; /* frecuencia palabra en el documento*/
      alist_open(&dat_list, &data->index, dat_info);
    } else {
      global_wrd_ref->number_docs++;
      alist_new(&dat_list, dat_info, sizeof(idx_data));
      alist_get_new_element(idx_list, (void **)&data, &tmp);
      data->freq = 1;
      data->index = dat_list->ihead;
      data->id_doc = *doc;
      /*dat.dptr = &tmp;
	dat.dsize= sizeof(tmp);
	gdbm_store(trie->db_index, key, dat, GDBM_REPLACE);*/
    }
  }
  alist_get_new_element(dat_list, (void **)&wrd_data, &tmp);
  wrd_data->doc_position = wrd->pos;
  wrd_data->tag = wrd->tag;
  frec = data->freq;
  alist_put_element(dat_list, &tmp, 1);
  alist_close(dat_list);
  alist_close(idx_list);
  return frec;
}

#else /* without TEXT_TRIE_WITH_DATLIST */

int text_trie_add_doc(text_trie *trie, trie_node *node,
		      id_wrd_t *wrd, id_doc_t *doc,
		      int index_pos, u_int8_t is_final)
{
  index_t tmp;
  idx_global_wrd_data global_wrd, *global_wrd_ref;
  idx_wrd_data *wrd_data;
  int frec;
  idx_data *data;
  cache1 *idx_info = ((cache1 **)trie->arg)[0];
  alist *idx_list;
  if ((is_final) ? TNODE_FINAL_IS_NIL(node) : TNODE_IS_NIL(node, index_pos)) {
    alist_new(&idx_list, idx_info, sizeof(idx_data));
    global_wrd.global_freq = wrd->freq;
    global_wrd.number_docs = 1;
    alist_insert_new(idx_list, &global_wrd, &tmp, sizeof(global_wrd));
    alist_get_new_element(idx_list, (void **)&data, &tmp);
    data->freq = wrd->freq;
    data->id_doc = *doc;
    if (is_final) {
      trie_node_set_at_final_field(node, &idx_list->ihead);
    } else {
      trie_node_set_word_end(node, index_pos, &idx_list->ihead);
    }
 } else {
    if (is_final) {
      tmp = *trie_node_get_final(node);
    } else {
      tmp = *trie_node_get_symbol_by_index(node, index_pos);
    }
    alist_open(&idx_list, &tmp, idx_info);
    /* El primer elemento de la lista es el contador de la palabra para
       todos los documentos */
    global_wrd_ref = (idx_global_wrd_data *)ALIST_DATA_POINTER(idx_list->head);
    global_wrd_ref->global_freq+=wrd->freq;
    /* works because we use backward find, this could result in erroneus
       results if global_wrd_ref matchs with the cmp */
    if (alist_find(idx_list, doc, idx_man_doc_find_cmp, &tmp)) {
      alist_get_by_index(idx_list, (void **)&data, &tmp);
      data->freq += wrd->freq; /* frecuencia palabra en el documento*/
    } else {
      global_wrd_ref->number_docs++;
      alist_get_new_element(idx_list, (void **)&data, &tmp);
      data->freq = wrd->freq;
      data->id_doc = *doc;
    }
  }
  frec = data->freq;
  alist_put_element(idx_list, &tmp, 1);
  /* alist_get_new_element(dat_list, (void **)&wrd_data, &tmp);
     wrd_data->doc_position = wrd->pos;
     wrd_data->tag = wrd->tag;
     alist_put_element(dat_list, &tmp, 1);
     alist_close(dat_list);*/
  alist_close(idx_list);
  return frec;
}

#endif /* TEXT_TRIE_WITH_DATLIST*/

#define text_trie_lock(_trie, _level, _key) \
    if (_level < (_trie)->mutex_level) \
        pthread_mutex_lock(&(_trie)->mutex_table[_level][_key]); 

#define text_trie_unlock(_trie, _level, _key) \
    if (_level < (_trie)->mutex_level) \
        pthread_mutex_unlock(&(_trie)->mutex_table[_level][_key]); 

#define TRIE_DEBUG(string) fprintf(stderr,string " trie: %p, idm: %p, trie->atrie: %p, itr->list: %p\n", trie, idm, trie->atrie, ((alist_iterator *) itr)->list); memset(NULL, 1, 1024); exit(-200);

int text_trie_insert_document(idx_man *idm, id_wrd_t *wrd_id, id_doc_t *doc_id)
{
  iterator *itr;
  int i, /* contador */
    length, /* tamao de la palabra (se decrementa) */
    index_pos, /* Posicin en nodo trie del caracter de cadena */
    frec=0; /*frecuencia de palabra en documento*/
  index_t index_node_new, index_tmp, index_tmp_tmp;
  char index_final_modif;
  trie_node *node, *tmp;
  text_trie *trie = (text_trie *) idm;
  alist_iterator_new(trie->atrie, &itr, 0);
  node = (trie_node *) itr->jump(itr, &trie->iroot);
  if (WRD_IS_KEYWORD(wrd_id)) {
    if (TNODE_FINAL_IS_NIL(node)) {
      alist_get_new_element(trie->atrie, (void **)&tmp, &index_node_new);
      trie_node_init(tmp, trie->init_node, trie->node_size);
      trie_node_set_at_final_field(node, &index_node_new);
      alist_put_element(trie->atrie, &index_node_new, 1);
      node = (trie_node *) itr->jump(itr, &index_node_new);
    } else {
      node = (trie_node *) itr->jump(itr, trie_node_get_final(node));
    }
    i = 1;
  } else {
    i = 0;
  }
  length = wrd_id->len;
  length--;
  /*fprintf(stderr,"\n\n\n =====***** PALABRA LONGITUD '%d'\n", wrd_id->len);
    fprintf(stderr, "ROOT MEM: %p\n", node);*/
  for (; i < length; i++) {
    index_pos = wrd_id->key[i];
    index_final_modif = 0;
    text_trie_lock(trie, i, index_pos);
    switch (trie_node_get_kind_by_index(node,index_pos)) {
    case TNODE_IDX_VALUE:
      /* Siguiendo el indice */
      index_tmp = *trie_node_get_symbol_by_index(node, index_pos);
      node = (trie_node *)itr->jump(itr,&index_tmp);
      break;
    case TNODE_END_VALUE:
      /* Haba una palabra hasta aqui, hay que recorrerla */
      index_final_modif = 1;
    default:
      if (index_final_modif == 0) {
	debug_index(&index_tmp, "==============================\nindex_tmp:");
	debug_index(&index_node_new, "index_node_new:");
	debug_index(&trie->atrie->cache->header.last_position, "last:");
	fprintf(stderr,"text_trie_insert_word:1 desconocido i:%d, tipo: %d, index_pos: %d\n",
		i,trie_node_get_kind_by_index(node,index_pos), index_pos);
#ifdef _DEVELOP_VERSION
	trie_node_inspect(node);
	TRIE_DEBUG("Error text_trie_insert_word:1");
#else
	/* Se debe verificar si an estamos en memoria reservada */
	TNODE_SET_NIL(node, index_pos);
	//fprintf(stderr, "Aun no funciona, primero arreglarlo");  exit(-2);
#endif
      }
    case TNODE_NIL_VALUE:
      /* hay que insertar */
      alist_get_new_element(trie->atrie, (void **)&tmp, &index_node_new);
      trie_node_init(tmp, trie->init_node, trie->node_size);
      if (index_final_modif) {
	trie_node_set_at_final_field
	  (tmp, trie_node_get_symbol_by_index(node,index_pos));
      }
      trie_node_set_index(node, index_pos, &index_node_new);
      ALIST_ITERATOR_SET_MODIFIED(itr);
      node = (trie_node *)itr->jump(itr, &index_node_new);
      /*returning from the alist_get_new_element call*/
      alist_put_element(trie->atrie, &index_node_new, 1); 
      break;
    }
    text_trie_unlock(trie,i,index_pos);
  }

  /* index_pos = alfa_get_index_by_char(wrd_id->key[length]); */
  index_pos = wrd_id->key[length];
#ifdef _DEVELOP_VERSION
  if (index_pos < 0) {
    fprintf(stderr,"index_pos s=%d u=%d l=%d\n", index_pos,
	    (u_int8_t)index_pos, length);
    TRIE_DEBUG("Error IndexpPos < 0");
  }
  if (length < 0) { 
    fprintf(stderr, "length: %d\n", length);
    TRIE_DEBUG("Error Length < 0");
  }
#endif

  text_trie_lock(trie, length, index_pos);
  index_final_modif = 0; /*usado como bandera en el siguiente bloque */
  switch (trie_node_get_kind_by_index(node,index_pos)) {
  case TNODE_END_VALUE:
    index_final_modif = 1;
    /*aade documento al ndice apuntado...*/
  default:
    if (index_final_modif == 0) {
      fprintf(stderr,"text_trie_insert_word:<2> desconocido i:%d, tipo: %d\n",
	      i,trie_node_get_kind_by_index(node,index_pos));
#ifdef _DEVELOP_VERSION_XXX
      trie_node_inspect(node);
      TRIE_DEBUG("Error text_trie_insert_word:<2>");
#else
      TNODE_SET_NIL(node, index_pos);
      /*fprintf(stderr, "Aun no funciona, primero arreglarlo");  exit(-2);*/
#endif
    }
  case TNODE_NIL_VALUE:
    frec = trie->add_info(trie, node, wrd_id, doc_id, index_pos, 0);
    ALIST_ITERATOR_SET_MODIFIED(itr);
    break;
  case TNODE_IDX_VALUE:
    if (trie_node_get_symbol_by_index(node,index_pos)->page > trie->atrie->cache->header.last_position.page) {
#warning "KILL MEM"
      memset(NULL,NULL,12230);
    }
    itr->get_current_index(itr, &index_tmp_tmp);
    index_tmp = *trie_node_get_symbol_by_index(node,index_pos);
    node = (trie_node *)itr->jump(itr, &index_tmp);
    ALIST_ITERATOR_SET_MODIFIED(itr);
    frec = trie->add_info(trie, node, wrd_id, doc_id, 0, 1);
    break;
  }
  text_trie_unlock(trie, length, index_pos);
  /*   alist_put_element(list, &index_node); /\*close C2*\/ */
  itr->close(itr);
  return frec;
}

int text_trie_extract_thesaurus(idx_man *idm, char ***thesaurus, int *size,
				char *prefix, int prefixlen, int maxlen,
				index_t *init_index)
{
  int index_pos; /* contador */
  trie_node *node;
  text_trie * trie = (text_trie *) idm;
  if (init_index == NULL) init_index = &trie->iroot;
  alist_get_by_index(trie->atrie, (void **)&node, init_index);
  if (trie_node_get_final(node)->flags == TNODE_END_VALUE) {
    (*size)++;
    /* fprintf(stderr, "---->");
    fwrite(prefix, sizeof(char), prefixlen, stderr);
    fprintf(stderr, "\n"); */
  }
  for (index_pos = 0; index_pos < trie->node_size; index_pos++) {
    switch (trie_node_get_kind_by_index(node,index_pos)) {
    case TNODE_IDX_VALUE:
      prefix[prefixlen] = alfaweb_get_char_by_index((unsigned int)index_pos);
      text_trie_extract_thesaurus(idm, NULL, size, prefix, prefixlen + 1,
		    maxlen, trie_node_get_symbol_by_index(node, index_pos));
      continue;
    case TNODE_END_VALUE:
      (*size)++;
#warning "PONER SALIDA A ALGUN LUGAR, SOLO SE CALCULA"
      /* fprintf(stderr, "oooo %d, %d>", prefixlen, maxlen);
	 fwrite(prefix, sizeof(char), prefixlen, stderr);
	 fprintf(stderr, "%c\n", alfaweb_get_char_by_index((unsigned int)index_pos)); */
      goto text_trie_extract_thesaurus_label;
    }
  }
 text_trie_extract_thesaurus_label:
  alist_put_element(trie->atrie, init_index, 1);
  return 0;
}

int text_trie_exists_word(idx_man *idm, id_wrd_t *wrd_id, iterator **res_itr)
{
  iterator *itr;
  text_trie *trie = (text_trie *) idm;
  u_int32_t i; /* contador */
  int ret, /* return value */
    index_pos; /* Posicin en nodo trie del carcter de cadena */
  index_t index_tmp;
  cache1 *idx_info = ((cache1 **)trie->arg)[0];
  itr = NULL;
  alist *list;
  trie_node *node;
  alist_iterator_new(trie->atrie, &itr, 0);
  ret = 0;
  node = (trie_node *)itr->jump(itr, &trie->iroot);
  if (WRD_IS_KEYWORD(wrd_id)) {
    if (TNODE_FINAL_IS_NIL(node)) goto text_trie_exists_word_label;
    node = (trie_node *) itr->jump(itr, trie_node_get_final(node));
    i = 1;
  } else {
    i = 0;
  }
  for (; i < wrd_id->len; i++) {
    /*index_pos = alfa_get_index_by_char(wrd_id->key[i]);*/
    index_pos = wrd_id->key[i];
    if (index_pos == -1) {
      fprintf(stderr,"Caracter '%d' no esta en el alfabeto\n", wrd_id->key[i]);
      return 0;
    }
    switch (trie_node_get_kind_by_index(node,index_pos)) {
    case TNODE_IDX_VALUE:
      /* Siguiendo el indice */
      index_tmp = *trie_node_get_symbol_by_index(node,index_pos);
      node = (trie_node *)itr->jump(itr, &index_tmp);
      if (i+1 == wrd_id->len && !TNODE_FINAL_IS_NIL(node)) {
	ret = 1;
	/*est mal!!! se necesita tener acceso lineal a los resultados, una
	  lista de los nodos head de un resultado para cada documento sera
	  una solucin simple al problema */
	alist_open(&list, trie_node_get_final(node), idx_info);
	alist_iterator_new(list, res_itr, 1);
	goto text_trie_exists_word_label;
      }
      break;
    case TNODE_END_VALUE:
      if (i+1 == wrd_id->len) {
	ret = 1;
	alist_open(&list, trie_node_get_symbol_by_index(node,index_pos),
		   idx_info);
	alist_iterator_new(list, res_itr, 1);
      }
      goto text_trie_exists_word_label;
    default:
      fprintf(stderr,"text_trie_exists_word:1 no conocido i:%d, tipo: %d\n",
	      i,trie_node_get_kind_by_index(node,index_pos));
#ifdef _DEVELOP_VERSION
      trie_node_inspect(node);
      TRIE_DEBUG("Error text_trie_exists_word:1");

#endif
    case TNODE_NIL_VALUE:
      ret = 0;
      goto text_trie_exists_word_label;
    }
  }
  /* buscando en campo final */
 text_trie_exists_word_label:
  /*   alist_put_element(list_cache, &index_node); /\*close C2*\/ */
  itr->close(itr);
  return ret;
}

