#ifndef _H_TEXT_TRIE__
#define _H_TEXT_TRIE__

#include <super_malloc.h>
#include <trie_node.h>
#include <idx_man.h>
#include <cache1.h>
#include <alist.h>
#include <db.h>


/*
  Faltara ver como hacer updates y perder espacio en disco, siempre
  crece. En cada registro a donde apunte el trie puede guardarse
  la llave con que se llego hasta ah. Y as se pueden lograr resoluciones
  inversas (al menos la mayora que necesitaremos).
*/

/* Estructura de trie */
struct text_trie {
  /* Interfaz idm_man. Debe ir al inicio para hacer cast */
  idx_man idm;
  pthread_mutex_t **mutex_table;
  int8_t mutex_level;
  u_int8_t alpha_size;
  /* Indices de la raz y la cabecera */
  index_t iroot;

  
  /* Caracteres en el trie_node */
  u_int16_t node_size;
  /* Nodo para iniciar todos los dems */
  trie_node *init_node;

  
  /* trie y lista */
  alist *atrie;
  
  /* Manejador para cuando se inserta informacin, por default
     text_trie_add_doc. Se debe manejar los casos TNODE_FINAL_IS_NIL y
     TNODE_IS_NIL. Regresa la frecuencia de la palabra en el documento
     en text_trie_add_doc.
     |trie| Es el trie en el que se est trabajando
     |node| Nodo sobre el cual se est insertando
     |word| Palabra insertada
     |docu| Documento de donde se saco est palabra
     |index_pos| Posicin en el nodo en el cual termino la palabra y se quiere
                 poner informacin
     |is_final| Verdadero si se inserto sobre la parte final del trie_node
  */
  int (*add_info)(struct text_trie *trie, trie_node *node, id_wrd_t *word,
		   id_doc_t *docu, int index_pos, u_int8_t is_final);
  /* Un argumento arbitrario (puede ser usado en add_info) */
  void *arg;
  /* Indice para insercin: muy lento */
  
  /*  GDBM_FILE db_index;*/
};

typedef struct text_trie text_trie;

/* Inserta un documento en la lista invertida de wrd_id */
int text_trie_insert_document(idx_man* idm, id_wrd_t *wrd_id,id_doc_t *doc_id);
/* Borra un documento de la lista invertida de wrd_id */
int text_trie_remove_document(idx_man* idm, id_wrd_t *wrd_id,id_doc_t *doc_id);
/* Pregunta si existe una palabra y crea un iterador (FALTA AQUI)*/
int text_trie_exists_word(idx_man* idm, id_wrd_t *wrd, iterator **itr);
/* Cierra el text_trie */
int text_trie_close(idx_man* idm);
/* Abre un trie, crea las caches */
int text_trie_open_master(idx_man **idxm, const char *file,
			  int mode, int max_pages, int alpha_size, int mutex_level);
/* Crea un text_trie con las caches trie y list, pueden ser la misma.
   El trie tiene la raiz en el indice root y el alfabeto es de |alfa_size|
   elementos */
int text_trie_open(idx_man **idxm, alist *atrie, int alfa_size, index_t *root, int mutex_level);

/* Pone las caches de indice y datos. Adems pone add_info<-text_trie_add_doc*/
void text_trie_set_caches(idx_man* idm, cache1 *idx, cache1 *dat);

/* Cierra las caches */
void text_trie_close_caches(idx_man* idm);

/* Aade un documento (manejador) */
int text_trie_add_doc(text_trie *trie, trie_node *node, id_wrd_t *wrd,
		       id_doc_t *doc, int index_pos, u_int8_t is_final);

/* Extrae el vocabulario (Por el momento solo imprime)*/
int text_trie_extract_thesaurus(idx_man *idm, char ***thesaurus, int *size,
				char *preffix, int preffixlen, int maxlen,
				index_t *init_index);
#endif
