#ifndef _H_TRIE_NODE_
#define _H_TRIE_NODE_

#include <alfabeto.h>
#include <super_malloc.h>
#include <cache1.h>
#include <id_word_t.h>

struct trie_node {
  index_t final;
  index_t alfabeto[1]; /* 1 para que nos permita hacer ref por indice */
} __attribute__ ((__packed__));

/*index_t alfabeto[LONGITUD_ALFABETO];*/

typedef struct trie_node trie_node;

#define SIZEOF_TRIENODE(_size) (sizeof(index_t)*(1+_size))

/* Accede a la posicin _index del alfabeto */
/*index_t* trie_node_get_symbol_by_index(trie_node *n, int index);*/
#define trie_node_get_symbol_by_index(trie_node_x_n, int_x_index) \
  (&((trie_node_x_n)->alfabeto[int_x_index]))

/* Accede para escritura en la posicin _index del alfabeto */
/*void trie_node_set_symbol_by_index(trie_node *n, int index, index_t* value);*/
#define trie_node_set_symbol_by_index(_n, _index, p_value) \
  ((_n)->alfabeto[_index] = *(p_value))

/* Devuelve el tipo de Nodo por valor */
/*int trie_node_get_kind_by_index(trie_node *n, int index);*/
#define trie_node_get_kind_by_index(_n, _index) \
  (TNODE_GET_FLAG(trie_node_get_symbol_by_index(_n,_index)))

/* Pone en la posicin _index del trie "_trie" un nodo TNODE_INDEX
   el indice _value, se ponen las flags necesarias para un nodo indice,
   es decir se modifica value */
int trie_node_set_index(trie_node *n, int index, index_t *value);

/* Se usan 4 bytes como banderas o estados, los 12 bits restantes
   pueden ser utilizados para otros fines */

#define TNODE_NIL_VALUE 1
#define TNODE_IDX_VALUE 2
#define TNODE_END_VALUE 3

#define TNODE_CMP_FLAGS(_field,_flag) (_field == _flag)

#define TNODE_FINAL_IS_NIL(_node) \
 TNODE_CMP_FLAGS((_node)->final.flags, TNODE_NIL_VALUE)

#define TNODE_CMP_FIELD(_node, _index, _flag) \
 TNODE_CMP_FLAGS(trie_node_get_symbol_by_index(_node, _index)->flags, _flag)

#define TNODE_IS_NIL(_node, _index) \
 TNODE_CMP_FIELD(_node,_index,TNODE_NIL_VALUE)

#define TNODE_IS_IDX(_node, _index) \
 TNODE_CMP_FIELD(_node,_index,TNODE_IDX_VALUE)

#define TNODE_IS_END(_node, _index) \
 TNODE_CMP_FIELD(_node,_index,TNODE_END_VALUE)

#define TNODE_SET_FLAG(_index,_edo) ((_index)->flags = _edo)

#define TNODE_GET_FLAG(_index) ((_index)->flags)

#define TNODE_SET_FLAG_FIELD(_node,_index,_flag) \
 TNODE_SET_FLAG(trie_node_get_symbol_by_index(_node, _index), _flag)

#define TNODE_SET_NIL_FINAL(_node) \
 TNODE_SET_FLAG(&(_node)->final, TNODE_NIL_VALUE)

#define TNODE_SET_IDX_FINAL(_node) \
 TNODE_SET_FLAG((_node)->final, TNODE_IDX_VALUE)

#define TNODE_SET_NIL_BY_ADDR(_node) \
 TNODE_SET_FLAG(_node, TNODE_NIL_VALUE)

#define TNODE_SET_NIL(_node, _index) \
 TNODE_SET_FLAG_FIELD(_node, _index, TNODE_NIL_VALUE)

#define TNODE_SET_IDX(_node, _index) \
 TNODE_SET_FLAG_FIELD(_node, _index, TNODE_IDX_VALUE)

#define TNODE_SET_END(_node, _index) \
 TNODE_SET_FLAG_FIELD(_node, _index, TNODE_END_VALUE)

/* Pone el index_t |index| como palabra final en |n| en la posicin |char_pos|
   modifica |index| con las flags necesarias para un nodo TNODE_END_VALUE */
void trie_node_set_word_end(trie_node *n, int char_pos, index_t *index);
/* Obtiene el campo para indicar fin de palabra de |n| */
/*index_t *trie_node_get_final(trie_node *n);*/
#define trie_node_get_final(_n) (&((_n)->final))

/* Pone el campo final de |n| a |index|*/
/*void trie_node_set_at_final_field(trie_node *n, const index_t *index);*/
#define trie_node_set_at_final_field(_n, _index) (_n)->final = *(_index)


/* Muestra una representacin de |n|, util para debug*/
void trie_node_inspect(trie_node *n);
/* Crea un nuevo nodo, no se usa con caches pero crea un nodo inicializado
   para luego copiarlo en trie_node_init */
trie_node * trie_node_new(int size);
/* Libera un nodo, no se usa con caches */
void trie_node_free(trie_node *node);
/* Inicializa los valores de |n| copiando |init_node| de |size| elementos*/
trie_node * trie_node_init(trie_node *n, const trie_node *init_node, int size);
/* Igual pero no copia un nodo, lo inicializa (lento) */
void trie_node_init_new(trie_node *n, int size);

#endif
