#coding: utf-8;
import warn
import sre

def pager(length, begin, count, create, max, msg_next, msg_back):
    """
    length: Number of documents (total number)
    begin: First document
    count: Number of documents in the page
    create: A function to ceate hyperlinks to the pager links.
     It recieves the list [begin, count, name]  where name:
     begin is the begin for the pointed page, count is the number
     of pages to be showned and name the possible name
    max: number of pages to show
    msg_next: The message to print the next page shortcut
    msg_back: The message to print the previous page
    """

    if length < count: return ()
    idx = 0
    pag = count * max # tamaño de paginacion
    # pagina de posicion actual (0 se promueve a 1)
    pos = int(begin / count) + 1
    # La primera pagina que se muestra en la lista de paginas, la posicion
    # actual siempre trata de estar en medio
    ini = int(((begin  - pag/2)) / count) * count
    if ini < 0: ini = 0
    res = []
    if pos > 1: res.append(create(((pos-2)*count,count,msg_back)))
    while idx < max:
        beg = ini+idx*count
        if beg >= length: break
        idx+=1
        _pos = int(beg / count) + 1
        if pos == _pos:
            res.append(str(pos))
        else:
            res.append(create((beg, count, _pos)))
    if pos*count < length: res.append(create((pos*count,count,msg_next)))
    return res

def parse_query(querystring, harvest, support_related = True):
    urlrelated = ""
    queryvector = []
    keywordvector = []
    for sq in sre.split(r"\s+", querystring):
        if sq.startswith("keyword:"):
            sq = sq[8:]
            keywordvector.append(sq)
        elif sq.find(":") != -1:
            if support_related:
                reltype = 0
                if sq.startswith("related-as-keywords:"):
                    reltype = 1
                    urlrelated = sq[20:]
                if sq.startswith("related:"):
                    reltype = 2
                    urlrelated = sq[8:]
                if reltype:
                    try:
                        s = harvest.open(urlrelated)
                        if s is None: raise "Repository returns None"
                        if reltype == 1:
                            keywordvector.append(s)
                        else:
                            queryvector.append(s)
                    except:
                        warn.debug("URL: '", urlrelated, "' does not exists")
                    continue
            keywordvector.append(sq)
        else:
            queryvector.append(sq)
    return queryvector, keywordvector, urlrelated

def search(query, keywords, begin, count, qtype, index, log = True, decode=None, encode='latin1'):
    """ Returns the tuple (results, approx, terms) where
    |results| are the results of the search
    |approx| is the approximated number of total documents returned by a
    complete search
    |terms| is a list of tuples (term,term's freq)

    The arguments are:
    |query| The query (in a valid format to spydihub or spydex)
    |begin| Returns documents since |begin| document
    |count| Number of document to retrieve
    |qtype| rank function
    |keywords| The keywords (in a valid format to spydihub or spydex)
    |index| The indexer to perform the search
    |log| Do log or not
    |decode| Decode from encoding to |encode| (default supported)
    |encode| Encoding  (default latin1)
    """
    if encode and decode:
        try:
            if isinstance(query, str):
                query = query.decode(decode).encode(encode)
            if isinstance(keywords, str):
                keywords = keywords.decode(decode).encode(encode)
        except:
            warn.warn()
    gdata, results = index.search(query, keywords, begin, count, qtype)
    if log:
        warn.log("QUERY(%s,%s,%s): %s, %s"%(begin, count, qtype,
                                            query, keywords))
    terms = map(lambda x: (x[0],x[1][2]), gdata)
    approx = 2147483647
    for entry in terms:
        if entry[1] > 0: approx = min(approx,entry[1])
    if approx == 2147483647:
        approx = 0
    return results, approx, terms

