#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'Repository.py'                                             #
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

###
# BUGS:
# Berkeley databases are not working in threading mode.
# Specifically index_times
# locks all threads in remove and updateInfo
# We use a workaround very simple but not very efficient
from dbsupport import bsddb3
import time
import urllib2
import os
import thread, threading
from threading import Thread, Condition
import monarca
import monarca.storage
from monarca.storage import Storage
import sre
import traceback
from cStringIO import StringIO
import socket
#import config
from threadutils import ThreadPool, PeriodicExecution, PCExecution
import warn
import shttp
from shttp import uri
from shttp.uri import URI
from shttp.sHTTP import HTTPError, HTTPUnexpectedStatus, HTTPBadResponse
import threading

from gzip import GzipFile
try:
    import bz2
except:
    warn.debug("There is not bzip2 module, unable to use it")

class UnknownURIFormat(Exception): pass

class SiteInformation:
    """ Represents the document on the disk and web. Provides methods to
    retrieve new versions of the document and track the fails in the retrieval
    """
    date_format="%a, %d %b %Y %H:%M:%S GMT"
    is_ip=sre.compile(r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}")
    #expires=604800 # One week after last_fetched: 60*60*24*7

    def get_request(self):
        req = "%s%s%s"%(self.path or '/', self.query or '', self.sharp or '')
        return os.path.normpath(req)

    last_slash = sre.compile(r"/+$")
    def get_url(self,useIP=True):
        if useIP:
            host = self.ip
        else:
            host = self.host
        return uri.normurldict(self.__dict__, host)

    firstfield=sre.compile("\s+")
    splithost=sre.compile("\.")
    def __init__(self, url, ip = None, last_fetched = 0, mime = '',
                 keywords = None, datafilter = None):
        """ Create new SiteInformation. To reuse the object,
        the SiteInformation  is stored into a |shelve|
        The |filter| is the filter that must be applied to recover
        the data from the repository it can be any of the following
        string:
         |gzip| Compress with gzip before saving
         |bzip2| Compress using bzip2
         |mime| to guess
         |None,'',False,0| used to no apply any transformation
         Additionally to any supported keyword in subclasses os specializations
         of the repository """
        self.last_fetched = last_fetched # Never, yet
        self.errors_count = 0 # Clean init
        self.errors_count404 = 0 # Clean init
        self.url = url
        self.ip = ip
        self.time_index = 0
        u=URI.match(url)
        if not u:
            #raise UnknownURIFormat, url
            self.path = url
            self.port = ''
            self.protocol = 'none'
            self.host = ''
            self.query = ''
            self.sharp = ''
        else:
            self.__dict__.update(u.groupdict()) # All stuff and the real host
        self.mime = mime # The real mime
        self.datafilter = datafilter
        mime = SiteInformation.firstfield.split(mime)[0]
        if keywords is None:
            self.keywords = ['access:public']
        else:
            self.keywords = keywords
            if not 'access:public' in self.keywords:
                self.keywords.append('access:public')
        #self.keywords.append('site:%s'%(self.host))
        if not self.is_ip.match(self.host):
            s = SiteInformation.splithost.split(self.host)
            h = s[-1]
            for i in range(len(s) - 2, -1, -1):
                h = ".".join((s[i], h))
                self.keywords.append('site:%s'%(h))
        if mime != '': self.keywords.append('mime:%s'%(mime))

    def _failRetrieve(self, repository, response_status):
        if response_status == 200: return False
        # Manage 2xx,3xx,4xx, and 5xx
        if 400 <= response_status <= 500:
            self.errors_count404 = self.errors_count404 + 1
            if self.errors_count404 >= repository.config.HTTPLib_maxRetries404:
                warn.debug('Too many 404 errors ', self.url, ' ',
                      repository.config.HTTPLib_maxRetries404)
                repository.remove(self.url)
            return True
        elif 500 <= response_status:
            self.errors_count = self.errors_count + 1 # No check, but increment
            return True
        else:
            return False

    def needRetrieve(self,repository,last_modified,date_now,response_status):
        """ Verify if we need retrieve the document based on headers info and 
        localtime """
        if self._failRetrieve(repository, response_status): return False
        #warn.debug('needRetrive status ', headers.status, ' on ', self.url)
        time_range=0
        if last_modified and date_now:
            time_now = time.time()
            # Expires works? should I believe it?
            # We use a static expires date, like week. But the refresh can be
            # in demand (much better)
            last_modified = time.strptime(last_modified,self.date_format)
            date_now = time.strptime(date_now,self.date_format)
            time_range = time.mktime(date_now) - time.mktime(last_modified)
        return self.last_fetched < (time.time() - time_range)

    removesharp = sre.compile(r"\#[^\#]+$")
    def httprequest(self, _type, location, host, port):
        if port == '':
            port = 80
            strport = ""
        else:
            port = int(port)
            strport = ":%s"%(port)
        headers = {'Host' : host + strport,
#                   'User-Agent' : config.HTTPUserAgent,
#                   'Accept' : 'text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,text/plain;q=0.8,image/png,*/*;q=0.5',
#                   'Accept-Encoding' : 'gzip,deflate',
#                   'Connection' : 'keep-alive',
#                   'Keep-Alive' : '100',
#                   'Accept-Language': 'en-us,en;q=0.5',
                   }
        conn = shttp.sHTTP.HTTPClient((self.ip, port),
                                      defaultheaders=headers,
                                      sockettimeout=10)
        try:
            conn.send_request(location, method=_type, version="HTTP/1.1")
        except HTTPUnexpectedStatus:
            pass
        return conn, conn.get_response_status()

    removequery=sre.compile(r"\?.+$")
    def retrieveHTTP(self, realurl, repository, deep = 0):
        """ Retrieve with optimized HTTP method.
        |url| Is the url to get. If we get a status 3xx it follows the
        locations |config.HTTPLib_maxRedirects| times.
        Returns |True| if the document was retrieved or |False| otherwise. """
        if self.path == '': self.path = '/'
        u = URI.match(realurl).groupdict()
        #u['sharp'] = ''
        #u['query'] = ''
        #location = uri.normlocation(u)
        location = u['path']
        #location = Repository.removequery.sub("",realurl)
        conn,response_status = self.httprequest('HEAD', location, u['host'], u['port'])
        #print "Host:",u['host'],", Location:", location, ", Deep: ",deep
        #c = conn.get_content()
        if 300 <= response_status < 400 and deep < repository.config.HTTPLib_maxRedirects:
            location = conn.headers['Location']
            conn.close()
            u = URI.match(location)
            if u:
                realurl = location
            else:
                d = URI.match(realurl).groupdict()
                d['path'] = '' #location
                d['query'] = ''
                d['sharp'] = ''
                realurl = uri.normurldict(d,d['host']) + location
            return self.retrieveHTTP(realurl, repository, deep+1)
        h = conn.headers
        if self.needRetrieve(repository, h.get('Last-Modified',None),
                             h.get('Date',None),
                             response_status):
            r = None
            try:
                conn.send_request(location, method='GET', version="HTTP/1.1")
                r = conn.get_content()
            except:
                conn.close_socket()
                conn.send_request(location, method='GET', version="HTTP/1.1")
                r = conn.get_content()
            self.save(r, h.get('Content-Type','text/plain'), repository)
            conn.close()
            return True
        else:
            conn.close()

    def retrieveFile(self, repository):
        stat = os.stat(self.path)
        if stat.st_mtime <= self.last_fetched:
            return False
        rfile=file(self.path)
        self.save(rfile, 'text/plain', repository)
        rfile.close()
        return True
    
    def retrieve(self, repository):
        """ Retrieve a document and save it in |repository |
        Returns |True| if the document was retrieved or |False| otherwise. """
        rfile = None
        #u=URI.search(self.url)
        #if not u: return False        
        #self.__dict__.update(u.groupdict())
        try:
            if not self.ip: # or many days ago since the last fetched
                if sre.match(r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$',self.host):
                    self.ip = self.host
                if self.host == 'localhost':
                    self.ip = '127.0.0.1'
                else:
                    self.ip = socket.gethostbyname(self.host)
        except socket.gaierror, (errno, errstr):
            self.errors_count = self.errors_count + 1
            warn.debug("Whilst resolving name ", self.host,
                       ". Errno ", errno, ": " , errstr)
            return None
        try:
            if self.protocol == 'http':
                return self.retrieveHTTP(self.url,repository)
            elif self.protocol == 'file':
                #self.callback.route(None, None, "callback", "Saved", site.url)
                return None
            #return self.retrieveFile(repository)
            elif self.protocol == 'push' or self.protocol == 'mailto':
               return None
            else:
                rfile=urllib2.urlopen(self.get_url())
        except (urllib2.URLError, socket.error, HTTPError,
                socket.gaierror), err:

            self.errors_count = self.errors_count + 1
            if self.errors_count >= repository.config.HTTPLib_maxRetriesHost:
                warn.debug("Removing '%s': HTTPLib_maxRetriesHost"%(self.url))
                repository.remove(self.url)
            warn.warn("Error while retrieving: ", err,
                       ". URL: '%s'"%(self.url))
            return None
        except Exception, err:
            warn.warn("Exception catched on retrieve: %s, %s"%(err,
                      err.__class__), ". URI: ", self.url)
            return None
        else:
            self.errors_count=0
        resp = rfile.info() 
        if self.needRetrieve(resp.getheader('Last-Modified'),
                             resp.getheader('Date'), resp.status):
            self.save(rfile,resp.getheader('Content-Type'),repository)
        rfile.close()

    def save(self, rfile, mime, repository, datafilter = 'mime'):
        self.last_fetched = time.time()
        repository.save(self,rfile,mime, datafilter = datafilter)

class Repository:
    """ A Document Repository Abstraction. Usage like a dictionary """
    def __init__(self, root, callback, mode = 0600, poolSize=64,
                 bufferlength=400, textfilter=None, circular_harvest = False,
                 config = None):
        if not os.path.isdir(root):
            warn.debug("Directory '%s' don't exists\n"%(root))
            if 0600 & mode == 0000:
                raise IOError, "Can't open root directory '%s'"%(root)
            os.mkdir(root, 0700)
        
        # Variables
        # root: Root of the repository
        # index_sites: maps url -> site object
        # index_times: maps time -> url (Don't need exists, every
        #  request must be inserted as site)
        # storage: maps  url -> Compressed data
        # callback: Callback distributor
        self.textfilter = textfilter
        self.callback = callback
        if len(root) > 0 and root[-1] != os.path.sep: root = root + os.path.sep
        self.root=root
        db = bsddb3.db
        dbobj = bsddb3.dbobj
        #from bsddb3 import dbobj
        #print dir(bsddb3)
        #dbobj = bsddb3.dbobj
        # Databases
        envflags = (db.DB_THREAD |
                    db.DB_INIT_MPOOL |
                    #db.DB_INIT_LOCK |
                    #db.DB_INIT_LOG |
                    #db.DB_INIT_TXN |
                    db.DB_INIT_CDB |
                    db.DB_CREATE )
        #OK:envflags=db.DB_INIT_CDB|db.DB_INIT_MPOOL|db.DB_CREATE|db.DB_THREAD
        #envflags=( db.DB_INIT_MPOOL | db.DB_CREATE )
        self.dbEnvRepository = dbobj.DBEnv()
        self.dbEnvRepository.set_lk_detect(db.DB_LOCK_DEFAULT)
        self.dbEnvRepository.open(self._path(''), envflags)
        self.index_sites = Storage(self._path('index_sites.db'),mode=mode)
        self.index_times = dbobj.DB(self.dbEnvRepository)
        #dbflags=(db.DB_CREATE | db.DB_AUTO_COMMIT)
        dbflags=(db.DB_CREATE|db.DB_THREAD)
        #dbflags=db.DB_CREATE
        #dbflags=db.DB_CREATE
        self.index_times.open(self._path('index_times.db'),
                              dbname="index_time",
                              dbtype=db.DB_RECNO,
                              flags=dbflags,
                              mode=mode)
        self.storage=dbobj.DB(self.dbEnvRepository, 0)
        self.storage.open(self._path('storage.db'),
                          dbname="storage",
                          dbtype=db.DB_BTREE,
                          flags=dbflags,
                          mode=mode)
#         self.hostlist=dbobj.DB(self.dbEnvRepository, 0)
#         self.hostlist.open(self._path('hostlist'),
#                           dbname="hostlist",
#                           dbtype=db.DB_BTREE,
#                           flags=dbflags,
#                           mode=mode)

        #self.storage=bsddb3.btopen(self._path('storage'), mode=mode)
        # Other
        self.config = config
        self.safe=PCExecution(self.sync, 3, None)
        self.last_harvest = 0.1
        self.replacePool(poolSize, False)
        self.cursor = None
        self.priorityindex = self.index_times
        #self.lock_harvest = threading.Lock()
        self.cond_harvest = threading.Condition()
        #self.lock_cursor = threading.Lock()
        self.lock_repository = threading.Lock()
        self.lock_get_next = threading.Lock()
        self.count_harvest = 0
        self.doharvest = True
        self.circular_harvest = circular_harvest
        self.bufferlength = bufferlength
        self.buffersites = None
        self.poolX.execute(None, self.__harvest)

    def sync(self):
        self.lock_repository.acquire()
        try:
            #self.hostlist.sync()
            self.storage.sync()
            self.index_sites.sync()
            self.index_times.sync()
        except:
            warn.warn("Repository.sync")
        self.lock_repository.release()
        
    def _path(self,path):
        """ Return the real path to get a |path| in the repository """
        return self.root + path

    def replacePool(self,poolSize, stop=True):
        self.poolSize = poolSize
        if stop: self.pool.close()
        self.pool = ThreadPool(poolSize)
        self.poolX = ThreadPool(2)

    def close(self):
        """ Close the repository """
        self.index_times.close()
        self.index_sites.close()
        #self.hostlist.close()
        self.storage.close()
        self.dbEnvRepository.close()
        self.cond_harvest.acquire()
        self.doharvest = False
        self.cond_harvest.notify()
        self.cond_harvest.release()
        self.pool.close()
        self.poolX.close()

    def regurllist(self, urllist):
        for i in urllist:
            self.regurl(i)

    def regurl(self, url):
        """ Register an URL """
        url = SiteInformation.removesharp.sub("",uri.normurl(url))
        if Repository.DoNotDownloadExtensions.search(url): return
        if not self.index_sites.has_key(url):
            warn.log("Registering ", url)
            s = SiteInformation(url)
	    self.lock_repository.acquire()
            #txn = self.dbEnvRepository.txn_begin(None,bsddb3.db.DB_TXN_NOWAIT)
            try:
                #s.time_index = self.index_times.append(url, txn)
                s.time_index = self.index_times.append(url)
                #self.hostlist.put(s.host, self.hostlist.get(s.host,0))
                self.index_sites[url] = s
                #txn.commit()
            except:
                #txn.abort()
                warn.warn("111111xxxxxxxxxxxxxxxxxxxxxxxxx" * 4)
	    self.lock_repository.release()
        #self.callback.route(None, None, "callback", "Registered", s.url)
    
    def remove(self, url):
        #warn.debug("REMOVE AAA")
        self.lock_repository.acquire()
        try:
            #warn.debug("REMOVE BBB")
            site = self.index_sites[url]
            #warn.debug("REMOVE CCC")
            del self.index_times[site.time_index]
            #warn.debug("REMOVE DDD")
            del self.index_sites[url]
        except Exception, exc:
            warn.debug(exc)
        self.lock_repository.release()
        #warn.debug("REMOVE EEE")

    def updateInfo(self, site):
        #warn.debug("UPDATEINFO AAA")
        #self.remove(site.url)
        #warn.debug("UPDATEINFO BBB")
        self.lock_repository.acquire()
        try:
            if self.circular_harvest:
                site.time_index = self.index_times.append(site.url)
            self.index_sites[site.url] = site
        except:
            warn.warn("222222222222222222xxxxxxxxxxxxxxxxxx"*4)
        self.lock_repository.release()
        #warn.debug("UPDATEINFO CCC")
        self.safe.changed()

    MIME_Text = sre.compile(r'^text/')
    ValidExtensions = sre.compile(r'\.(pdf\.gz|ps\.gz)$')
    # Do not download extensions
    DoNotDownloadExtensions = sre.compile(r'\.(iso|tar\.(gz|Z|bz2)|rar|ace|zip|gz|tbz2|bz2|rpm|deb|avi|jpg|gif|png|bin|exe)$')
    # No compress extensions (are valid because of push method)
    ValidExtensionsNoCompress = sre.compile(r'\.(gz|Z|rar|ace|zip|gz|tbz2|bz2|rpm|deb|avi|jpg|gif|png|)$')

    # We have not support for big files
    def _gzipHandlerSave(self, data):
        rfile = StringIO()
        zfile = GzipFile(mode='wb', fileobj=rfile)
        if hasattr(data,'read'):
            data = data.read()
        zfile.write(data)
        zfile.close()
        string = rfile.getvalue()
        rfile.close()
        return string

    def _noneHandlerSave(self, data):
        if hasattr(data,'read'):
            data = data.read()
        return data

    def _gzipHandlerOpen(self, data):
        rfile = StringIO(data)
        zfile = GzipFile(mode='rb', fileobj=rfile)
        string = zfile.read()
        zfile.close()
        rfile.close()
        return string
    
    def _noneHandlerOpen(self, data):
        return data

    def _getHandler(self, name, mime, write=True):
        if self.ValidExtensionsNoCompress.search(name):
            return write and self._noneHandlerSave or self._noneHandlerOpen
        #if self.MIME_Text.match(mime) or self.ValidExtensions.search(name):
        return write and self._gzipHandlerSave or self._gzipHandlerOpen
        #return None

    def save(self, site, data, mime, docallback=True, datafilter = 'mime'):
        """ Binds the |data| (document's data) to |site| information"""
        #warn.debug("SAVE AAA")            
        try:
            url = site.pop('url')
            site = SiteInformation(url,**site)
        except:
            # It is not a dictionary or equivalent
            warn.warn(type(site), site)
	    url = site.url
        handler = None
        if not datafilter:
            pass
        elif datafilter == 'gzip':
            handler = self._gzipHandlerSave
        elif datafilter == 'bzip2':
            handler = bz2.compress
        elif datafilter == 'mime':
            handler = self._getHandler(url, mime)
        else:
            warn.warn("Unknown filter, using none")
        site.datafilter = datafilter
        site.mime = mime
        self.lock_repository.acquire()
        try:
            if self.textfilter: data = self.textfilter(data)
            if handler:
                self.storage[url] = handler(data)
            else:
                self.storage[url] = data
	except:
            warn.warn("Repository.save")
        self.lock_repository.release()
        self.updateInfo(site)
        warn.log('URL Saved: ', url, ", mime-type:", mime)
        if docallback:
            self.callback.route(None, None, "callback", "Saved", url)
    
    def get(self, url, default=None):
        """ Return the site object of |url|.
        |get| returns compressed data """
        # return self.index_sites.get(url, default)
        x =  self.index_sites[url]
        return x
    
    def get_recno(self, time_index, default=None):
        return self.get(self.index_times.get(time_index), default)
    
    def open(self, url):
        """ Returns the document's record data and filecontent.
            |open| always returns uncompressed data """
        site = self.index_sites[url]
        #if not site: return None
        dfilter = site.datafilter
        handler = None
        if not dfilter:
            pass
        elif dfilter == 'gzip':
            handler = self._gzipHandlerOpen
        elif dfilter == 'bzip2':
            handler = bz2.decompress
        elif dfilter == 'mime':
            handler = self._getHandler(url, site.mime, False)
        else:
            warn.debug("There is no handler for url:", url,
                       ", mime:", site.mime," Using None filter")
        data = self.storage.get(url)
        if data:
            if handler:
                return handler(data)
            else:
                return data
        else:
            raise KeyError, "Repository has not " + url
    
    def getandopen(self, url):
        return (self.get(url), self.open(url))

    def __reset_cursor(self):
        self.cursor.close()
        self.cursor = None

    def __get_next(self):
        x = None
        self.lock_get_next.acquire()
        try:
            x = self.__get_next__()
        except:
            warn.warn("get_next")
        self.lock_get_next.release()
        return x
    
    def __get_next__(self):
        if self.buffersites is None:
            self.buffersites = []
        else:
            if len(self.buffersites) > 0:
                return self.buffersites.pop()
            else:
                self.buffersites = None
                return None
        i = self.bufferlength
        cursor = self.priorityindex.cursor(None, bsddb3.db.DB_WRITECURSOR)
        record = cursor.first()
        if record is None:
            return None
        while i >= 0 and record:
            cursor.delete()
            i = i - 1
            self.buffersites.append(record)
            record = cursor.next()
        cursor.close()
        return self.buffersites.pop()
        
    def set_priority_index(self, priorityindex):
        self.lock_cursor.acquire()
        self.priorityindex = priorityindex
        if not self.cursor is None: self.__reset_cursor()
        self.lock_cursor.release()
        
    def harvest(self, priorityindex = None, count = 0):
        self.cond_harvest.acquire()
        self.count_harvest += count
        self.cond_harvest.notify()
        self.cond_harvest.release()

    def __harvest(self):
        """ Try to harvest the Sites, according to the index.
        The policies can be tuned in the configuration file """
        while self.doharvest:
            self.cond_harvest.acquire()
            #warn.debug("PRE WAIT")
            self.cond_harvest.wait()
            #warn.debug("POS WAIT")
            self.cond_harvest.release()
            if not self.doharvest: break
            self.last_harvest = time.time()
            warn.log("Using %s threads "%(self.poolSize))
            try:
                self.__harvestretrieve()
            except bsddb3._db.DBError, e:
                #self.lock_cursor.release()
                warn.warn("Repository harvest ", e)
    
    def __harvestretrieve(self):
        while self.count_harvest != 0:
            #warn.debug("PRE GET NEXT")
            record = self.__get_next()
            #warn.debug("POS GET NEXT", record)
            if record is None: break
            self.count_harvest = self.count_harvest - 1
            queueposition, uri = record
            try:
                site = self.get(uri)
                #warn.debug("PRE EXECUTE")
                self.pool.execute(None, site.retrieve, self)
                #warn.debug("POS EXECUTE")
                #site.retrieve(self)
            except (IOError, AttributeError, HTTPError), err:
                warn.warn("Harvest unexpected error ",err)
