#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'Sendmail.py'
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import monarca
import monarca.storage
from monarca.storage import Storage
import smtplib
import warn
from threading import Lock
from Queue import Queue, Empty
from threadutils import PeriodicExecution, ThreadPool
import sys

class SMTPErrorSendingMail:
    pass

class SendmailRequest:
    def __init__(self, sendfrom, sendto, subject, message):
        self.sendfrom = sendfrom
        self.sendto = sendto
        self.subject = subject
        self.message = message
        self.finished = False

    def send(self, smtpconn):
        r = None
        warn.debug(self.message)
        try:
            r = smtpconn.sendmail(self.sendfrom, self.sendto,
                       "Subject: %s\n\n%s"%(self.subject,self.message))
        except smtplib.SMTPDataError, (errno, errstr):
            if errno == 250:
                self.finished = True
                r = {}
        if {} != r:
            warn.debug("Failed to send mail %s\n"%(self.sendto))
            raise SMTPErrorSendingMail, "Error on " + str(self.sendto)
        else:
            self.finished = True

class Sendmail(Queue):
    """ Send e-mails througth an open connection with an SMTP Server,
    If can't send, it saves and periodically try to flush the
    messages. Every single mail use a thread, so we don't need
    to be lazy while mail is sended. """
    def __init__(self, SMTP_Server, SMTP_Port, filename, periodic=5,
		    maxthreads=10):
        """ Initialize Sendmail object.
        |SMTP_Server| Address of the SMTP  server
        |SMTP_Port| The Port where SMTP Server is listening
        |filename| The name where we put the messages until it's sended
        |periodic| Amount of time between flushes """
        Queue.__init__(self)
        self.smtp_server = SMTP_Server
        self.smtp_port = SMTP_Port
        self.storage = Storage(filename, 1)
        self.__perexe = PeriodicExecution(self.flush,periodic,periodic)
        self.pool = ThreadPool(maxthreads)
        # Raise exception, in connect no exception
        #self.connection = smtplib.SMTP(SMTP_Server,port=SMTP_Port)
        self.connection = None
        self.__lock = Lock()
        self.locks = {}
        
    def connect(self):
        try:
            self.connection.noop()
        except Exception, e:
            print >>sys.stderr, "Is alive SMTP Server '%s' at port %s?: %s\n"%(
                self.smtp_server, self.smtp_port, e)
            try:
                self.connection=smtplib.SMTP(self.smtp_server,self.smtp_port)
            except Exception, ex:
                print >>sys.stderr, "Connecting to SMTP Server %s at %s %s\n"%(
                    self.smtp_server, self.smtp_port, ex)
                return False
        return True

    def acquire_lock(self, addr):
        lock = self.locks.get(addr, None)
        if not lock: self.locks[addr] = lock = Lock()
        lock.acquire()
        
    def release_lock(self, addr):
        lock = self.locks.get(addr, None)
        lock.release()
        if not lock.locked():
            del self.locks[addr]
        
    def send(self,sendfrom, sendto, subject, message):
        s=SendmailRequest(sendfrom, sendto, subject, message)
        self.put(s)
        self.sendOne(s)

    def sendOne(self, mail):
        self.acquire_lock(mail)
        if mail.finished:
            try:
                self.storage.pop(mail.sendto)
            except KeyError, err:
                warn.debug("Sendmail sendOne")
            return
        self.connect()
        def _whenEnd(res,exc):
            if exc == None: # No exceptions
                try:
                    self.storage.pop(mail.sendto)
                except Exception, err:
                    warn.warn(err)
            self.release_lock(mail)
        self.pool.execute(_whenEnd, mail.send, self.connection)

    def put(self, mail):
        Queue.put(self,mail)
        self.storage[mail.sendto] = mail

    def close(self):
        if self.connection:
            self.connection.quit()
        self.__perexe.stop()
        self.pool.close()
        self.storage.close()
        
    def flush(self):
        self.__lock.acquire()
        while True:
            try:
                s=self.get(timeout=3)
            except Empty, e:
                break
            self.sendOne(s)

        # Too much to do in memory?
        try:
            for k,v in self.storage.items():
                self.sendOne(v)
        except:
            warn.warn("flushing mail")
        self.__lock.release()
