#!/usr/bin/python -OO
#coding: iso-8859-1;

############################################################################
#                                                                          #
#    This file 'pyidxctl.py'
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

#import SPyRO as spyro
import sys, os
import warn
from threadutils import ThreadPool, ThreadPoolShared
from threading import Thread, Lock, Condition
from router import Routing
import SPyRO as spyro

class IndexRouter(Routing):
    def __init__(self, dist, ttl = 10):
        Routing.__init__(self, dist, ttl)

    def before(self, methodname, res, args, kwargs):
        if methodname != "search": return
        (kwargs["collectionsize"], kwargs["gdata"]) = self.route(None, args[0], "searchstats", args[0], args[1])[0]
    
    def fusion(self, methodname, results, args, kwargs):
        if methodname == "search":
            gdata = None
            resdict = {}
            for item in results:
                gdata, res = item
                if isinstance(res, dict):
                    resdict.update(res)
                else:
                    resdict.update(dict(res))
            res = resdict.items()
            res.sort(self.cmp_results)
            results = gdata, res
        elif methodname == "searchstats":
            results = reduce(self.join_stats, results)
        else:
            return results
        return [ results ]

    def join_stats(self, x, y):
        return x[0] + y[0], self.join_gdata(x[1],y[1])

    def join_gdata(self, gdatax, gdatay):
        res = dict(gdatay)
        for x in gdatax:
            y = res[x[0]]
            res[x[0]] = (max(x[1][0],y[0]), x[1][1]+y[1], x[1][2]+y[2])
        return res.items()
        
#     def join_gdataitem(self, x, y):
#         return (max(x[1][0],y[1][0]), x[1][1]+y[1][1], x[1][2]+y[1][2])
#         # return (x[0],(max(x[1][0],y[1][0]), x[1][1]+y[1][1], x[1][2]+y[1][2]))

    def cmp_results(self, x,y):
        a =  x[1] - y[1]
        if a < 0: return 1
        if a > 0: return -1
        return 0

class IdxControl:
    def __spyro_error_handler__(self,peerconn,request,exception,description):
        """ SPyRO error handler """
        warn.debug(peerconn, " ", request.__dict__, " ", exception, " ",
                   description)

    def _setPool(self,poolSize):
        self.poolSize = poolSize
        self.threadPool = ThreadPool(poolSize)
    
    def __getstate__(self):
        return {}
    
    def __init__(self, indexrouter, poolSize=7):
        self._setPool(poolSize)
        self.indexrouter = indexrouter

    def regindexer(self, idx):
        """ Register an indexer as part of a bigger indexing system """
        #self.indexrouter.register(idx, idx.vectorize())
        self.indexrouter.register(idx, None)
        warn.log("Registered %s"%(repr(idx)))

    def regindexerbyaddress(self, addressport, location, objname, auth=None):
        """ Register a new indexer but providing the (address, port) and
        the location of the server """
        try:
            comm = spyro.TClient(addressport, location)
            self.indexrouter.register(comm.getobj(objname,auth), None)
            warn.log("Registered %s:%d at %s"%(addressport[0], addressport[1],
                                               location))
        except:
            warn.warn("registering indexer by address")

    def indexFromRepositoryList(self, urllist):
        self.indexrouter.route(None, urllist, "indexFromRepositoryList", urllist)
        
    def indexFromRepository(self, url, name=None):
        self.indexrouter.route(None, url, "indexFromRepository", url, name)
                               
    def indexString(self, name, string, doc_id=None):
        """ Index an string """
        self.indexrouter.route(None, name, "indexString", name, string, doc_id)

    def indexMagicFiles(self, paths):
        """ Same as in pyidx """
        self.indexrouter.route(None, paths, "indexMagicFiles", paths)

    def indexMagicFile(self, name, path, doc_id=None):
        """ Same as in pyidx """
        self.indexrouter.route(None, path, "indexMagicFile", name, path, doc_id)

    def indexPlainFile(self, name, path, doc_id=None, bufflen=4096):
        """ Same as in pyidx """
        self.indexrouter.route(None, path, "indexPlainFile", name, path, doc_id=doc_id,
                                 bufflen=4096)
            
    def indexPlainFiles(self, path):
        """ Same as in pyidx """
        self.indexrouter.route(None, path, "indexPlainFiles", path)

    def indexXMLFile(self, name, path, doc_id=None):
        """ Same as in pyidx """
        self.indexrouter.route(None, path, "indexXMLFile", name, path, doc_id)

    def indexXMLFiles(self, path):
        """ Same as in pyidx """
        self.indexrouter.route(None, path, "indexXMLFiles", path)

    def sync(self):
        """ Synchronize indices. WISH: How many? how much?"""
        self.indexrouter.route(None, None, "sync")

    def search(self, vectorquery, vectorkeywords,
               begin, count, qtype="rank_vector"):
        res = self.indexrouter.route(None, None, "search", vectorquery,
                                     vectorkeywords, begin, count, qtype)
        return res[0]

    def close(self):
        self.threadPool.close()

    def callback(self, event, uri):
        if event == 'Saved':
            warn.log('Redirecting index request ',uri)
            if uri.startswith("file://"):
                self.indexrouter.route(None,uri,'indexMagicFile', uri, uri)
            else:
                self.indexrouter.route(None,uri,'indexFromRepository',uri)

