#!/usr/bin/python -OO
#coding: iso-8859-1;

############################################################################
#                                                                          #
#    This file 'pyidx.py'
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import sys,os
import SPyDI
import sre
import SPyRO as spyro
import warn
import Repository

from threadutils import ThreadPool, ServerInThread
#import citeseer

class DocumentCloser:
    def __init__(self, name, doc):
        self.name = name
        self.doc = doc

    def close(self, *args):
        self.doc.close()
        warn.log("Indexed ", self.name)

class DocumentIndexer:
    """ Creates a text Indexer using C binding of Python of
    the indexer module. """

    def _setPool(self, poolSize):
        self.poolSize = poolSize
        self.threadPool = ThreadPool(poolSize)

    def __init__(self, name, poolSize=30, cacheSize=1<<12,
                 divCache=3, idxCacheSize=1<<12, perm=0640,
                 repositoryrouter = None):
        self.replicas = 1 # fake argument
        self.basename = name
        self.repositoryrouter = repositoryrouter
        self.idx_info = SPyDI.cache1(name+'.idx', perm, 0, cacheSize, divCache)
        self.dat_info = SPyDI.cache1(name+'.dat', perm, 0, cacheSize, divCache)
        # WISH: change the file names inside document_env module
        self.docenv = SPyDI.document_env(self.idx_info, perm,
                                         SPyDI.DOCENV_NAME,name+".docenv")
        self.idm = SPyDI.idx_man(name + '.wrd', self.idx_info, self.dat_info,
                                 perm, idxCacheSize)
        self._setPool(poolSize)

    def __keywords(self, y):
        return " \001%s "%(y)

    def indexFromRepositoryList(self, urllist):
        for url in urllist:
            try:
                self.indexFromRepository(url)
            except:
                warn.warn("indexFromRepository:", url)
    
    def indexFromRepository(self, url, name = None):
        """ Index something from the repository """
        site, string = self.repositoryrouter.routeone(None, url, "getandopen", url)
        #print "*****************DICT:>",site.__dict__
        keywords = "\n".join(map(self.__keywords, site.keywords))
        #print "*****************KEYWRD:>",type(keywords)
        #print "*****************STRING:>",type(string)
        if name is None: name = url
        self.indexString(name, keywords + string, None)

    def indexString(self, name, string, doc_id=None):
        """ Index a string """
        if not doc_id: doc_id = SPyDI.id_doc(*self.docenv.get_id(name))
        if not string: string = ''
        string = name + string
        doc = SPyDI.document(string, SPyDI.DOCUMENT_STRING)
        self.threadPool.execute(DocumentCloser(name, doc).close,
                                doc.insert, self.docenv, self.idm, doc_id)


    def indexMagicFile(self, name, path, doc_id=None):
        """ Index a file trying to guess useful structure of the file """
        print "Not implemented... yet!"
        pass

    def indexMagicFiles(self, paths):
        """ Index a file trying to guess useful structure of the file """
        for path in paths: self.indexMagicFile(path, path)

    def indexPlainFiles(self, paths):
        for path in paths: self.indexPlainFile(path, path)

    def indexPlainFile(self, name, path, doc_id=None, bufflen=4096):
        """ Index a plain text file"""
        if not doc_id: doc_id = SPyDI.id_doc(*self.docenv.get_id(name))
        doc = SPyDI.document(path, SPyDI.DOCUMENT_FILE, bufflen)
        self.threadPool.execute(DocumentCloser(path,doc).close, doc.insert,
                                self.docenv, self.idm, doc_id)
    
    def indexXMLFile(self, name, path, doc_id=None):
        """Index a XML file. Using its structure. """
        print "En cola XML File", path
        #self.threadPool.execute(None, self.__indexXML__, path)
        oai = citeseer.CiteseerOAI(self)
        self.indexString(name, oai.insert(path, path), doc_id)
        print "Termino XML File", path

    def indexXMLFiles(self, paths):
        for path in paths: self.indexXMLFile(path, path)

    def sync(self):
        self.docenv.sync()
        self.idm.sync()
    
    def close(self):
        self.idm.close()
        self.threadPool.close()
        #idx_info e dat_info se cierran en idx_man.close()
        self.docenv.close()

    def search(self, vectorquery, begin, count, qtype="rank_vector",
               vectorkeywords = None):
        q = self.newQuery(vectorquery, vectorkeywords)
        q.init()
        gdata = q.getgdata()
        results = q.doquery(begin, count, qtype)
        q.close()
        return gdata, results

    def newQuery(self, argv, keys=None, size=None):
        """ Create a new query with argv as query terms """
        tquery = type(argv)
        if tquery in (tuple, list):
            argv = argv + keys
        elif tquery is str:
            argv = argv + " " + keys
        return SPyDI.query(self.docenv, self.idm, argv)


def test():
    addr=('0.0.0.0',59195)
    import socket
    try:
        import psyco
        psyco.profile()
    except:
        warn.debug("The psyco module is not present")

    ConnectionBroken = True
    numports = 10 
    while ConnectionBroken and numports:
        try:
            server=spyro.CommunicationServer(addr, '/SPyRO?PICKLE2')
        except socket.error, e:
            warn.warn(e)
            addr = (addr[0], addr[1]+numports)
            numports = numports - 1
        else:
            ConnectionBroken = False

    basename = 'datafiles/dbX' + str(numports) # para pruebas

    index = DocumentIndexer(basename, poolSize=16)
    #atexit.register(index.close)
    server.setobj('index',index)
    #comm=spyro.CommunicationClient(('127.0.0.1',7777),'/SPyRO?PICKLE2')
    comm=spyro.TClient(('127.0.0.1',3001),'/SPyRO?PICKLE2',local_server=server)
    idxcontrol = comm.getobj('index')
    index.repositoryrouter = comm.getobj("RepositoryRouter") 
    print "regindexer"
    idxcontrol.regindexer(index)
    comm.close()
    try:
        print "Sirviendo en ", addr
        server.serve_forever()
    except Exception, e:
        warn.warn(e)

    server.close()
    index.close()

if __name__ == '__main__':
    test()
