#coding: utf-8;
############################################################################
#                                                                          #
#    This file 'monarca/httpserver.py'                                     #
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

__doc__ = """
Package of the Monarca webserver
"""

import shttp
import shttp.http
import sre
import os
import sys
from select import select
import socket, SocketServer
from threading import Lock
from SocketServer import ThreadingTCPServer
import Queue
import time
import warn
import mime

__doc__ = """ A Simple and (hopefully) fast HTTP Server """

from StringIO import StringIO
import urllib
    
class PythonCache:
    def __init__(self, makeCode=True):
        """ Caches templates. 
        It loads and (possibily compiles) if its new
        
        Don't compile (or build python byte-code) templates if
        you use if you will transmit them with spyro"""
        self.dict = {}
        self.makeCode = makeCode

    def get(self, filename):
        mtime = os.stat(filename).st_mtime
        r = self.dict.get(filename)
        if r and r[0] >= mtime: return r[1]
        f = file(filename)
        r = f.read()
        f.close()
        if False and self.makeCode:
            source = compile(r, filename, 'single')
        else:
            source = r
        self.dict[filename] = (mtime, source)
        return source

class Monarca(shttp.http.HTTPServer):
    """ The Monarca HTTP server
    
    Monarca supports plug-ins, managed via register_handler and unregister_handler
    functionhandler(MonarcaRequest, parts)
    
    """
    def __init__(self, server_address, DocumentRoot = None,
                 WeyPath = None, config = None, globalvars = None,
                 **httpoptions):
        """
        |server_address| is the address where the server listen on
        |DocumentRoot| The document root
        |WeyPath| The path to wey controller
	|config| is the configuration pool

        Additional keyword options will be send to the constructor of the parent
        class shttp.http.HTTPServer
        """
        if DocumentRoot: httpoptions['document_root'] = DocumentRoot
        shttp.http.HTTPServer.__init__(self, server_address, **httpoptions)
        if config:
            for path in config.python_path: sys.path.insert(0, path)
        if not WeyPath:
            global __file__
            WeyPath=os.path.join(os.path.dirname(__file__), 'wey.py')
        if not globalvars: globalvars = {}
        if not globalvars.has_key("config"): globalvars['config'] = config
        self.WeyPath = WeyPath
        self.cache = PythonCache()
        self.globalvars = globalvars
        self.register_urihandler('/html', self.wey_handler)
        # self.register_urihandler('/cgi-bin', self.handle_python_cgi)

    def wey_handler(self, httpreq, environment):
        """ wey handler

        It receives:
        |httpreq| The HTTP Request Handler object
        |environment| The environment
        """
        #cambiar parts a leer el wey y __filename__ a template...
        extra_globals = {'__filename__' : httpreq.script }
        if self.globalvars: extra_globals.update(self.globalvars)
        self.handle_python_cgi(httpreq, environment, filename = self.WeyPath, extra_globals = extra_globals)

    def handle_python_cgi(self, httpreq, environment, filename = None, extra_globals = {}):
        """
        The python cgi handler:
        |httpreq| is the HTTP Request Handler object
        |environment| CGI/1.1 Environment
        |filename| the path filename of the script
        |extra_globals| extra global variables to be passed to the cgi
        """
        
        if not filename:
            filename = httpreq.translate_path(httpreq.cgi_info[1])
        source = self.cache.get(filename)
        glob = {
            '__input__': httpreq.rfile,
            '__output__': httpreq.wfile,
            '__filename__': filename,
            '__environ__': environment
        }
        glob.update(extra_globals)
        loc = {}
        exec source in glob, loc

def test(rootpath='.'):
    """ Run an http server in |rootpath| directory"""
    s = Monarca(('0.0.0.0', 8000),rootpath)
    s.serve_forever()
    
if __name__ == '__main__':
    test()
