#coding: iso-8859-1;

############################################################################
#                                                                          #
#    This file 'Storage.py'
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################


__doc__ = """ Storage in BerkeleyDB shelves. Manages threading and automatic
synchronization policies """

import sys
import warn
from threading import Thread, Lock, Condition
from shelve import BsdDbShelf
#from dbsupport import *
try:
    import bsddb3
except ImportError, err:
    try:
        import bsddb as bsddb3
    except:
        warn.warn("Error importing bsddb hopefully it could work")
from threadutils import PeriodicExecution, PCExecution

class Storage:
    """ Storage class.
    |filename| Filename of the storage database
    |sync_every| How many transactions will ocurre before synchronize
    In-Memory database with the DB on the Disk
    |sync_time| synchronize every |sync_time| seconds """
    
    def __init__(self, filename,sync_every=10,sync_time=None,mode=0600):
        """ Initialize """
        self.__db = BsdDbShelf(bsddb3.btopen(filename, mode=mode))
        self.__safeexec = PCExecution(self.sync, sync_every, sync_time)
        # needs the lock?
        self.__lock = Lock()
    
    def close(self):
        self.__safeexec.stop()
        return self.__db.close()
    
    def get(self,key,default=None):
        return self.__db.get(key,default)

    def __len__(self):
        return len(self.__db)
    
    def __getitem__(self,key):
        return self.__db[key] # No raise exception

    def sync(self):
        self.__lock.acquire()
        try:
            self.__db.sync()
        except:
            warn.warn("Storage.sync")
        self.__lock.release()
        
    def set(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self,key,value):
        self.__lock.acquire()
        try:
            self.__db[key] = value
        except:
            warn.warn("Storage.__setitem__")
        self.__lock.release()
        self.__safeexec.changed()

    def clear(self):
        return self.__db.clear()
    
    def has_key(self, k):
        return self.__db.has_key(k)

    def items(self):
        return self.__db.items()

    def keys(self):
        return self.__db.keys()

    def values(self):
        return self.__db.values()

    def pop(self, key):
        x = self.__db.pop(key)
        self.__safeexec.changed()
        return x

    def __delitem__(self, key):
        del self.__db[key]
        self.__safeexec.changed()
    
    def popitem(self, key):
        x = self.__db.popitem(key)
        self.__safeexec.changed()
        return x
    
    def setdefault(self, key, default):
        return self.__db.setdefault(key,default)

    def update(self, other):
        return self.__db.update(other)

    def values(self):
        return self.__db.values()

    def last(self):
        return self.__db.last()

    def first(self):
        return self.__db.first()

    def previous(self):
        return self.__db.previous()

    def next(self):
        return self.__db.next()
