#!/usr/bin/python
#coding: iso-8859-1;

############################################################################
#                                                                          #
#    This file 'SPyDI.py'                                                  #
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2006      by Eric Sadit Téllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import os, sys
import cgi
import SPyRO as spyro
import session
import monarca
from monarca import template
import socket
import warn as debug
import threadutils
from threadutils import ServerInThread
import warn
import sre
import os, sys

class Config:
    def __init__(self, root = None, cwd = None):
        if not cwd: cwd = os.getcwd()
        self.cwd = cwd
	if not root: root = cwd
        self.root = root
        self.python_path = [cwd, root] # the python's module path

        self.AdminMail='sadit@lsc.fie.umich.mx'
        self.datadir = self.cwd + '/datafiles/'
        # Mail shown when register 
        self.RegistryMail='register@lsc.fie.umich.mx'
        self.AuthentityFile = self.datadir+"sessions.db"
        self.PasswordFile = self.datadir+"passwd.db"
        self.SpoolMailFile = self.datadir+"spool.db"

        #SPyRO Configuration
        self.SPyRO_ServerPort = 3001
        self.SPyRO_ServerAddr = '127.0.0.1'
        self.SPyRO_Server = (self.SPyRO_ServerAddr, self.SPyRO_ServerPort)
        #Servers

        # Run HTTP Server
        self.Run_HTTP = True
        self.HTTP_Root = self.root + '/'
        self.HTTP_CGIRoot = "/cgi-bin/" # In Web Location
        self.HTTP_CGIPath = self.HTTP_Root + self.HTTP_CGIRoot
        self.python_path.append(self.HTTP_CGIPath)
        self.HTTP_IPName = True
        weypath=os.path.join(os.path.dirname(monarca.__file__), 'wey.py')
        self.httpd_options = {
            'server_address' : ('0.0.0.0',8000),
            'DocumentRoot' : self.HTTP_Root,
            'WeyPath' : weypath,
            }

        self.HTTPUserAgent = """Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.7.12) Gecko/20051010 Firefox/1.0.7 (Ubuntu package 1.0.7)"""
        # Template Directory (absolute or relative to CGI program CWD)
        self.TemplatePath = [ self.cwd + "/templates/" ]
        self.LanguagePath = [ self.cwd + "/lang/" ]
        self.DefaultTemplate = "indextemplate"

        # Pyew settings
        self.pyew_administrator = 'admin'

        #Language Settings
        self.pyewdefaults = { 'language': 'es' }

        # modules
        self.initModuleGlobal = True
        self.initModuleAuth = True
        # Use psyco is a suggested behaviour, may cause memory leaks
        self.usePsyco = False

        # names by default
        self.name_Authority = "authority"
        self.name_Passwd = "password"
        self.name_TemplateCache = "templates"
        self.name_LanguageCache = "languages"
        self.name_Sendmail = "sendmail" #faked
                                                                              
class Server:
    def __init__(self, config):
        self.config = config

    def auth_init(self):
        self.auth=session.Authority(self.config.AuthentityFile)
        self.user=session.User(self.config.PasswordFile)
        self.spyro.setobj(self.config.name_Sendmail, None)
        self.spyro.setobj(self.config.name_Authority, self.auth)
        self.spyro.setobj(self.config.name_Passwd, self.user)

    def auth_close(self):
        self.user.close()
        self.auth.close()

    def global_init(self):
        #global serv, auth, user, sendmail, temp, mesgcache
        config = self.config
        self.serv=ServerInThread(spyro.CommunicationServer,
                                 config.SPyRO_Server)
        self.spyro = self.serv.server
        self.temp=template.TemplateCache(config.TemplatePath,
                                         makeCode=True, makeCompiled = True)
        self.mesgcache=template.LanguageCache(config.LanguagePath)
        self.spyro.setobj(self.config.name_TemplateCache,self.temp)
        self.spyro.setobj(self.config.name_LanguageCache,self.mesgcache)
        self.auth = None
        self.user = None
        self.spyro.setobj(self.config.name_Sendmail, None)
        self.spyro.setobj(self.config.name_Authority, self.auth)
        self.spyro.setobj(self.config.name_Passwd, self.user)
        
    def global_close(self):
        #global serv, auth, user, sendmail, temp, mesgcache
        self.serv.close()


    def init(self):
        config = self.config
        if config.initModuleGlobal:
            self.global_init()
        if config.initModuleAuth:
            self.auth_init()
        self.httpd = None

    def run(self):
        config = self.config
        if config.Run_HTTP:
            globalvars = {
                'config' : config,
                'authorization' : self.auth,
                'passwd': self.user,
                'templates': self.temp,
                'language': self.mesgcache,
                }
            self.httpd = monarca.Monarca(config = config,
                                         globalvars = globalvars,
                                         **config.httpd_options)
            self.httpd.serve_forever()
        else:
            time.sleep(1<<30 | ((1<<30)-1))

    def close(self):
        close_list = []
        config = self.config
        if config.initModuleGlobal:
            close_list.append(self.global_close)
        if config.initModuleGlobal:
            close_list.append(self.auth_close)
        for close_function in close_list:
            try:
                close_function()
            except:
                warn.warn("Error in close function " + repr(close_function))

def run(DocumentRoot='.',Cwd='.', options = {}):
    conf = Config(DocumentRoot,Cwd)
    conf.__dict__.update(options)
    s = Server(conf)
    try:
        s.init()
        s.run()
        s.close()
    except:
        warn.warn()
        s.close()


if __name__ == '__main__':
    run()
