#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'session.py'
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

#coding: iso-8859-1
__doc__ = """ Web Site Session Management """
from md5 import md5
import os, sys
import random
import time
from crypt import crypt
from storage import Storage
from threading import Lock
import sre
from threadutils import PeriodicExecution
import warn

def getRandomString():
    return str(random.random() + time.time())

def getBetterRandomString():
    if os.path.exists("/dev/urandom"):
        f=file("/dev/urandom","r")
        s=md5(f.read(512)).hexdigest()
        f.close()
    else:
        warn.debug("""*Warning*: session.getBetterRandomString can't
        use /dev/urandom!! using a low quality replacement""")
        s=getRandomString()
    return s

SessionTimeout = 3600*3 # Timeout to expires (AuthorityId and Session)

class UserAlreadyExists(Exception): pass
class UserNotRecognized(Exception): pass
class ActivationTimeout(Exception): pass

class User(Storage):
    """ User authenticity based on password and login """
    PASSWORD=0
    NAME=1
    EMAIL=3

    def __init__(self,filename):
        """ Initialize User object.
        |filename| is the name of databasefile to authorize users """
        Storage.__init__(self,filename)
        self.noactivated=Storage(filename + ".tmp")

    def update(self, login, password = None, name = None, email = None):
        """ Update user info, variables binded to None are ignored """
        if not self.has_key(login):
            raise UserNotRecognized, "Updating user info"
        p,n,e = self[login]
        if password and not sre.search(r"^\s*$",password):
            p = self.crypt(password)
        if name: n = name
        if email: e = email
        self[login] = (p,n,e)

    def crypt(self, password):
        """ Create a crypted string of |password|"""
        return crypt(password, str(random.random()*1000))
    
    def authorize(self, user, password):
        """ Authorize user |user| with password |password|"""
        if not self.has_key(user): return False
        info = self.get(user)
        p = info[self.PASSWORD]
        return (crypt(password, p) == p)

    def activate(self, user):
        """ Activate user |user| account """
        if self.has_key(user) or not self.noactivated.has_key(user):
            raise UserAlreadyExists, "User '%s' already activated or trying to activate no registered user"%(user)
        old, regtime, regtimeout = self.noactivated[user]
        if (regtime + regtimeout) < time.time():
            del self.noactivated[user]
            raise ActivationTimeout, "User '%s' has timeout"%(user)
        del self.noactivated[user]
        self[user] = (old[0],old[1],old[2])

    def create(self, user, password, name=None, email=None,
               timetoactivate=604800):
        """ Create a new user with info |user|, |password|, |name|, |email|.
        It must be activated before |timetoactivate| seconds or will be
        removed"""
        if self.has_key(user) or self.noactivated.has_key(user):
            raise UserAlreadyExists, "User '%s' already exists"%(user)
        passwd = self.crypt(password)
        self.noactivated[user] = ((passwd,name,email), time.time(), timetoactivate)

    def delete(self,user,password):
        """ Remove an user by the same user """
        if not self.authorize(user, password):
            raise UserNotRecognized, "User identity '%(user)s' can't be proved"%locals()
        del self[user]

    def close(self):
        """ Close the User object"""
        self.noactivated.close()
        Storage.close(self)

class Authority(Storage):
    """ Authority about sessions. It knows every session. Can be used as remote
    object.

    Gets all the information in dictionary style
    
    Problems: Non-Atomic operations, Don't remove olds Sessions (easy job,
    but not done) """
    timeout = SessionTimeout
    def __init__(self, filename, timeout=None):
        Storage.__init__(self,filename)
        if timeout == None: timeout = Authority.timeout
        self.timeout = timeout
        # Last and Previous Authority Id. It's rotative over a timeout. We
        # need 2 Ids in order to authorize old sessions with almost one
        # |timeout| of age
        self.prevId = self.authId = getBetterRandomString()
        self.prevtime = self.lasttime = time.time()
        # The timeout-rotative id implementation. Must be the last instruction
        # in the Constructor
        self.perexe = PeriodicExecution(self.updateAuthId, self.timeout+1, self.timeout)
        
    def close(self):
        """ Close Authority """
        self.perexe.stop()
        Storage.close(self)
        
    def updateAuthId(self):
        """ Refresh authentication seeds and timeouts"""
        self.sync() # Save the information on cache (security)
        self.perexe.timeinterval = self.timeout
        self.prevtime = self.lasttime
        self.lasttime = time.time()
        self.prevId = self.authId
        self.authId = getBetterRandomString()

    def getId(self):
        """ Return Authentication Id """
        return self.authId

    def createAuth(self, seed, authid):
        """ Create an authority with session seed |seed| and
        authority Id |authid|"""
        return md5(seed + authid).hexdigest()

    def regsession(self, session, user):
        """ Registry a new Session |session| and binding session to |user|"""
        auth = self.createAuth(session.seed, self.authId)
        self[auth] = (session, user)
        return auth

    def authorize(self, authstr, userstr):
        """ Authorize the authentication string |authstr| and user |userstr|"""
        if not self.has_key(authstr): return False
        sess,user = self[authstr]
        if userstr == user and self.authorizeX(sess, authstr): return True
        return False

    #authstr sirve igual que el tiempo, se debe incluir otra cadena
    # una para usuario otro para validar (authstr). Por el momento solo
    # se usara una
    def authorizeX(self, session, authstr):
        """ Authorize method, lower lever """
        if session.lasttime < self.prevtime: return False
        if session.lasttime < self.lasttime:
            return authstr == self.createAuth(session.seed,self.prevId)
        else:
            return authstr == self.createAuth(session.seed,self.authId)

class Session:
    """ Session class with timeout. It must be created and then saved with a
    (fast) pickle implementation (as cPickle and a fast index as shelve).
    When it's loaded from an unpickler we need call
    |authorize| method in the authority. An authority is a (maybe remote)
    object that knows every sessions and has an AuthorityId that will be used
    in the session creation and authorization of the sessions. """
    timeout = SessionTimeout
    def __init__(self, timeout=None):
        """ Initialize Session
        |idsession| Identificator of the session. If not specified create one
        |timeout| Timeout of the session. If not specified it will be
        Session.timeout and must be less that Authority.timeout """
        self.seed = getRandomString()
        if timeout is None: timeout = Session.timeout
        self.timeout = timeout
        self.lasttime = time.time()

