#coding: iso-8859-1;
############################################################################
#                                                                          #
#    This file 'Template.py'
#    is part of 'SPyDI: Simple Python Distributed Indexing'                #
#    Copyright (C) 2004-2005 by Eric Sadit Tllez Avila                    #
#    sadit@lsc.fie.umich.mx or donsadit@gmail.com                          #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

#coding: iso-8859-1
import os,sys
import warn
from sys import argv
import sre
try:
    #from cStringIO import StringIO
    from StringIO import StringIO
except:
    from StringIO import StringIO

class Template:
    """ Load files and uses like templates. The Template File must use
    the tags <?python CODE ?> to use a handler to the code inside
    tags (Actually only |python| is supported).
    If |python| is used CODE is python code. It has full access to the process
    and access to the variables cgi, forms (FieldStorage object of the
    request) and local-variables given by the calling Template.execute.
    All ouputs must be with the echo(*args) function (global in templates)
    We can see the outputs as echo calls. So, if the indentation level is
    diferent of the current we need to change it explicitelly. Example:

    <?python import time>
    <html>
    <?python echo(time()) ?>
    </html>

    Is translated as:
    
    import time; echo(\"\"\"
    <html>
    \"\"\"%locals()); echo (time()); echo(\"\"\"
    </html>\"\"\"%locals());

    So if we need to change the indentation level, we need to make the follow
    <?python
           echo(time())
    ?>

    The indentation is the spaces following the '<?python' so if we need
    8 blank characters we need to do it explicity 
    The substitute version replace the tags by the resulting of echo calls
    inside  <?python ?>

    The syntax '%(localvariable)s' must be used to interpolate variables into
    the document HTML definition.
    Must be used in to optimize substitution of local-variable
    in a template. Useful to use graphical web design packages to create
    templates 
    NOTE: To display %(something)s is necessary to escape % as %%
    (e.g %%(something)s)
    """
    codeRE = sre.compile(r"<\? (?P<handler> python)  (?P<code> .+?) \?>",
                       sre.MULTILINE | sre.VERBOSE | sre.DOTALL)
    #quoteRE = sre.compile(r"(?!\%) %\(([\w_][\w\d_]*)\)s", sre.VERBOSE)
    quoteRE = sre.compile(r"%\(([^\)]*)\)s", sre.VERBOSE)
    endBlankLine = sre.compile(r"[\n:;,\.]\s*$")
    replaceDC = sre.compile(r'"')

    def __init__(self, filename, content):
        self.filename = filename
        self.content = content
        self.compiled = None
        self.output = StringIO()

    #varname = "___template_vars___"

    def _addvar(self, var):
        #idvar = len(self.vars)
        #self.vars.append(var)
        #return 'echo(%s[%s]);'%(self.varname,idvar)
        return 'echo("""' + Template.replaceDC.sub(r'\"', var) + '""");'

    def _quoteX(self, string, comp):
        s = self.quoteRE.search(string)
        if not s:
            comp.append(self._addvar(string))
            return comp
        ini, fin = s.span()
        comp.append(self._addvar(string[0:ini]))
        comp.append('echo(%s);'%(s.group(1)))
        return self._quoteX(string[fin:], comp)

    def _quote(self, string):
        return "".join(self._quoteX(string, []))
        
    def _compile(self, string):
        """ Create a fast version of the Template. """
        s = Template.codeRE.search(string)
        if not s:
            return self.compiled.append(self._quote(string))
        ini,fin=s.span()
        self.compiled.append(self._quote(string[:ini]))
        handler, code = s.groups()
        if handler != 'python':
            raise RuntimeException, "Unknown template handler '%s'"%(handler)
        # append ';' if it's in the same indentation level
        if not Template.endBlankLine.search(code): code = code + ';'
        self.compiled.append(code)
        self._compile(string[fin:])
        return

    def compile(self, makeCode=True):
        self.compiled = []
        #self.vars = []
        self._compile(self.content)
        self.compiled = "".join(self.compiled)
        if makeCode:
            fname = "Template " + self.filename
            self.compiled = compile(self.compiled, fname, 'exec')

    def echo(self, *args):
        self.output.write(" ".join(map(str,args)))

    def substitute(self, match):
        handler, code = match.groups()
        if handler != 'python':
            raise RuntimeException, "Unknown template handler '%s'"%(handler)
        try:
            exec code in self.gvars, self.lvars
        except Exception, e:
            warn.warn(self.filename, ": "  , code)
            raise
        return self.output.getvalue()
    
    def execute(self, global_vars, local_vars, _echo=None):
        """ Execute template with the args as local variables. """
        tmp_echo = global_vars.get('echo', None)
        if _echo == None: _echo = self.echo
        global_vars['echo'] = _echo
        self.stdout = []
        self.output.truncate(0)
        try:
            # f = file('/tmp/stemplate' + self.filename.replace("/",'-'),"wb")
            # f.write(self.compiled)
            # f.close()
            exec self.compiled in global_vars, local_vars
        except SystemExit, err:
            if err.code != 0: warn.warn("SystemExit: ", self.filename, ": ",
                                        err, " exit with code ", err.code)
        except Exception, err:
            #warn.warn("GenericException: ",self.filename, ": ", err)
            raise
        except:
            warn.warn("UnknownTypeOfException: ",self.filename, ": Unknown exception ")
        #if tmp: global_vars[self.varname] = tmp
        if tmp_echo: global_vars['echo'] = tmp_echo
        return self.output.getvalue()

class ExceptionUseOld(Exception):
    def __init__(self, item):
        Exception.__init__(self)
        self.item = item

class CacheItem:
    def __init__(self, mtime, content):
        self.mtime = mtime
        self.content = content    

class FileCache(dict):
    def _readifnew(self, olditem, filename):
        mtime = os.stat(filename).st_mtime
        if olditem and olditem.mtime >= mtime:
            raise ExceptionUseOld, olditem
        f=file(filename,"rb")
        r=f.read()
        f.close()
        return CacheItem(mtime, r)

    def get(self, filename):
        item = dict.get(self, filename, None)
        item = self._readifnew(item, filename)
        self[filename] = item
        return item

class TemplateCache:
    def __init__(self, path = ['.'],
                 makeCode=False, makeCompiled=True):
        """ Caches templates. 
        It loads and possibily compiles if its needed
        
        Don't compile (or build python byte-code) templates if
        you use if you will transmit them with spyro"""
        self.cache = FileCache()
        self.path = path
        self.makeCode = makeCode
        self.makeCompiled = makeCompiled

    def get(self, filename):
        for base in self.path:
            fname = os.path.join(base, filename)
            try:
                return self._get(fname)
            except SyntaxError, err:
                raise
            except OSError, err:
                continue
        raise "Unable to get template %s"%(filename)
    
    def _get(self, filename):
        try:
            item = self.cache.get(filename)
        except ExceptionUseOld, exc:
            return exc.item.content
        f = Template(filename, item.content)
        if self.makeCompiled: f.compile(self.makeCode)
        item.content = f
        return item.content

class LanguageCache:
    def __init__(self, path = ['.']):
        self.msgdict = {}
        self.path = path
        self.cache = FileCache()

    def _get(self, name):
        for base in self.path:
            try:
                filename = os.path.join(base,name)
                item = self.cache.get(filename)
                return item
            except ExceptionUseOld:
                raise
            except:
                pass
        warn.debug("No language found in LanguagePath")
        
    def get(self, lang, *modules):
        for m in modules:
            try:
                item = self._get(m)
            except ExceptionUseOld:
                continue
            gvars = {}
            item.content = eval(item.content, gvars, gvars)
            for lang, msgs in item.content.items():
                langdict = self.msgdict.get(lang, {})
                for name, content in msgs.items():
                    langdict[".".join((m,name))] = content
                if self.msgdict.has_key(lang):
                    self.msgdict[lang].update(langdict)
                else:
                    self.msgdict[lang] = langdict
            item.content = None
        return self.msgdict[lang]
